/*
 *  consts.c(pp) - constant variable/array definitions
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "cmpile.h"
#include "hobbsh.h"



// Misc. definitions

//const char *empty_str = "\0";


// Sizes of Yazoo data types

const Ulong PrimitiveSizes[10] = { sizeof(unsigned char), sizeof(signed char), sizeof(unsigned short), sizeof(signed short),
                                sizeof(unsigned long), sizeof(signed long), sizeof(float), sizeof(double), 0, sizeof(linkedlist)  };


                                
// For cmpile.c(pp)

const double MaxInteger[6] = { 255, 127, 65535, 32767, 4294967295., 2147483647. };
const double MinInteger[6] = { 0, -128, 0, -327678, 0, -2147483648. };



// The next two lists match bytecode command words (in the CommandIDs[] list) to written names (the CommandNames[] list).

const Slong CommandIDs[written_commands_num] = 
    {  exit_Yazoo, func_return, get_args, user_function,
            step_to_memberID, resize, step_to_all, step_to_all, step_to_all, step_to_all,
                    insert_indices_up, insert_indices_down, step_to_index, define, sub_code, this_var, that_var, no_var,
            if_eq_at, if_eq_at, if_ne_at, if_ne_at, if_eq, if_ne, if_ge, if_gt, if_le, if_lt,
            def_equate_at, def_equate_at, var_define, member_define, proxy_define, def_equate, append_code, equate_at, equate_at,
                    forced_equate, equate, remove_member, delete_indices,
            if_and, if_or, if_xor, if_not,
            addf, subf, mulf, divf, powerf, modi,
            code_block, constant_string,
            ubyte_cmd, sbyte_cmd, ushort_cmd, sshort_cmd, ulong_cmd, slong_cmd, float_cmd, double_cmd, block_cmd, string_cmd,
            code_marker, code_marker, code_number  };


// Note that the '= @' must come before the '=' in the CommandNames[] list, and the ':= @' before ':=' (which precedes :),
// in order for the parser to be able to find it.  Also, '/=' must come before '/'.
// In instances where symbols overlap (e.g. *, -) only one is listed, and the other is handled separately.

const char* CommandNames[written_commands_num] =
            {  "exit", "return", "args", "(",
                    ".", "[^", "[*]", "[ *]", "[* ]", "[ * ]", "[+", "+[", "[", "::", "<<", "this", "that", "nothing",
                        "==@", "== @", "/=@", "/= @", "==", "/=", ">=", ">", "<=", "<",
                    ":=@", ":= @", "@::", "*::", "#::", ":=", ":", "=@", "= @", "=!", "=", "remove", "delete",
                    "and", "or", "xor", "not",
                    "+", "-", "*", "/", "^", "mod",
                        "{", "\42",
                    "ubyte", "sbyte", "ushort", "sshort", "ulong", "slong", "single", "double", "block", "string",
                    ";", "code", "#"};


// ControlCommandNames[] contains the names for control statements (which compile into sequences involving gotos, etc.)

const char* ControlCommandNames[control_commands_num] =
         {  "if", "elseif", "else", "end if", "endif",
            "while", "end while", "endw", "do", "until",
            "for", "in", "step", "end for", "endf"  };


// BuiltInCommandNames[]:  holds the names of the built-in Yazoo functions, in order of ID number.

const char* BuiltInCommandNames[bi_commands_num] =
        {  "call", "compile", "transform", "load", "save", "input", "print", "read_string", "print_string", "clip", "trap",
            "throw", "top", "size", "abs", "round_down", "round_up", "log", "cos", "sin", "tan", "acos", "asin", "atan", "random", "extract", "find",
            "variable_type", "variable_member_top", "variable_code_top", "member_type", "member_code_top",
            "member_indices", "member_ID", "if_member_targeted", "if_hidden_member", "variable_code", "member_code",
            "variable_code_ID", "member_code_ID", "variable_code_offset", "member_code_offset", "GetParentScriptInfo", "SpringCleaning"  };


// InternalVarNames[]:  the names of the registers

const char* InternalVarNames[internal_vars_num] =
         {  "R_slong", "R_double", "R_string", "R_composite",
            "R_error_code", "R_error_index", "R_error_script", "R_warning_code", "R_warning_index", "R_warning_script"  };


// OoOlevel[]:  gives the order-of-operations level for each bytecode command.

const char OoOlevel[commands_num] = {
    OoO_nolevel, OoO_command, OoO_no_args, OoO_stepvar, OoO_nolevel,
        OoO_stepvar, OoO_no_args, OoO_nolevel, OoO_nolevel, OoO_nolevel,
    OoO_no_args, OoO_define, OoO_define, OoO_stepvar, OoO_subcode,              // 10-
        OoO_append, OoO_command, OoO_no_args, OoO_stepvar, OoO_stepvar,
    OoO_stepvar, OoO_stepvar, OoO_stepvar, OoO_stepvar, OoO_stepvar,            // 20-
        OoO_command, OoO_no_args, OoO_no_args, OoO_no_args, OoO_no_args,
    OoO_cond_test, OoO_cond_test, OoO_cond_test, OoO_cond_test, OoO_cond_test, // 30-
        OoO_cond_test, OoO_lbool, OoO_lbool, OoO_lbool, OoO_lnot,
    OoO_cond_test, OoO_cond_test, OoO_addsub, OoO_addsub, OoO_muldiv,                   // 40-
        OoO_muldiv, OoO_pow, OoO_muldiv, OoO_nolevel, OoO_nolevel,
    OoO_no_args, OoO_no_args, OoO_no_args, OoO_no_args, OoO_no_args,           // 50-
        OoO_no_args, OoO_no_args, OoO_no_args, OoO_no_args, OoO_no_args,
    OoO_no_args, OoO_no_args, OoO_block, OoO_no_args, OoO_no_args, OoO_no_args  };   // 60-


// OpDirection[] tells whether the operators in each order-of-ops level are grouped left-to-right or right-to-left.
// For example, a = b = c is grouped:  a = (b = c), so the grouping should be right-to-left.  From the OoOlevel[] array above,
// def-general has level 12 (the levels are defined in cmpile.c(pp)), which gives r_to_l in the OpDirection[] array.

const char OpDirection[OoOLevels] = { r_to_l, l_to_r, l_to_r, l_to_r, l_to_r,        // should make 3-4 one OoO?
                                    l_to_r, l_to_r, l_to_r, l_to_r, l_to_r,
                                      l_to_r, r_to_l, r_to_l, r_to_l, r_to_l,
                                    l_to_r, l_to_r };



// BIF_CanStartSentence[op]:  determines whether each operator can start a sentence.
// If not (and if it begins a sentence), a token is created, if possible.

Boolean BIF_CanStartSentence[bi_commands_num] =
                 {  yz_true, yz_true, yz_true, yz_true, yz_true, yz_false, yz_true, yz_true, yz_true,
                    yz_true, yz_true, yz_true, yz_false, yz_false, yz_false, yz_false, yz_false, yz_false, yz_false,
                    yz_false, yz_false, yz_false, yz_false, yz_false, yz_false, yz_false, yz_false, yz_false, yz_false,
                    yz_false, yz_false, yz_false, yz_false, yz_false, yz_false, yz_false, yz_false, yz_false, yz_false,
                    yz_false, yz_false, yz_false, yz_true, yz_true  };



// left_args[] and right_args[] determine the type of arguments that each operator can take on the left/right.
// For example, equate takes a variable on the left, and any type of data (variable, constant) on the right.
// The elements of these lists correspond to the various AllowedCMDJumpTable[]s below.

char left_args[commands_num] =
{   no_arg, no_arg, no_arg, no_arg, no_arg,
        var_arg, no_arg, no_arg, no_arg, no_arg,
    no_arg, var_arg, var_arg, var_arg, var_arg,
        code_arg, no_arg, no_arg, var_arg, var_arg,
    var_arg, var_arg, var_arg, var_arg, var_arg,     // 20
        no_arg, no_arg, no_arg, no_arg, no_arg,
    data_arg, data_arg, data_arg, data_arg, data_arg,
        data_arg, log_arg, log_arg, log_arg, no_arg,
    var_arg, var_arg, data_arg, data_arg, data_arg,
        data_arg, data_arg, data_arg, no_arg, no_arg,
    no_arg, no_arg, no_arg, no_arg, no_arg,         // 50
        no_arg, no_arg, no_arg, no_arg, no_arg,
    no_arg, no_arg, no_arg, no_arg, no_arg, no_arg     };

char right_args[commands_num] =
{   no_arg, var_arg, no_arg, no_arg, no_arg,
        no_arg, no_arg, no_arg, log_arg, log_arg,
    no_arg, type_arg, data_arg, data_arg, code_arg,   //var_arg,
        code_arg, var_arg, no_arg, no_arg, no_arg,
    no_arg, no_arg, no_arg, no_arg, no_arg,          // 20
        var_arg, no_arg, no_arg, no_arg, no_arg,
    data_arg, data_arg, data_arg, data_arg, data_arg,
        data_arg, log_arg, log_arg, log_arg, log_arg,
    var_arg, var_arg, data_arg, data_arg, data_arg,
        data_arg, data_arg, data_arg, no_arg, no_arg,
    no_arg, no_arg, no_arg, no_arg, no_arg,          // 50
        no_arg, no_arg, no_arg, no_arg, no_arg,
    no_arg, no_arg, data_arg, no_arg, no_arg, no_arg       };



// Jump tables/lists for determining whether a given command is allowed, based on the type of command
// expected at the given position (variable, data, etc.).
// The functions contained here are called by CheckHobbish(), in hobbsh.c(pp), when checking code.
// This table is also used, as a list (i.e. the elements are just compared to &IllegalCmd, not run) by cmpile.c(pp).

void(*AllowedCMDJumpTable[6][commands_num])(Slong) =  {

{   &IllegalCmd, &Check_YazooCmd, &IllegalCmd, &Check_FunctionCmd, &IllegalCmd,     // 'starting' commands
        &Check_FunctionCmd, &IllegalCmd, &Check_JumpCmd, &Check_JumpCmd, &Check_JumpCmd,
    &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
        &IllegalCmd, &Check_YazooCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &Check_IndicesCmd, &Check_IndicesCmd, &Check_IndicesCmd,
        &Check_YazooCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,         // code block is illegal; compiler adds :=@ func() to it.
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &Check_YazooCmd      },

{   &IllegalCmd, &IllegalCmd, &Check_LongCmd, &Check_FunctionCmd, &IllegalCmd,    // variable
        &Check_FunctionCmd, &Check_YazooCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
        &IllegalCmd, &IllegalCmd, &Check_LongCmd, &Check_LongCmd, &Check_IndicesCmd,
    &Check_IndicesCmd, &Check_YazooCmd, &Check_IndicesCmd, &Check_IndicesCmd, &Check_IndicesCmd,
        &IllegalCmd, &IllegalCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd         },

{   &IllegalCmd, &IllegalCmd, &Check_LongCmd, &Check_FunctionCmd, &IllegalCmd,     // logical commands
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
        &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
    &Check_YazooCmd, &Check_YazooCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd         },

{   &IllegalCmd, &IllegalCmd, &Check_LongCmd, &Check_FunctionCmd, &IllegalCmd,      // data (var, const. string, etc)
        &Check_FunctionCmd, &Check_YazooCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
        &IllegalCmd, &IllegalCmd, &Check_LongCmd, &Check_LongCmd, &Check_IndicesCmd,
    &Check_IndicesCmd, &Check_YazooCmd, &Check_IndicesCmd, &Check_IndicesCmd, &Check_IndicesCmd,
        &IllegalCmd, &IllegalCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
        &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &IllegalCmd, &IllegalCmd,
    &Check_LongCmd, &Check_DoubleCmd, &Check_StringCmd, &IllegalCmd, &Check_YazooCmd,    // Check_CodeCmd not allowed; compiler adds :=
        &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
    &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &IllegalCmd         },

{   &IllegalCmd, &IllegalCmd, &Check_LongCmd, &Check_FunctionCmd, &IllegalCmd,          // code
        &Check_FunctionCmd, &Check_YazooCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
        &Check_YazooCmd, &IllegalCmd, &Check_LongCmd, &Check_LongCmd, &Check_IndicesCmd,
    &Check_IndicesCmd, &Check_YazooCmd, &Check_IndicesCmd, &Check_IndicesCmd, &Check_IndicesCmd,
        &IllegalCmd, &IllegalCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &Check_CodeCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd         },

{   &IllegalCmd, &IllegalCmd, &Check_LongCmd, &Check_FunctionCmd, &IllegalCmd,       // any type definition (code, var, etc.)
        &Check_FunctionCmd, &Check_YazooCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
        &Check_YazooCmd, &IllegalCmd, &Check_LongCmd, &Check_LongCmd, &Check_IndicesCmd,
    &Check_IndicesCmd, &Check_YazooCmd, &Check_IndicesCmd, &Check_IndicesCmd, &Check_IndicesCmd,
        &IllegalCmd, &IllegalCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
    &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
        &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd, &IllegalCmd,
    &IllegalCmd, &IllegalCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
        &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &IllegalCmd, &IllegalCmd,
    &Check_LongCmd, &Check_DoubleCmd, &Check_StringCmd, &Check_CodeCmd, &Check_YazooCmd,
        &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd,
    &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &Check_YazooCmd, &IllegalCmd         }  };



// This is the code that, when run, creates Zero and the top of Yazoo's memory.

const char *ZeroWrittenCode = "\
this[1] :: {\
this[1] :: slong,\
this[2] :: double,\
this[3] :: string,\
this[4] :: {},\
this[5] :: ulong,\
this[6] :: ulong,\
this[7] :: {},\
this[8] :: ulong,\
this[9] :: ulong,\
this[10] :: {} },\
this[2] :: {}\
\0";
