/*
 *  extrnl.c(pp) - built-in Yazoo functions
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <math.h>
#include "extrnl.h"
#include "userfn.h"
#include "yazooh.h"
#include "cmpile.h"
#include "intrpt.h"
#include "hobbsh.h"
#include "yzmain.h"
#include <limits.h>




// FileReadBufferSize:  number of bytes to fread() at a time when loading a file
// don't set too high because that has given PCs trouble, in my experience

const Ulong FileReadBufferSize = 10000;



// Jump tables for the built-in functions and variables.

void(*BuiltInFunctionJumpTable[bi_commands_num])() =
         {  &ext_call, &ext_compile, &ext_transform, &ext_load, &ext_save,
                &ext_input, &ext_print, &ext_read_string, &ext_print_string, &ext_clip,
            &ext_trap, &ext_throw, &ext_top, &ext_size, &ext_abs,
                &ext_round_down, &ext_round_up, &ext_log, &ext_cos, &ext_sin,
            &ext_tan, &ext_acos, &ext_asin, &ext_atan, &ext_random,
                &ext_extract, &ext_find, &ext_YZ_variable_type, &ext_YZ_variable_member_top, &ext_YZ_variable_code_top,
            &ext_YZ_member_type, &ext_YZ_member_code_top, &ext_YZ_member_indices, &ext_YZ_member_ID, &ext_YZ_if_member_targeted,
                &ext_YZ_if_hidden_member, &ext_YZ_get_variable_code, &ext_YZ_get_member_code, &ext_YZ_get_variable_code_ID, &ext_YZ_get_member_code_ID,
            &ext_YZ_get_variable_code_offset, &ext_YZ_get_member_code_offset, &ext_YZ_ParentScriptInfo, &SpringCleaning  };

void(*InternalVarJumpTable[internal_vars_num])() =
         {  &ext_slong, &ext_double, &ext_string, &ext_composite,
            &ext_error_code, &ext_error_index, &ext_error_script, &ext_warning_code, &ext_warning_index, &ext_warning_script  };



yz_external_global_struct yz_external_globals;



// The built-in Yazoo functions follow.
// Each is titled 'ext_', then the Yazoo name of the function.


// ext_call():  runs a user-defined function in, or referenced in, userfn.c(pp).  Calls UserFunctionCall().

void ext_call()  {  ExtCallMode = full_call;  UserFunctionCall();  }

void UserFunctionCall()
{
    bundle *FNameBundle, *loopArgBundle;
    linkedlist *FunctionNameString;
    char **PointerList, **ArgvHandles;
    Ulong FunctionID, char_counter, ArgCounter, args_num, DataBundleNum, StringNum, DataSize, StringSize;
    arg_info *InfoList, *ArgvInfo;
    Boolean match;
    
    args_num = MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr);
    if (args_num < 1)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
    
    
        // get first argument:  the function name/ID number
        
    FNameBundle = GetBIF_Member(1);
    if (FNameBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (FNameBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    

        // if the user passed a name, then find its position in the UserFunctionSet list

    if (FNameBundle->variable_ptr->type == string_type)    {
    
        FunctionNameString = (linkedlist *) FElement(&(FNameBundle->variable_ptr->mem.data), FNameBundle->bundle_offset+1);    
        if (FunctionNameString->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }

        for (FunctionID = 0; FunctionID < UF_Top; FunctionID++)   {
            match = yz_true;
            for (char_counter = 0; char_counter < FunctionNameString->elementnum; char_counter++)   {
                if ((*(char *) FElement(FunctionNameString, char_counter+1) != *(UserFunctionSet[FunctionID].FunctionName+char_counter)) ||
                                    (*(UserFunctionSet[FunctionID].FunctionName+char_counter) == 0))   {
                    {  match = yz_false;  break;  }    // 2nd condition to prevent crash if 0 is in the FunctionID string
            }   }
            if (*(UserFunctionSet[FunctionID].FunctionName+char_counter) != 0)  match = yz_false;
            if (match == yz_true)  break;     }
    
        if (FunctionID == UF_Top)
            {  SetError(user_function_not_found_err, PCCodePtr-1);  return;  }    }
        
        
        // if the user passed the position directly, we have less work to do
    
    else if (FNameBundle->variable_ptr->type <= double_type)  {

        if (FNameBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
        LoadDPRegister(FElement(&(FNameBundle->variable_ptr->mem.data), FNameBundle->bundle_offset+1), FNameBundle->variable_ptr->type);
        if (!((DPRegister >= 1) && (DPRegister <= UF_Top)))  {  SetError(user_function_not_found_err, PCCodePtr-1);  return;  }
        FunctionID = (Ulong) DPRegister - 1;     }

    else  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    
    
        // Now count the number of primitive arrays to pass to the function (note that variables count as arrays of 1)
        // These come from arguments 2-the last.
    
    DataBundleNum = 0;
    StringNum = 0;
    DataSize = 0;
    StringSize = 0;
    
    for (ArgCounter = 2; ArgCounter <= args_num; ArgCounter++)  {
        loopArgBundle = GetBIF_Member(ArgCounter);
        if (loopArgBundle != NULL)   {
            CountDataLists(loopArgBundle, loopArgBundle->bundle_offset, loopArgBundle->instances, (void *) &DataBundleNum, (void *) &StringNum);
            if (err_code != passed)  break;
    }   }
    if (err_code != passed)  return;
    
    
        // Allocate space both for our argument pointers and for the info list that Yazoo provides.
    
    PointerList = (char **) malloc((DataBundleNum+StringNum+1)*sizeof(char *));
    if (PointerList == NULL)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    InfoList = (arg_info *) malloc((DataBundleNum+StringNum)*sizeof(arg_info) + 4);     //  avoid size-0 malloc
    if (InfoList == NULL)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    ArgvHandles = PointerList;
    ArgvInfo = InfoList;
    
    
        // Now fill the argument pointer/info lists.
    
    for (ArgCounter = 2; ArgCounter <= args_num; ArgCounter++)  {
        loopArgBundle = GetBIF_Member(ArgCounter);
        if (loopArgBundle != NULL)  {
            Fill_argv_Handle(loopArgBundle, loopArgBundle->bundle_offset, loopArgBundle->instances, (void *) &ArgvHandles, (void *) &ArgvInfo);
            if (err_code != passed)  break;
    }   }
    *(PointerList+DataBundleNum+StringNum) = (char *) InfoList;
    
    
        // Set the return coordinate to R_double:  the number that the function returns.
    if (err_code == passed)   {
        *LL_Double(&(BIF_ReturnBundle(R_double)->variable_ptr->mem.data), 1) =
                    UserFunctionSet[FunctionID].FunctionPtr(DataBundleNum, (char **) PointerList);
        SetReturnCoordinate(R_double);    }
    
    free((void *) PointerList);     // clean up
    free((void *) InfoList);
}


// ext_load() reads the contents of a file into R_string.

void ext_load()
{
    bundle *FileBundle;
    linkedlist *FileNameLL, *InputStringLL;;
    size_t bytes_read;
    long file_size;
    FILE *ToReadFrom;
    char *FileString, close_retrn = passed;
    LL_retrn linkedLretrn;

    
        // get the name of the file (arg 1)
    
    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 1)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }

    FileBundle = GetBIF_Member(1);
    if (FileBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (FileBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    if (FileBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    FileNameLL = (linkedlist *) FElement(&(FileBundle->variable_ptr->mem.members), FileBundle->bundle_offset+1);
    if (FileNameLL->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }
    
    
        // copy the file name, plus a null terminator, into a separate buffer
    
    FileString = (char *) malloc(FileNameLL->elementnum+1);
    if (FileString == NULL)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    FGetElements(FileNameLL, 1, FileNameLL->elementnum, FileString);
    *(FileString + FileNameLL->elementnum) = 0;
    
    
        // try to open the file; throw an error if unsuccessful
    
    ToReadFrom = fopen(FileString, "rb");
    if (ToReadFrom == NULL)  {  SetError(IO_error, PCCodePtr-1);  }
        
    else  {    

            // get the size of the file
        
        fseek(ToReadFrom, 0, SEEK_END);
        file_size = ftell(ToReadFrom);
        rewind(ToReadFrom);

            // allocate memory and read the file
        
        InputStringLL = (linkedlist *) FElement(&(BIF_ReturnBundle(R_string)->variable_ptr->mem.data), 1);
        if (InputStringLL->memory != NULL)  FDeleteLinkedList(InputStringLL);
        
        linkedLretrn = FNewLinkedList(InputStringLL, file_size, sizeof(char), 0, uncleared);
        if (linkedLretrn == passed)  linkedLretrn = FDefragmentLinkedList(InputStringLL);
        if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }

        if (file_size > 0)    {
            bytes_read = fread(FElement(InputStringLL, 1), 1, file_size, ToReadFrom);
            if (bytes_read != file_size)  {  SetError(IO_error, PCCodePtr-1);  }    }
        
        close_retrn = fclose(ToReadFrom);   }
    if ((err_code == passed) && (close_retrn != 0))  {  SetError(IO_error, PCCodePtr-1);  }
    
    free((void *) FileString);      // clean up

    SetReturnCoordinate(R_string);
}


// ext_input() copies input from the keyboard, up until the first end-of-line, into R_string.

void ext_input()
{
    linkedlist *InputStringLL;
    Ulong counter;
    char CharBuffer[1001];
    char *read_retrn;
    size_t bytes_read;
    Boolean done;
    LL_retrn linkedLretrn;
    
    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 0)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }

    
        // initialize R_string
    
    InputStringLL = (linkedlist *) FElement(&(BIF_ReturnBundle(R_string)->variable_ptr->mem.data), 1);
    if (InputStringLL->memory != NULL)  FDeleteLinkedList(InputStringLL);

    linkedLretrn = FNewLinkedList(InputStringLL, 0, sizeof(char), 0, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    
    
        // loop until all of the data has been read
    
    done = yz_false;
    do  {
    
            // load the data into the inlined buffer (on the stack)
        
        bytes_read = FileReadBufferSize;
        read_retrn = fgets(CharBuffer, FileReadBufferSize+1, stdin);
        if (read_retrn == NULL)  {  SetError(IO_error, PCCodePtr-1);  break;  }
        for (counter = 0; counter < FileReadBufferSize; counter++)  {
            if (*(CharBuffer+counter) == '\n')  {
                done = yz_true;
                bytes_read = counter;    // don't include the terminator
                break;
        }   }


            // copy from the buffer into R_string
        
        linkedLretrn = FAddElements(InputStringLL, bytes_read, uncleared);         // will be contiguous if spare_room = 0
        if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
        FSetElements(InputStringLL, InputStringLL->elementnum-bytes_read+1, InputStringLL->elementnum, CharBuffer);
    }  while (done == yz_false);
    
    FDefragmentLinkedList(InputStringLL);

    SetReturnCoordinate(R_string);
}


// ext_save() saves a string into a file.  Overwrites the file if it already exists.

void ext_save()
{
    FILE *ToWriteTo;
    bundle *FileBundle, *StringBundle;
    linkedlist *FileNameLL, *StringLL;
    char *FileString, close_retrn;
    Ulong BytesWritten;
    LL_retrn linkedLretrn;

    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 2)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }


        // get first argument:  file name
    
    FileBundle = GetBIF_Member(1);
    if (FileBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (FileBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    if (FileBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    FileNameLL = (linkedlist *) FElement(&(FileBundle->variable_ptr->mem.members), FileBundle->bundle_offset+1);
    if (FileNameLL->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }


        // get second argument:  string to write
    
    StringBundle = GetBIF_Member(2);
    if (StringBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (StringBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    if (StringBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    StringLL = (linkedlist *) FElement(&(StringBundle->variable_ptr->mem.members), StringBundle->bundle_offset+1);
    if (StringLL->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }
    
    
        // defragment the string; copy the file name into a separate buffer
    
    linkedLretrn = FDefragmentLinkedList(StringLL);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }

    FileString = (char *) malloc(FileNameLL->elementnum+1);
    if (FileString == NULL)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    FGetElements(FileNameLL, 1, FileNameLL->elementnum, FileString);
    *(FileString + FileNameLL->elementnum) = 0;
    
    
        // try to open the file; throw error if unsuccessful
    
    ToWriteTo = fopen(FileString, "wb");
    if (ToWriteTo == NULL)  {  SetError(IO_error, PCCodePtr-1);  free((void *) FileString);  return;  }
    
    
        // write the string to the file
    
    if (StringLL->elementnum > 0)  BytesWritten = fwrite(FElement(StringLL, 1), 1, StringLL->elementnum, ToWriteTo);
    else  BytesWritten = 0;
    if (BytesWritten != StringLL->elementnum)  {  SetError(IO_error, PCCodePtr-1);  }
    
    
        // close the file
    
    close_retrn = fclose(ToWriteTo);
    if ((err_code == passed) && (close_retrn != 0))  {  SetError(IO_error, PCCodePtr-1);  }

    free((void *) FileString);    // clean up
    SearchCoordinate._Bundle = NULL;
}


// ext_print_string() writes the contents of its arguments into a string.
// Numeric data is written out in ASCII.

void ext_print_string()
{
    bundle *BundleBundle, *StringBundle, *PrecisionBundle;
    linkedlist *theString;
    char *StringPositionMarker;
    Ulong DataSize, StringSize, counter, ArgsNum, StringBundlePosition;
    linkedlist TempPrintLL;
    LL_retrn linkedLretrn;

    ExtCallMode = no_call;
    ArgsNum = MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr);
    if (ArgsNum < 1)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
    
    
        // read in the (optional) floating-point precision (first argument if one is given); otherwise use default 
    
    PrecisionBundle = GetBIF_Member(1);
    if (PrecisionBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (PrecisionBundle->variable_ptr->type <= double_type)  {
        StringBundlePosition = 2;
        if (ArgsNum < 2)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
        if (PrecisionBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
        LoadDPRegister(FElement(&(PrecisionBundle->variable_ptr->mem.data), PrecisionBundle->bundle_offset+1), PrecisionBundle->variable_ptr->type);
        if (!((DPRegister >= 0) && (DPRegister <= 255)))  {  SetError(out_of_range_err, PCCodePtr-1);  return;  }
        GL_MaxDigits = (Ulong) DPRegister;      }
    else if (PrecisionBundle->variable_ptr->type == string_type)  {
        StringBundlePosition = 1;
        GL_MaxDigits = 0;     }
    else  {  SetError(string_expected_err, PCCodePtr-1);  return;  }


        // get the destination string (1st or 2nd argument, depending on whether a precision was specified)

    StringBundle = GetBIF_Member(StringBundlePosition);
    if (StringBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (StringBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    if (StringBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    theString = (linkedlist *) FElement(&(StringBundle->variable_ptr->mem.data), StringBundle->bundle_offset+1);
    
    
        // compute the byte size of the final string without printing it yet
    
    DataSize = 0;  StringSize = 0;
    for (counter = StringBundlePosition+1; counter <= ArgsNum; counter++)  {
        BundleBundle = GetBIF_Member(counter);
        if (BundleBundle != NULL)  {
            SizeBundleString(BundleBundle, BundleBundle->bundle_offset, BundleBundle->instances, &DataSize, (void *) &StringSize);
    }   }
    
    
        // allocate contiguous memory in the target string

    linkedLretrn = FNewLinkedList(&TempPrintLL, DataSize+StringSize, sizeof(char), 0, uncleared);
    if (linkedLretrn == passed)  linkedLretrn = FDefragmentLinkedList(&TempPrintLL);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }


        // print the string
    
    if (DataSize+StringSize > 0)  {
        StringPositionMarker = (char *) FElement(&TempPrintLL, 1);
        for (counter = StringBundlePosition+1; counter <= ArgsNum; counter++)  {
            BundleBundle = GetBIF_Member(counter);
            if (BundleBundle != NULL)  {
                ReadBundleString(BundleBundle, BundleBundle->bundle_offset, BundleBundle->instances,
                                                (void *) &StringPositionMarker, (void *) &StringSize);
            if (warning_code != passed)  warning_index--;
            if (err_code != passed)  {  FDeleteLinkedList(&TempPrintLL);  return;  }
    }   }   }
    
    
        // rewire the output string to point to the linked list we just created
    
    if (theString->memory != NULL)  FDeleteLinkedList(theString);
    *theString = TempPrintLL;

    SearchCoordinate._Bundle = NULL;
}


// ext_read_string() copies data from a string into its subsequent argument variables

void ext_read_string()
{
    bundle *BundleBundle, *StringBundle;
    linkedlist *theString;
    char *StringPositionMarker, *HoldString, *StringOverflow, OneLetterType;
    Ulong counter, ArgsNum;

    ExtCallMode = no_call;
    ArgsNum = MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr);
    if (ArgsNum < 1)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
    
    
        // read in the source string:  argument 1
    
    StringBundle = GetBIF_Member(1);
    if (StringBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (StringBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    if (StringBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    theString = (linkedlist *) FElement(&(StringBundle->variable_ptr->mem.data), StringBundle->bundle_offset+1);
    if (theString->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }
    
    
        // make a copy of the source string, just in case the copy process would overwrite the source string
    
    HoldString = (char *) malloc(theString->elementnum+1);
    if (HoldString == NULL)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    FGetElements(theString, 1, theString->elementnum, (void *) HoldString);
    *(HoldString+theString->elementnum) = 0;
    StringOverflow = HoldString+theString->elementnum;     // the end of the string


        // get past any weird characters at the beginning
    
    StringPositionMarker = HoldString;
    OneLetterType = lettertype(*StringPositionMarker);
    while ((OneLetterType != a_symbol) && (OneLetterType != a_number) && (OneLetterType != a_letter) && (StringPositionMarker < StringOverflow))  {
        StringPositionMarker++;
        OneLetterType = lettertype(*StringPositionMarker);     }


        // Read the string into the subsequent argument variables.  String variables are read up through the next space.
    
    for (counter = 2; counter <= ArgsNum; counter++)   {
        BundleBundle = GetBIF_Member(counter);
        if (BundleBundle != NULL)  {
            WriteBundleString(BundleBundle, BundleBundle->bundle_offset, BundleBundle->instances,
                                                    (void *) &StringPositionMarker, StringOverflow);
            if (warning_code != passed)  warning_index--;
            if (err_code != passed)  {  free((void *) HoldString);  return;  }
    }   }
    
    
        // Finally, check to see if there seem to be any left-over useful fields in the string -- if so, set a warning.
    
    OneLetterType = lettertype(*StringPositionMarker);
    while ((OneLetterType != a_symbol) && (OneLetterType != a_number) && (OneLetterType != a_letter) && (StringPositionMarker < StringOverflow))  {
        StringPositionMarker++;
        OneLetterType = lettertype(*StringPositionMarker);         }

    if (StringPositionMarker != HoldString+theString->elementnum)  {  SetWarning(type_mismatch_err, PCCodePtr-1);  free((void *) HoldString);  return;  }

    free((void *) HoldString);      // clean up
    SearchCoordinate._Bundle = NULL;
}


// ext_print() calls PrintOneBundle() to print its entire argument list to the terminal (screen).

void ext_print()  {
    PrintOneBundle(BIF_ArgsCoordinate._Bundle, BIF_ArgsCoordinate._Offset, BIF_ArgsCoordinate._IndexWidth);  SearchCoordinate._Bundle = NULL;
    if (fflush(stdout) != passed)  {  SetError(IO_error, PCCodePtr-1);  }
}

void PrintOneBundle(bundle *theBundle, Ulong theOffset, Ulong theIndexWidth)
{
    void *StringStartPtr;
    linkedlist *theLL = &(theBundle->variable_ptr->mem.data), *StringLL;
    bundle *BundleToStep;
    member *loopMember;
    Ulong counter, byte_counter, offset_counter, index_counter, OffsetToStep, IndexWidthToStep;
    unsigned char theChar;
    unsigned char theType = theBundle->variable_ptr->type;
    
    
        // We use the business flags to avoid re-printing aliases that point back to places we've already been
        // (as this could cause infinite recursion).  If we encounter repeat variables, just set a warning and skip.
        
    if (IfBusy(theBundle, busy_print_flag) != yz_false)  {  SetWarning(self_reference_err, PCCodePtr-1);  return;  }
    SetBusy(theBundle, busy_print_flag);


        // composite types:  we print each member (this can be a mess since they aren't separated by anything)
        
    if (theType == composite_type)  {
        for (counter = 1; counter <= theBundle->variable_ptr->mem.members.elementnum; counter++)  {
            loopMember = (member *) FElement(theLL, counter);
            if ((loopMember->member_ptr != NULL) && (loopMember->if_hidden == yz_false))  {
                for (offset_counter = theOffset; offset_counter < theIndexWidth+theOffset; offset_counter++)   {
                for (index_counter = 0; index_counter < loopMember->indices; index_counter++)   {
                    BundleToStep = theBundle;
                    OffsetToStep = offset_counter;
                    IndexWidthToStep = 1;
                    
                    StepBundle(&BundleToStep, &OffsetToStep, &IndexWidthToStep, loopMember, index_counter, 1);
                    if (err_code != passed)  err_code = passed;    // a void member
                    else  PrintOneBundle(BundleToStep, OffsetToStep, IndexWidthToStep);
                    if (err_code != passed)  return;    // i.e.  self_reference_err
                }}
    }   }   }
    
    
        // block types:  write out each byte as a character, so we'll probably get a bunch of unprintables here
    
    else if (theType == block_type)  {
        for (counter = 1; counter <= theIndexWidth; counter++)  {
            StringStartPtr = FElement(theLL, counter+theOffset);
            for (byte_counter = 0; byte_counter < theLL->elementsize; byte_counter++)  {
                printf("%c", *(((char *) StringStartPtr) + byte_counter));
    }   }   }
    
    
        // strings:  write the printable bytes out verbatim;
        // write the unprintable ones symbolically as they would be entered in a script (\0D for example)
        
    else if (theType == string_type)  {
        for (counter = 1; counter <= theIndexWidth; counter++)  {
            StringLL = (linkedlist *) FElement(theLL, counter+theOffset);
            if (StringLL->memory != NULL)  {
            for (byte_counter = 1; byte_counter <= StringLL->elementnum; byte_counter++)  {
                theChar = *(unsigned char *) FElement(StringLL, byte_counter);
                if (((theChar >= ' ') && (theChar <= '~')) || (theChar == '\t') || (theChar == '\n'))
                    printf("%c", theChar);
                else if (theChar == 13)  {
                    printf("\n");
                    if (byte_counter < StringLL->elementnum)    {
                    if (*(unsigned char *) FElement(StringLL, byte_counter+1) == '\n')    {
                        byte_counter++;
                    }}      }
                else  printf("\\%c%c", ASCII_Hex(theChar/16), ASCII_Hex(theChar % 16));
    }   }   }}
    
    
        // floating point numbers:  write them out as such
        // This generally looks different (have fewer decimal places) from the output of print_string() since it uses C's printf().
    
    else if (theType >= single_type)  {
        for (counter = 1; counter <= theIndexWidth; counter++)  {
            LoadDPRegister(FElement(theLL, counter+theOffset), theType);
            printf("%g", (double) DPRegister);
    }   }
    
    
        // take care of ulongs separately just to make sure we won't break the slong upper limit
    
/*    else if (theType == Ulong_type)  {
        for (counter = 1; counter <= theIndexWidth; counter++)  {
            LoadDPRegister(FElement(theLL, counter+theOffset), theType);
            printf("%u", (unsigned int) DPRegister);
    }   }*/
    
    
        // output any integer using printf()
    
    else  {
        for (counter = 1; counter <= theIndexWidth; counter++)  {
            LoadDPRegister(FElement(theLL, counter+theOffset), theType);
            if ((double) (int) DPRegister != DPRegister)  printf("%g", DPRegister);
            else  printf("%i", (int) DPRegister);
    }   }

    ClearBusy(theBundle, busy_print_flag);    // OK, we can use this bundle again if it appears later
}

// ASCII_Hex():  converts a number 0-15 into a hexidecimal digit 0-F (used in printing strings)
char ASCII_Hex(unsigned char HexNumber)
{
    if (HexNumber <= 9)  return HexNumber+'0';
    
    return HexNumber-10+'A';
}



// ext_clip() removes stems from fibrils, thus breaking a search path at a given point.

void ext_clip()
{
    bundle *BundleToClip;
    fibril *loopFibril, *StemCrawler;
    linkedlist *FibrilDataLL, *FibrilRefLL;
    Ulong counter, bundle_count, top_bundle;
    Boolean ClipAll = yz_false;
    
    top_bundle = MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr);
    
    
        // Clip each argument at the point closest to the tip where that path leads into PCCoordinate._Fibril,
        // which is the active search path of the current code.
        // For example, running "a :: { b :: { .. } }, clip(a.b)" will allow 'b' to explore 'a' but not the current function.
        // CURRENT BUG:  seems to clip at the very top of the path...
        // If any of the arguments is * (the void), this will clip PCCoordinate._Fibril's stem.
    
    for (bundle_count = 1; bundle_count <= top_bundle; bundle_count++)  {
        BundleToClip = GetBIF_Member(bundle_count);
        if (BundleToClip == NULL)  ClipAll = yz_true;
        
        if (ClipAll == yz_false)   {
    
            FibrilDataLL = &(BundleToClip->variable_ptr->fibril_list.data);
            FibrilRefLL = &(BundleToClip->variable_ptr->fibril_list.references);
    
            for (counter = 1; counter <= FibrilDataLL->elementnum; counter++)  {
            if (*LL_ULong(FibrilRefLL, counter) != 0)  {
                loopFibril = (fibril *) FElement(FibrilDataLL, counter);
                StemCrawler = loopFibril;
                while (StemCrawler != NULL)   {
                    if (StemCrawler == PCCoordinate._Fibril)  {
                        DerefFibril(&(loopFibril->stem));
                        break;    }
                    StemCrawler = StemCrawler->stem;   }
        }   }}
        else    {
            if (PCCoordinate._Fibril->stem != NULL)   {
                DerefFibril(&(PCCoordinate._Fibril->stem));   }
        }
    }
    
    SearchCoordinate._Bundle = NULL;
}


// ext_top() returns the number of indices (things that can be rerefenced by [..]) contained in the given variable.
// This is not the same as the number of members!  Each member can have 0-N indices, and/or be hidden in which case it does not contribute.

void ext_top()
{
    Ulong member_counter, top_index = 0;
    bundle *BundleToTop;
    variable *VarToTop;
    member *loopMember;
    
    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 1)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
    

        // get the argument variable
        
    BundleToTop = GetBIF_Member(1);
    if (BundleToTop == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    VarToTop = BundleToTop->variable_ptr;
    if (VarToTop->type != composite_type)  {  SetError(not_composite_err, PCCodePtr-1);  return;  }

    
        // add up the non-hidden indices of each member
    
    for (member_counter = 1; member_counter <= VarToTop->mem.members.elementnum; member_counter++)  {
        loopMember = (member *) FElement(&(VarToTop->mem.members), member_counter);
        if (loopMember->if_hidden == yz_false)  {
            top_index += ((member *) FElement(&(VarToTop->mem.members), member_counter))->indices;
    }   }


        // return R_slong
    
    *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = top_index;
    SetReturnCoordinate(R_slong);
}


// compile() generates bytecode from a script (a string).  Writes errors/warnings into the error/warning registers.

void ext_compile()
{
    linkedlist HobbishCode, *YazooCode, *HobbishString, *NamesLL = NULL, *LinePositionLL;
    bundle *TextBundle, *HobbishBundle, *NamesBundle = NULL, *LinePositionBundle = NULL, *ErrCodeBundle, *ErrPtrBundle;
    Ulong counter, names_num = 0, new_names_characters;
    char *FirstNameLetter = NULL, *TextBeginning, *NamesLL_char, *loop_character, **loop_string_ptr;
    char *YazooTextBlock;
    Boolean use_unique_nameLL = yz_true, get_line_positions = yz_false;
    parse_retrn Compile_retrn;
    LL_retrn linkedLretrn;

    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) == 0)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
    
    
        // read in the script string (argument 1)

    TextBundle = GetBIF_Member(1);
    if (TextBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (TextBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    if (TextBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    
    
        // if the optional namespace string is present, read that in too (arg 2)
    
    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) > 1)  {
        if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) > 3)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
        NamesBundle = GetBIF_Member(2);
        if (NamesBundle != NULL)  {
            if (NamesBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
            if (NamesBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
            use_unique_nameLL = yz_false;           }
        
        
            // if the optional newline-positions argument is specified, read that (arg 3)
            
        if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) == 3)  {
            LinePositionBundle = GetBIF_Member(3);
            if (LinePositionBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
            if (LinePositionBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
            if (LinePositionBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
            get_line_positions = yz_true;
    }   }
    
    
        // here's where we will write our compiled bytecode or error codes/indices
    
    HobbishBundle = BIF_ReturnBundle(R_string);
    ErrCodeBundle = BIF_ReturnBundle(R_error_code);
    ErrPtrBundle = BIF_ReturnBundle(R_error_index);
    
    
        // copy the script into a new buffer, and insert the null terminator that Compile() requires
    
    YazooCode = (linkedlist *) FElement(&(TextBundle->variable_ptr->mem.data), TextBundle->bundle_offset+1);    
    if (YazooCode->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }

    YazooTextBlock = (char *) malloc(YazooCode->elementnum+1);
    if (YazooTextBlock == NULL)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    FGetElements(YazooCode, 1, YazooCode->elementnum, YazooTextBlock);
    *(YazooTextBlock+YazooCode->elementnum) = 0;         // the null terminator
    TextBeginning = YazooTextBlock;


        // if we're using an existent namespace, initialize the compiler with that
    
    if (use_unique_nameLL == yz_false)  {

        NamesLL = (linkedlist *) FElement(&(NamesBundle->variable_ptr->mem.data), NamesBundle->bundle_offset+1);
        if (NamesLL->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  free((void *) YazooTextBlock);  return;  }
        if (NamesLL->elementnum < sizeof(Ulong))  {  SetError(argument_err, PCCodePtr-1);  free((void *) YazooTextBlock);  return;  }


            // check the namespace string and count the number of names in it
        
        names_num = 0;
        if (NamesLL->elementnum > sizeof(Ulong))    {
            if (*LL_UByte(NamesLL, NamesLL->elementnum) != 0)  {  SetError(argument_err, PCCodePtr-1);  free((void *) YazooTextBlock);  return;  }
            linkedLretrn = FDefragmentLinkedList(NamesLL);
            if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  free((void *) YazooTextBlock);  return;  }
    
            FirstNameLetter = (char *) FElement(NamesLL, sizeof(Ulong)+1);
            for (counter = 0; counter < NamesLL->elementnum-sizeof(Ulong); counter++)   {
                if (*(FirstNameLetter+counter) == 0)  names_num++;
        }   }


            // initialize the compiler with the number of names..
        
        Compile_retrn = InitCompile(names_num);
        if (Compile_retrn != passed)  {  SetError(Compile_retrn, PCCodePtr-1);  free((void *) YazooTextBlock);  return;  }
        NullNamesNum = *(Ulong *) FElement(NamesLL, 1);         // it was defragmented
        
        
            // and give the compiler pointers to each name
        
        for (counter = 1; counter <= names_num; counter++)  {
            if (counter != 1)  {  while (*FirstNameLetter != 0)  {  FirstNameLetter++;  }  FirstNameLetter++;  }
            *(void **) FElement(&AllUserNames, counter) = FirstNameLetter;   }   }


        // if no namespace was provided, then just tell the compiler to use its own
    
    else  {
        SetError(InitCompile(0), PCCodePtr-1);
        if (err_code != passed)  {
            free((void *) YazooTextBlock);
            return;
    }   }


        // Run the compiler.  Store the error code/index (which may be zero -- no error).
    
    Compile_retrn = Compile(&TextBeginning, &HobbishCode);
    *(Ulong *) FElement(&(ErrCodeBundle->variable_ptr->mem.data), ErrCodeBundle->bundle_offset+1) = Compile_retrn;
    *(Ulong *) FElement(&(ErrPtrBundle->variable_ptr->mem.data), ErrPtrBundle->bundle_offset+1) = errPtr - YazooTextBlock + 1;
    
    
        // store the bytecode string in R_string

    HobbishString = (linkedlist *) FElement(&(HobbishBundle->variable_ptr->mem.data), 1);
    if (HobbishString->memory != NULL)  FDeleteLinkedList(HobbishString);
    linkedLretrn = FNewLinkedList(HobbishString, HobbishCode.elementnum*sizeof(Ulong), sizeof(char), 0, uncleared);
    if (linkedLretrn != passed)  {
        SetError(out_of_mem_err, PCCodePtr-1);
        free((void *) YazooTextBlock);
        FDeleteLinkedList(&AllUserNames);
        FDeleteLinkedList(&NewlinePositions);
        FDeleteLinkedList(&HobbishCode);
        return;
    }
    if (HobbishCode.elementnum > 0)
        FGetElements(&HobbishCode, 1, HobbishCode.elementnum, FElement(HobbishString, 1));    // assumes 1 sublist for HobbishString
    FDeleteLinkedList(&HobbishCode);


        // store the newline positions in the provided variable, if it was given
    
    if (get_line_positions == yz_true)   {
        LinePositionLL = (linkedlist *) FElement(&(LinePositionBundle->variable_ptr->mem.data), LinePositionBundle->bundle_offset+1);
        if (LinePositionLL->memory == NULL)  linkedLretrn = FNewLinkedList(LinePositionLL, 0, sizeof(char), 0, uncleared);
        else  linkedLretrn = passed;
        
        if (linkedLretrn == passed)  {
            FDeleteElements(LinePositionLL, 1, LinePositionLL->elementnum);
            linkedLretrn = FAddElements(LinePositionLL, NewlinePositions.elementnum*NewlinePositions.elementsize, uncleared);    }
        if (linkedLretrn == passed)  linkedLretrn = FDefragmentLinkedList(LinePositionLL);
        if (linkedLretrn != passed)  {
            SetError(out_of_mem_err, PCCodePtr-1);
            free((void *) YazooTextBlock);
            FDeleteLinkedList(&AllUserNames);
            FDeleteLinkedList(&NewlinePositions);
            return;     }
        if (NewlinePositions.elementnum != 0)
            FGetElements(&NewlinePositions, 1, NewlinePositions.elementnum, FElement(LinePositionLL, 1));   }


        // if we provided a namespace, then update that w/ the new names
    
    if (use_unique_nameLL == yz_false)  {
    
            // get the number of extra bytes occupied by the new names
        
        new_names_characters = 0;
        for (counter = names_num+1; counter <= AllUserNames.elementnum; counter++)  {
            loop_character = *(char **) FElement(&AllUserNames, counter);
            while (*loop_character != 0)  {  new_names_characters++;  loop_character++;  }
            new_names_characters++;   }       // to account for the final 0
        
        
            // make room for new names
        
        linkedLretrn = FAddElements(NamesLL, new_names_characters, uncleared);
        if (linkedLretrn == passed)  linkedLretrn = FDefragmentLinkedList(NamesLL);
        if (linkedLretrn != passed)  {
            SetError(out_of_mem_err, PCCodePtr-1);
            free((void *) YazooTextBlock);
            FDeleteLinkedList(&AllUserNames);
            FDeleteLinkedList(&NewlinePositions);
            return;       }


            // store the new number of unprintable names (tokens)
        
        *(Ulong *) FElement(NamesLL, 1) = NullNamesNum;     // NamesLL was defragmented


            // copy all the new ordinary (non-token) names over

        if (new_names_characters > 0)   {
            NamesLL_char = (char *) FElement(NamesLL, NamesLL->elementnum-new_names_characters+1);    // they are all in a row
            for (counter = names_num+1; counter <= AllUserNames.elementnum; counter++)  {
                loop_string_ptr = (char **) FElement(&AllUserNames, counter);
                loop_character = *loop_string_ptr;

                while (*loop_character != 0)  {  *NamesLL_char = *loop_character;  loop_character++;  NamesLL_char++;  }
                *NamesLL_char = 0;  NamesLL_char++;         // to account for the final 0

//                if (*loop_string_ptr != empty_str)
                    {  free((void *) *loop_string_ptr);  *loop_string_ptr = NULL;  }   // free memory manually instead of calling FinishCompile(), so we
        }   }                                                                          // don't lose the first 1-names_num name pointers, which are ours
    }
    else  FinishCompile();

    FDeleteLinkedList(&AllUserNames);         // regardless of whether we created these here..  Compile() doesn't clean up after itself.
    FDeleteLinkedList(&NewlinePositions);
    
    SetReturnCoordinate(R_error_code);        // return R_error_code
    free((void *) YazooTextBlock);
}


// ext_transform() takes bytecode and stores it in Yazoo's code registry.
// Checks the code first to make sure it won't crash the interpreter.

void ext_transform()
{
    linkedlist *InputStringLL;
    member *loopMember;
    bundle *CodeBundle, *CodeIDBundle = NULL, *CompiledCodeBundle, *FibrilPathBundle, *member_bundle;
    variable *FPB_var;
    fibril *NewCodeFibril, *NewCodeStemFibril = PCCoordinate._Fibril;
    Ulong FPB_member_counter, FirstFPBMember = 0, CodeIndex, counter, ArgsNum;
    Ulong member_offset, step_offset, step_indices, dummy;
    Slong *HoldPCCodePtr, *CodeEntryPtr, *HoldStartCodePtr, *HoldEndCodePtr;
    Boolean return_code_ID = yz_false;
    run_retrn retrn;
    LL_retrn linkedLretrn;
    
    ArgsNum = MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr);
    if ((ArgsNum == 0) || (ArgsNum > 3))  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
    
    
        // get the bytecode string (argument 1)
    
    CodeBundle = GetBIF_Member(1);
    if (CodeBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (CodeBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    if (CodeBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    
    
        // if a code-ID variable was provided, read that too
    
    if (ArgsNum >= 2)  {
        CodeIDBundle = GetBIF_Member(2);
        if (CodeIDBundle != NULL)  {
            if (CodeIDBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
            if (CodeIDBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
            return_code_ID = yz_true;
    }   }
    
    
    // if a code-ID variable was provided, read that too
    
    if (ArgsNum == 3)  {
        FibrilPathBundle = GetBIF_Member(3);
        if (FibrilPathBundle == NULL)  NewCodeStemFibril = NULL;
        else    {
            FPB_var = FibrilPathBundle->variable_ptr;
            
            if (FibrilPathBundle->variable_ptr->type != composite_type)  {  SetError(not_composite_err, PCCodePtr-1);  return;  }
            if (FibrilPathBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
            
            FPB_member_counter = 0;
            while (1 == 1)  {
                FPB_member_counter++;
                retrn = FindMemberIndex(FPB_var, FPB_member_counter, &loopMember, &dummy, &member_offset);
                if (retrn == invalid_index_err)  break;
                member_bundle = FibrilPathBundle;
                if (loopMember->if_hidden == yz_false)  {
                    step_offset = FibrilPathBundle->bundle_offset;
                    step_indices = FibrilPathBundle->instances;
                    StepBundle(&member_bundle, &step_offset, &step_indices, loopMember, member_offset, 1);
                    if ((err_code == void_member_err) || (err_code == void_proxy_err))  NewCodeStemFibril = NULL;
                    else  {  FirstFPBMember = FPB_member_counter;  break;  }
                    err_code = passed;
            }   }
            err_code = passed;
    }   }
    
    
        // check & defragment the bytecode string
    
    InputStringLL = (linkedlist *) FElement(&(CodeBundle->variable_ptr->mem.data), CodeBundle->bundle_offset+1);
    if (InputStringLL->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }
    if (InputStringLL->elementnum == 0)  {
        *LL_ULong(&(BIF_ReturnBundle(R_error_code)->variable_ptr->mem.data), 1) = 0;
        SetReturnCoordinate(R_error_code);
        return;         }
    linkedLretrn = FDefragmentLinkedList(InputStringLL);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    
    CompiledCodeBundle = BIF_ReturnBundle(R_composite);     // get the return variable ready
    
    
        // store some important variables since we need to overwrite these temporarily while checking the bytecode
    
    HoldPCCodePtr = PCCodePtr;
    HoldStartCodePtr = StartCodePtr;
    HoldEndCodePtr = EndCodePtr;
    
    
        // set the PC at the beginning of the bytecode, and run it in 'check' mode
    
    StartCodePtr = LL_SLong(InputStringLL, 1);
    EndCodePtr = StartCodePtr + (Ulong) ceil(1.*InputStringLL->elementnum/sizeof(Ulong));
    PCCodePtr = StartCodePtr;
    CheckHobbish();
    
    
        // restore the important variables
    
    PCCodePtr = HoldPCCodePtr;
    StartCodePtr = HoldStartCodePtr;
    EndCodePtr = HoldEndCodePtr;
    
    
        // store the error code; if it didn't pass then store the error index and return
    
    *LL_ULong(&(BIF_ReturnBundle(R_error_code)->variable_ptr->mem.data), 1) = err_code;
    if (err_code != passed)  {
        *LL_ULong(&(BIF_ReturnBundle(R_error_index)->variable_ptr->mem.data), 1) = err_index;
        err_code = passed;
        
        SetReturnCoordinate(R_error_code);
        return;         }
    
    
        // OK, it worked.  Make a new anchor for the code.
    
    if (NewCodeStemFibril != NULL)  RefFibril(NewCodeStemFibril);       // temporarily over-reference each new fibril as we create it
    if (FirstFPBMember != 0)    {

        FPB_member_counter = FirstFPBMember-1;
        while (1 == 1)  {
            FPB_member_counter++;
            retrn = FindMemberIndex(FPB_var, FPB_member_counter, &loopMember, &dummy, &member_offset);
            if (retrn == invalid_index_err)  break;
            member_bundle = FibrilPathBundle;
            if (loopMember->if_hidden == yz_false)  {
                step_offset = FibrilPathBundle->bundle_offset;
                step_indices = FibrilPathBundle->instances;
                StepBundle(&member_bundle, &step_offset, &step_indices, loopMember, member_offset, 1);
                if (err_code == passed)     {
                    retrn = DrawFibril(member_bundle, NewCodeStemFibril, &NewCodeFibril, member_bundle->bundle_offset,
                                       member_bundle->instances, yz_true, PCCodeRef.PLL_index, 0, NULL);
                    if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
                    RefFibril(NewCodeFibril);                               // so it doesn't get deleted by a later fibril in the same bundle
                    if (NewCodeStemFibril != NULL)  DerefFibril(&NewCodeStemFibril);
                    NewCodeStemFibril = NewCodeFibril;
                }
                else  err_code = passed;
        }   }
        err_code = passed;
        
        /*for (FPB_member_counter = FirstFPBMember; FPB_member_counter <= FPB_var->mem.members.elementnum; FPB_member_counter++)  {
            loopMember = (member *) FElement(&(FPB_var->mem.members), FPB_member_counter);
            if ((loopMember->member_ptr != NULL) && (loopMember->if_hidden == yz_false))  {
                retrn = DrawFibril(loopMember->member_ptr, NewCodeStemFibril, &NewCodeFibril, loopMember->member_ptr->bundle_offset,
                                   PCCoordinate._IndexWidth*loopMember->indices, yz_true, PCCodeRef.PLL_index, 0, NULL);
                if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
                RefFibril(NewCodeFibril);                               // so it doesn't get deleted by a later fibril in the same bundle
                if (NewCodeStemFibril != NULL)  DerefFibril(&NewCodeStemFibril);
                NewCodeStemFibril = NewCodeFibril;*/
    }   //}   }
    retrn = DrawFibril(CompiledCodeBundle, NewCodeStemFibril, &NewCodeFibril, 0, 1, yz_true, PCCodeRef.PLL_index, 0, NULL);
    if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
    RefFibril(NewCodeFibril);
    if (NewCodeStemFibril != NULL)  DerefFibril(&NewCodeStemFibril);
    
    
        // wipe up any existing codes in R_composite
    
    for (counter = 1; counter <= CompiledCodeBundle->variable_ptr->code_list.elementnum; counter++)
        Deref_CodeRef((code_ref *) FElement(&(CompiledCodeBundle->variable_ptr->code_list), counter));
    FDeleteElements(&(CompiledCodeBundle->variable_ptr->code_list), 1, CompiledCodeBundle->variable_ptr->code_list.elementnum);
    
    
        // add our new code to the Registry..
        
    retrn = AddCode(LL_SLong(InputStringLL, 1), &CodeEntryPtr, &CodeIndex, (Ulong) ceil(1.*InputStringLL->elementnum/sizeof(Ulong)));
    if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
    
    
        // and to the code list of R_composite
    
    retrn = AddCodeRef(&(CompiledCodeBundle->variable_ptr->code_list), NewCodeFibril, CodeEntryPtr, CodeIndex);
    if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
    DerefFibril(&NewCodeFibril);
    
    
        // if we were asked for the code ID back, give them that
    
    if (return_code_ID == yz_true)  {
        DPRegister = (double) CodeIndex;
        SaveDPRegister(FElement(&(CodeIDBundle->variable_ptr->mem.data), CodeIDBundle->bundle_offset+1), CodeIDBundle->variable_ptr->type);
        if (err_code == out_of_range_err)  {  err_code = passed;  SetWarning(out_of_range_err, PCCodePtr-1);  }
    }
    
    SetReturnCoordinate(R_error_code);     // return R_error_code
}


// ext_trap() catches runtime errors in its constructor, and allows execution to resume after the trap.
// Uses the registers to return error/warning codes/indices/offsets.

void ext_trap()
{
    code_ref ArgCodeRef;
    Ulong counter;
    Boolean IfSetWarning = yz_false, IfSetError = yz_false;
    run_retrn SCR_retrn;


        // run each code in our args constructor
        
    for (counter = 1; counter <= BIF_ArgsCoordinate._Bundle->variable_ptr->code_list.elementnum; counter++)  {

            // get a handle on the code
        
        ArgCodeRef = *(code_ref *) FElement(&(BIF_ArgsCoordinate._Bundle->variable_ptr->code_list), counter);
        ArgCodeRef.anchor = PCCoordinate._Fibril;
        Ref_CodeRef(&ArgCodeRef);   // in case BE() changes something
        
        
            // run the code
        warning_code = passed;
        BeginExecution(&ArgCodeRef, yz_true, BIF_ArgsCoordinate._Offset, BIF_ArgsCoordinate._IndexWidth, 0);
        Deref_CodeRef(&ArgCodeRef);
        if (err_code == return_flag)  err_code = passed;


            // save warnings in the warning registers
        
        if (warning_code != passed)   {
            *LL_ULong(&(BIF_ReturnBundle(R_warning_code)->variable_ptr->mem.data), 1) = warning_code;
            *LL_ULong(&(BIF_ReturnBundle(R_warning_index)->variable_ptr->mem.data), 1) = warning_index;
            warning_code = passed;
            SCR_retrn = SetCodeRef(BIF_ReturnBundle(R_warning_script)->variable_ptr, &warning_script);
            if (SCR_retrn != passed)  {  SetError(SCR_retrn, PCCodePtr-1);  return;  }
            IfSetWarning = yz_true;   }


            // save errors in the error registers
        
        if (err_code != passed)  {
            *LL_ULong(&(BIF_ReturnBundle(R_error_code)->variable_ptr->mem.data), 1) = err_code;
            *LL_ULong(&(BIF_ReturnBundle(R_error_index)->variable_ptr->mem.data), 1) = err_index;
            err_code = passed;
            SCR_retrn = SetCodeRef(BIF_ReturnBundle(R_error_script)->variable_ptr, &err_script);
            if (SCR_retrn != passed)  {  SetError(SCR_retrn, PCCodePtr-1);  return;  }
            IfSetError = yz_true;
            break;       }          }
    
    
        // clear R_error_code/index if no error, and R_warning_code/index if no warning
    
    if (IfSetError == yz_false)  {
        *LL_ULong(&(BIF_ReturnBundle(R_error_code)->variable_ptr->mem.data), 1) = 0;
        *LL_ULong(&(BIF_ReturnBundle(R_error_index)->variable_ptr->mem.data), 1) = 0;    }
        
    if (IfSetWarning == yz_false)   {
        *LL_ULong(&(BIF_ReturnBundle(R_warning_code)->variable_ptr->mem.data), 1) = 0;
        *LL_ULong(&(BIF_ReturnBundle(R_warning_index)->variable_ptr->mem.data), 1) = 0;   }

    SetReturnCoordinate(R_error_code);      // return R_error_code
}


// Used by trap():  SetCodeRef() sets a variable's code list to hold element: theCR.  (Used for storing error/warning_script).

run_retrn SetCodeRef(variable *CR_Variable, code_ref *theCR)
{
    Ulong counter;
    LL_retrn linkedLretrn;
    
    
        // get rid of all existing code refs in that register
    
    for (counter = 1; counter <= CR_Variable->code_list.elementnum; counter++)
        Deref_CodeRef((code_ref *) FElement(&(CR_Variable->code_list), counter));
        
    FDeleteElements(&(CR_Variable->code_list), 1, CR_Variable->code_list.elementnum);
    
    
        // add and reference our code refs of interest
    
    linkedLretrn = AddElements(&(CR_Variable->code_list), 1, uncleared);
    if (linkedLretrn != passed)  return out_of_mem_err;
    
    *(code_ref *) FElement(&(CR_Variable->code_list), 1) = *theCR;
    Ref_CodeRef((code_ref *) FElement(&(CR_Variable->code_list), 1));
    
    return passed;
}


// throw() generates an error.  throw(err_code[, err_index])

void ext_throw()
{
    bundle *ParamBundles[4] = { NULL, NULL, NULL, NULL };
    code_ref *EW_script = &err_script, *TargetCodeRef = &PCCodeRef;
    Ulong ArgsNum, cA, *EW_code = &err_code, *EW_index = &err_index;
    Ulong nums[4] = { 0, 0, PCCodePtr-1 - PCCodeRef.code_ptr + 1, yz_false };

    ArgsNum = MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr);
    
    if ((ArgsNum < 1) || (ArgsNum > 4))  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }

    
        // read in the error code, throwing real errors if something's wrong with the argument
    
    for (cA = 0; cA < ArgsNum; cA++)    {
        ParamBundles[cA] = GetBIF_Member(cA+1);
        if (ParamBundles[cA] != NULL)  {
            if (cA != 1)    {
                if (ParamBundles[cA]->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
                if (ParamBundles[cA]->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
                
                LoadDPRegister(FElement(&(ParamBundles[cA]->variable_ptr->mem.data), ParamBundles[cA]->bundle_offset+1),
                                            ParamBundles[cA]->variable_ptr->type);
                if (!((DPRegister >= 0) && (DPRegister <= ULONG_MAX)))
                    {  SetError(out_of_range_err, PCCodePtr-1);  return;  }
                if ((cA == 3) && (DPRegister != 0.) && (DPRegister != 1.))
                    {  SetError(out_of_range_err, PCCodePtr-1);  return;  }
                nums[cA] = (Ulong) DPRegister;          }
            else    {
                if (ParamBundles[cA]->variable_ptr->code_list.elementnum != 1)  {  SetError(argument_err, PCCodePtr-1);  return;  }
                TargetCodeRef = (code_ref *) FElement(&(ParamBundles[cA]->variable_ptr->code_list), 1);
            }
    }   }
    
    
    
        // throw the error/warning
    
    if (nums[3] == 1)  {
        EW_code = &warning_code;
        EW_script = &warning_script;
        EW_index = &warning_index;      }
    
    *EW_code = nums[0];
    
    if (EW_script->code_ptr != NULL)  Deref_CodeRef(EW_script);
    *EW_script = *TargetCodeRef;
    EW_script->anchor = NULL;
    Ref_CodeRef(EW_script);
    
    *EW_index = nums[2];
        
    SearchCoordinate._Bundle = NULL;         // probably redundant
}


// ext_size() returns the total number of bytes contained in its argument.  bytes_num = size(bundle_to_size)

void ext_size()
{
    bundle *BundleToSize;
    Ulong DataSize = 0, StringSize = 0;


        // read in the argument
    
    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 1)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
    BundleToSize = GetBIF_Member(1);
    if (BundleToSize != NULL)  {               // compute its size
        ExtCallMode = no_call;
        SizeBundle(BundleToSize, BundleToSize->bundle_offset, BundleToSize->instances, &DataSize, &StringSize);     }
    
    *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = DataSize+StringSize;
    SetReturnCoordinate(R_slong);      // return R_slong
}


// The following mathematical functions all invoke ExtMathTemplate(),
// and each is paired with a second function that does the actual computation.
// All math is done in double-precision floating point.

void ext_abs()  {  ExtMathTemplate(&Math_Abs);  }
double Math_Abs(double Argument)  {  return fabs(Argument);  }

void ext_round_down()  {  ExtMathTemplate(&Math_Round_Down);  }
double Math_Round_Down(double Argument)  {  return floor(Argument);  }

void ext_round_up()  {  ExtMathTemplate(&Math_Round_Up);  }
double Math_Round_Up(double Argument)  {  return ceil(Argument);  }

void ext_log()  {  ExtMathTemplate(&Math_Log);  }
double Math_Log(double Argument)  {  return log(Argument);  }

void ext_cos()  {  ExtMathTemplate(&Math_Cos);  }
double Math_Cos(double Argument)  {  return cos(Argument);  }

void ext_sin()  {  ExtMathTemplate(&Math_Sin);  }
double Math_Sin(double Argument)  {  return sin(Argument);  }

void ext_tan()  {  ExtMathTemplate(&Math_Tan);  }
double Math_Tan(double Argument)  {  return tan(Argument);  }

void ext_acos()  {  ExtMathTemplate(&Math_ArcCos);  }
double Math_ArcCos(double Argument)  {  return acos(Argument);  }

void ext_asin()  {  ExtMathTemplate(&Math_ArcSin);  }
double Math_ArcSin(double Argument)  {  return asin(Argument);  }

void ext_atan()  {  ExtMathTemplate(&Math_ArcTan);  }
double Math_ArcTan(double Argument)  {  return atan(Argument);  }

void ExtMathTemplate(double(MathFunction)(double Argument))
{
    bundle *ArgumentBundle;

    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 1)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }


        // get the first numeric argument, and store it in DPRegister
    
    ArgumentBundle = GetBIF_Member(1);
    if (ArgumentBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }

    LoadDPRegister(FElement(&(ArgumentBundle->variable_ptr->mem.data), ArgumentBundle->bundle_offset+1), ArgumentBundle->variable_ptr->type);


        // run the computation and return the result in R_double
    
    *LL_Double(&(BIF_ReturnBundle(R_double)->variable_ptr->mem.data), 1) = MathFunction(DPRegister);
    SetReturnCoordinate(R_double);
}


// ext_random() generates a random number and returns it in R_double.

void ext_random()
{
    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 0)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }

    *LL_Double(&(BIF_ReturnBundle(R_double)->variable_ptr->mem.data), 1) = ((double) rand())/RAND_MAX + ((double) rand())/RAND_MAX/RAND_MAX;
    SetReturnCoordinate(R_double);
}



// ext_extract() returns a portion of a string in R_string.

void ext_extract()
{
    bundle *ArgumentBundle;
    linkedlist *SourceString, *OutputString, OutputTemp;
    Ulong FirstToExtract, LastToExtract;
    LL_retrn linkedLretrn;
    
    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 3)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }


        // load first argument -- the source string
    
    ArgumentBundle = GetBIF_Member(1);
    if (ArgumentBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    SourceString = (linkedlist *) FElement(&(ArgumentBundle->variable_ptr->mem.data), ArgumentBundle->bundle_offset+1);
    if (SourceString->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }


        // load second argument -- beginning character of substring

    ArgumentBundle = GetBIF_Member(2);
    if (ArgumentBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    LoadDPRegister(FElement(&(ArgumentBundle->variable_ptr->mem.data), ArgumentBundle->bundle_offset+1), ArgumentBundle->variable_ptr->type);
    if (!((DPRegister >= 1) && (DPRegister <= SourceString->elementnum+1)))  {  SetError(argument_out_of_range_err, PCCodePtr-1);  return;  }
    FirstToExtract = (Ulong) DPRegister;


        // load third argument -- ending character of substring

    ArgumentBundle = GetBIF_Member(3);
    if (ArgumentBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    LoadDPRegister(FElement(&(ArgumentBundle->variable_ptr->mem.data), ArgumentBundle->bundle_offset+1), ArgumentBundle->variable_ptr->type);
    if (!((DPRegister >= 0) && (DPRegister <= SourceString->elementnum)))  {  SetError(argument_out_of_range_err, PCCodePtr-1);  return;  }
    LastToExtract = (Ulong) DPRegister;

    if (FirstToExtract > LastToExtract)  LastToExtract = FirstToExtract-1;     // common sense


        // copy the substring into R_string's linked list, and return R_string
    
    OutputString = (linkedlist *) FElement(&(BIF_ReturnBundle(R_string)->variable_ptr->mem.data), 1);
    linkedLretrn = NewLinkedList(&OutputTemp, LastToExtract-FirstToExtract+1, sizeof(char), 0, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    FCopyElements(SourceString, FirstToExtract, &OutputTemp, 1, LastToExtract-FirstToExtract+1);
    if (OutputString->memory != NULL)  FDeleteLinkedList(OutputString);
    *OutputString = OutputTemp;
    
    SetReturnCoordinate(R_string);
}


// ext_find() searches a string for a substring, and returns its first/last appearance,
// or the number of times it appears (without overlapping).

void ext_find()
{
    bundle *ArgumentBundle;
    linkedlist *StringToSearch, *SoughtString;
    char *StrPtrToSearch, *SoughtStrPtr;
    Ulong ArgsNum, StartingPosition, MatchedCharacters, NumberOfMatches, CharacterCounter;
    signed long *ReturnValue;
    signed char Mode;
    LL_retrn linkedLretrn;
    
    ArgsNum = MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr);
    if ((ArgsNum < 2) || (ArgsNum > 4))
        {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }


        // load argument 1 -- the string to search
    
    ArgumentBundle = GetBIF_Member(1);
    if (ArgumentBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    StringToSearch = (linkedlist *) FElement(&(ArgumentBundle->variable_ptr->mem.data), ArgumentBundle->bundle_offset+1);
    if (StringToSearch->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }


        // load argument 2 -- the string to look for inside the big string

    ArgumentBundle = GetBIF_Member(2);
    if (ArgumentBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->variable_ptr->type != string_type)  {  SetError(string_expected_err, PCCodePtr-1);  return;  }
    if (ArgumentBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    SoughtString = (linkedlist *) FElement(&(ArgumentBundle->variable_ptr->mem.data), ArgumentBundle->bundle_offset+1);
    if (SoughtString->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }


        // if argument 3 is present, load it -- it is the 'mode' (search forwards from beginning, backwards from end, or count)

    if (ArgsNum < 3)  {  Mode = 1;  StartingPosition = 1;  }
    else  {
        ArgumentBundle = GetBIF_Member(3);
        if (ArgumentBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
        if (ArgumentBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
        if (ArgumentBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
        LoadDPRegister(FElement(&(ArgumentBundle->variable_ptr->mem.data), ArgumentBundle->bundle_offset+1), ArgumentBundle->variable_ptr->type);
        if (!((DPRegister >= -1) && (DPRegister <= 1)))  {  SetError(argument_err, PCCodePtr-1);  return;  }
        Mode = (signed char) DPRegister;
    

            // if argument 4 is present, load it -- that is the point in the big string to start looking/counting (defaults to beginning/end)

        if (ArgsNum < 4)  {
            if (Mode == -1)  StartingPosition = StringToSearch->elementnum;
            else  StartingPosition = 1;         }
        else  {
            ArgumentBundle = GetBIF_Member(4);
            if (ArgumentBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
            if (ArgumentBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
            if (ArgumentBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
            LoadDPRegister(FElement(&(ArgumentBundle->variable_ptr->mem.data), ArgumentBundle->bundle_offset+1), ArgumentBundle->variable_ptr->type);
            if (!((DPRegister >= 1) && (DPRegister <= StringToSearch->elementnum)))  {  SetError(argument_out_of_range_err, PCCodePtr-1);  return;  }
            StartingPosition = (Ulong) DPRegister;
    }   }


        // prepare the return variable (R_slong)
    
    ReturnValue = LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1);
    linkedLretrn = DefragmentLinkedList(StringToSearch);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    linkedLretrn = DefragmentLinkedList(SoughtString);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    SetReturnCoordinate(R_slong);

    
        // get pointers to our two strings
    
    if (StringToSearch->elementnum == 0)  {  *ReturnValue = 0;  return;  }
    if (SoughtString->elementnum == 0)  {
        if (Mode == 0)  *ReturnValue = 0;
        else  *ReturnValue = (signed long) StartingPosition;
        return;     }
    SoughtStrPtr = (char *) FElement(SoughtString, 1);
    StrPtrToSearch = (char *) FElement(StringToSearch, 1);
    
    MatchedCharacters = 0;
    NumberOfMatches = 0;
    
    if (Mode > -1)  {    // search forwards
        for (CharacterCounter = StartingPosition-1; CharacterCounter < StringToSearch->elementnum; CharacterCounter++)  {
            if (*(StrPtrToSearch+CharacterCounter) == *(SoughtStrPtr+MatchedCharacters))    {
                MatchedCharacters++;
                if (MatchedCharacters == SoughtString->elementnum)  {

                    if (Mode == 1)   {      // we're finding the first match -- return
                        *ReturnValue = (signed long) CharacterCounter+2-MatchedCharacters;
                        return;   }

                    else   {     // we're counting -- increment the counter and keep going
                        MatchedCharacters = 0;
                        NumberOfMatches++;   }
            }   }
            else  MatchedCharacters = 0;
    }   }

    else    {     // search backwards
        for (CharacterCounter = StartingPosition; CharacterCounter > 0; CharacterCounter--)   {
            if (*(StrPtrToSearch+CharacterCounter-1) == *(SoughtStrPtr+SoughtString->elementnum-MatchedCharacters-1))    {
                MatchedCharacters++;

                if (MatchedCharacters == SoughtString->elementnum)   {
                    *ReturnValue = (signed long) CharacterCounter;      // we found the match -- return
                    return;
            }   }
            else  MatchedCharacters = 0;
    }   }
    
    if (Mode == 0)  *ReturnValue = (signed long) NumberOfMatches;
    else  *ReturnValue = 0;     // if we were searching, rather than counting, then if we made it this far we found nothing
}



// The next three funtions use GetYazooVariableInfo() to extract information about a given variable.

void ext_YZ_variable_type()  {  GetYazooVariableInfo(_ext_YZ_variable_type);  }
void ext_YZ_variable_member_top()  {  GetYazooVariableInfo(_ext_YZ_variable_member_top);  }
void ext_YZ_variable_code_top()  {  GetYazooVariableInfo(_ext_YZ_variable_code_top);  }

void GetYazooVariableInfo(Ulong FunctionID)
{
    bundle *HostBundle;


        // get the variable (arg 1)
    
    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 1)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
    HostBundle = GetBIF_Member(1);
    if (HostBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    
    
        // return the requested piece of information in R_slong
    
    if (FunctionID == _ext_YZ_variable_type)   {
        *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = HostBundle->variable_ptr->type;   }

    else if (FunctionID == _ext_YZ_variable_member_top)   {
        if (HostBundle->variable_ptr->type == composite_type)
            *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = HostBundle->variable_ptr->mem.members.elementnum;
        else
            *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = 0;    }

    else if (FunctionID == _ext_YZ_variable_code_top)   {
        if (HostBundle->variable_ptr->type == composite_type)
            *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = HostBundle->variable_ptr->code_list.elementnum;
        else
            *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = 0;     }

    SetReturnCoordinate(R_slong);      // return R_slong
}


// The next six funtions use GetYazooMemberInfo() to extract information about a given member.

void ext_YZ_member_type()  {  GetYazooMemberInfo(_ext_YZ_member_type);  }
void ext_YZ_member_code_top()  {  GetYazooMemberInfo(_ext_YZ_member_code_top);  }
void ext_YZ_member_indices()  {  GetYazooMemberInfo(_ext_YZ_member_indices);  }
void ext_YZ_member_ID()  {  GetYazooMemberInfo(_ext_YZ_member_ID);  }
void ext_YZ_if_member_targeted()  {  GetYazooMemberInfo(_ext_YZ_if_member_targeted);  }
void ext_YZ_if_hidden_member()  {  GetYazooMemberInfo(_ext_YZ_if_hidden_member);  }

void GetYazooMemberInfo(Ulong FunctionID)
{
    bundle *HostBundle, *MemberIndexBundle;
    Ulong MemberNumber;
    member *SoughtMember;

    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 2)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }


        // get the variable containing the member (arg 1)
    
    HostBundle = GetBIF_Member(1);
    if (HostBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (HostBundle->variable_ptr->type != composite_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }


        // get the member number (not the index!) -- arg 2
    
    MemberIndexBundle = GetBIF_Member(2);
    if (MemberIndexBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (MemberIndexBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
    if (MemberIndexBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }

    LoadDPRegister(FElement(&(MemberIndexBundle->variable_ptr->mem.data), MemberIndexBundle->bundle_offset+1), MemberIndexBundle->variable_ptr->type);
    if (!((DPRegister >= 1) && (DPRegister <= HostBundle->variable_ptr->mem.members.elementnum)))
            {  SetError(argument_out_of_range_err, PCCodePtr-1);  return;  }
    MemberNumber = (Ulong) DPRegister;
    SoughtMember = (member *) FElement(&(HostBundle->variable_ptr->mem.members), MemberNumber);


        // return the requested piece of info in R_slong
    
    if (FunctionID == _ext_YZ_member_type)  {
        *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = SoughtMember->type;         }

    else if (FunctionID == _ext_YZ_member_code_top)  {
        if (SoughtMember->type == composite_type)
            *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = SoughtMember->code_list.elementnum;
        else
            *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = 0;         }

    else if (FunctionID == _ext_YZ_member_indices)  {
        *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = SoughtMember->indices;         }

    else if (FunctionID == _ext_YZ_member_ID)  {
        *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = SoughtMember->member_ID;       }

    else if (FunctionID == _ext_YZ_if_member_targeted)  {
        if (SoughtMember->member_ptr == NULL)  {
            *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = yz_false;         }
        else  {
            *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = yz_true;         }    }

    else if (FunctionID == _ext_YZ_if_hidden_member)  {
        *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = SoughtMember->if_hidden;         }

    SetReturnCoordinate(R_slong);
}


// ext_YZ_ParentScriptInfo() returns the script ID and offset of a given code of a given function.

void ext_YZ_ParentScriptInfo()
{
    bundle *HostBundle, *ScriptIndexBundle, *ParentScriptIDBundle, *ScriptOffsetBundle = NULL;
    Ulong ScriptIndex, ArgsNum, ParentScriptID;

    ArgsNum = MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr);
    if ((ArgsNum != 3) && (ArgsNum != 4))  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }


        // load the function variable (arg 1)
    
    HostBundle = GetBIF_Member(1);
    if (HostBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (HostBundle->variable_ptr->type != composite_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }


        // load the ID variable (arg 2)
    
    ScriptIndexBundle = GetBIF_Member(2);
    if (ScriptIndexBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (ScriptIndexBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
    if (ScriptIndexBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }

    
        // load the script ID variable (arg 3)
    
    ParentScriptIDBundle = GetBIF_Member(3);
    if (ParentScriptIDBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (ParentScriptIDBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
    if (ParentScriptIDBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }


        // load the offset variable (arg 4) if that was provided
    
    if (ArgsNum == 4)  {
        ScriptOffsetBundle = GetBIF_Member(4);
        if (ScriptOffsetBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
        if (ScriptOffsetBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
        if (ScriptOffsetBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    }

    
        // find our code
    
    LoadDPRegister(FElement(&(ScriptIndexBundle->variable_ptr->mem.data), ScriptIndexBundle->bundle_offset+1), ScriptIndexBundle->variable_ptr->type);
    if (!((DPRegister >= 1) && (DPRegister <= HostBundle->variable_ptr->code_list.elementnum)))
            {  SetError(argument_out_of_range_err, PCCodePtr-1);  return;  }
    ScriptIndex = (Ulong) DPRegister;
    
    
        // save the parent script ID
    
    ParentScriptID = ((code_ref *) FElement(&(HostBundle->variable_ptr->code_list), ScriptIndex))->PLL_index;
    DPRegister = (double) ParentScriptID;
    SaveDPRegister(FElement(&(ParentScriptIDBundle->variable_ptr->mem.data), ParentScriptIDBundle->bundle_offset+1),
                                    ParentScriptIDBundle->variable_ptr->type);
    if (err_code == out_of_range_err)  {  err_code = passed;  SetWarning(out_of_range_err, PCCodePtr-1);  }


        // optionally, save the parent script offset
    
    if (ArgsNum == 4)   {
        DPRegister = (double) (((code_ref *) FElement(&(HostBundle->variable_ptr->code_list), ScriptIndex))->code_ptr
                                        - *(Slong **) FElement(&(MasterCodeList.data), ParentScriptID));
        SaveDPRegister(FElement(&(ScriptOffsetBundle->variable_ptr->mem.data), ScriptOffsetBundle->bundle_offset+1),
                                        ScriptOffsetBundle->variable_ptr->type);
        if (err_code == out_of_range_err)  {  err_code = passed;  SetWarning(out_of_range_err, PCCodePtr-1);  }     }
    
    SearchCoordinate._Bundle = NULL;     // we've stored all our info in the argument variable(s); don't need to return anything
}



// These next six routines use GetYazooCode() to extract information about member or variable codes.

void ext_YZ_get_variable_code()  {  GetYazooCode(_ext_YZ_get_variable_code, 0);  }
void ext_YZ_get_member_code()  {  GetYazooCode(_ext_YZ_get_member_code, 0);  }
void ext_YZ_get_variable_code_ID()  {  GetYazooCode(_ext_YZ_get_variable_code, 1);  }
void ext_YZ_get_member_code_ID()  {  GetYazooCode(_ext_YZ_get_member_code, 1);  }
void ext_YZ_get_variable_code_offset()  {  GetYazooCode(_ext_YZ_get_variable_code, 2);  }
void ext_YZ_get_member_code_offset()  {  GetYazooCode(_ext_YZ_get_member_code, 2);  }

void GetYazooCode(Ulong FunctionID, unsigned char Query)
{
    bundle *HostBundle, *MemberIndexBundle, *CodeIndexBundle;
    Ulong MemberNumber, CodeArgumentNumber = 0;
    linkedlist *ListOfCodeRefs = NULL, *HobbishString;
    code_ref *SoughtCodeRef;
    Slong *HoldPC;
    LL_retrn linkedLretrn;
    
    if ((FunctionID == _ext_YZ_get_variable_code) || (FunctionID == _ext_YZ_get_variable_code_ID))  CodeArgumentNumber = 2;
    else if ((FunctionID == _ext_YZ_get_member_code) || (FunctionID == _ext_YZ_get_member_code_ID))  CodeArgumentNumber = 3;
    
    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != CodeArgumentNumber)
        {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }


        // load the variable containing the member or variable code (arg 1)
    
    HostBundle = GetBIF_Member(1);
    if (HostBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (HostBundle->variable_ptr->type != composite_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }


        // if it's something about a member we want to know, then we have to provide the member number (arg 2)
    
    if (CodeArgumentNumber == 3)  {
        MemberIndexBundle = GetBIF_Member(2);
        if (MemberIndexBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
        if (MemberIndexBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
        if (MemberIndexBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    
        LoadDPRegister(FElement(&(MemberIndexBundle->variable_ptr->mem.data), MemberIndexBundle->bundle_offset+1), 
                                            MemberIndexBundle->variable_ptr->type);
        if (!((DPRegister >= 1) && (DPRegister <= HostBundle->variable_ptr->mem.members.elementnum))) 
                {  SetError(argument_out_of_range_err, PCCodePtr-1);  return;  }
        MemberNumber = (Ulong) DPRegister;
        ListOfCodeRefs = (linkedlist *) &(((member *) FElement(&(HostBundle->variable_ptr->mem.members), MemberNumber))->code_list);    }
    
    else if (FunctionID == _ext_YZ_get_variable_code)
        ListOfCodeRefs = (linkedlist *) &(HostBundle->variable_ptr->code_list);
    
    
        // get the code index (last argument)
    
    CodeIndexBundle = GetBIF_Member(CodeArgumentNumber);
    if (CodeIndexBundle == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
    if (CodeIndexBundle->variable_ptr->type > double_type)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
    if (CodeIndexBundle->instances != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }

    LoadDPRegister(FElement(&(CodeIndexBundle->variable_ptr->mem.data), CodeIndexBundle->bundle_offset+1), CodeIndexBundle->variable_ptr->type);
    if (!((DPRegister >= 1) && (DPRegister <= ListOfCodeRefs->elementnum)))  {  SetError(argument_out_of_range_err, PCCodePtr-1);  return;  }
    SoughtCodeRef = (code_ref *) FElement(ListOfCodeRefs, (Ulong) DPRegister);


        // if we just want the bytecode, copy that into R_string
    
    if (Query == 0)   {
        HoldPC = PCCodePtr;
        PCCodePtr = SoughtCodeRef->code_ptr;
        RunSkipMode(0);    // should always work since the code has already been checked; just in case, check err_code (index would be off, though)
        if (err_code != passed)  {  PCCodePtr = HoldPC;  return;  }
        PCCodePtr++;
        
        HobbishString = (linkedlist *) FElement(&(BIF_ReturnBundle(R_string)->variable_ptr->mem.data), 1);
        if (HobbishString->memory != NULL)  FDeleteLinkedList(HobbishString);
        linkedLretrn = FNewLinkedList(HobbishString, (PCCodePtr-SoughtCodeRef->code_ptr)*sizeof(Ulong), sizeof(char), 0, uncleared);
        PCCodePtr = HoldPC;
        if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
        
        FSetElements(HobbishString, 1, HobbishString->elementnum, (void *) SoughtCodeRef->code_ptr);     // HobbishString includes the final 0
        SetReturnCoordinate(R_string);   }
    
    
        // if we want the code ID, stick that into R_slong
    
    else if (Query == 1)   {
        *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = SoughtCodeRef->PLL_index;
        SetReturnCoordinate(R_slong);    }
    
    
        // if we want the code offset, put that into R_slong
    
    else     {
        *LL_SLong(&(BIF_ReturnBundle(R_slong)->variable_ptr->mem.data), 1) = 
                    SoughtCodeRef->code_ptr - (*(Slong **) FElement(&(MasterCodeList.data), SoughtCodeRef->PLL_index));
        SetReturnCoordinate(R_slong);    }
}


// SpringCleaning() removes inaccessible memory.

void SpringCleaning()
{
    if (MembersNum(BIF_ArgsCoordinate._Bundle->variable_ptr) != 0)  {  SetError(wrong_argument_count_err, PCCodePtr-1);  return;  }
    CombBundles();
    SearchCoordinate._Bundle = NULL;
}



// The register commands follow.

void ext_slong()  {  SetReturnCoordinate(R_slong);  }
void ext_double()  {  SetReturnCoordinate(R_double);  }
void ext_string()  {  SetReturnCoordinate(R_string);  }
void ext_composite()  {  SetReturnCoordinate(R_composite);  }
void ext_error_code()  {  SetReturnCoordinate(R_error_code);  }
void ext_error_index()  {  SetReturnCoordinate(R_error_index);  }
void ext_error_script()  {  SetReturnCoordinate(R_error_script);  }
void ext_warning_code()  {  SetReturnCoordinate(R_warning_code);  }
void ext_warning_index()  {  SetReturnCoordinate(R_warning_index);  }
void ext_warning_script()  {  SetReturnCoordinate(R_warning_script);  }



// These last three are generic support routines.


// MembersNum() returns the number of non-hidden members of theVariable.

Ulong MembersNum(variable *theVariable)
{
    Ulong counter, NumberOfMembers = 0;

    for (counter = 1; counter <= theVariable->mem.members.elementnum; counter++)   {
        if (((member *) FElement(&(theVariable->mem.members), counter))->if_hidden == yz_false)   {
            NumberOfMembers++;
    }   }
    
    return NumberOfMembers;
}


// GetBIF_Member() returns the Nth non-hidden member of BIF_ArgsCoordinate (N being denoted by 'SoughtMemberIndex').
// BIF = built-in-function.

bundle *GetBIF_Member(Ulong SoughtMemberIndex)
{
    Ulong counter, MemberCounter = 0;
    member *loopMember;
    
    for (counter = 1; counter <= BIF_ArgsCoordinate._Bundle->variable_ptr->mem.members.elementnum; counter++)   {
        loopMember = (member *) FElement(&(BIF_ArgsCoordinate._Bundle->variable_ptr->mem.members), counter);
        if (loopMember->member_ptr == NULL)   {
            MemberCounter++;
            if (MemberCounter == SoughtMemberIndex)  return loopMember->member_ptr;     }
        else if ((loopMember->if_hidden == yz_false) && (loopMember->member_ptr->jam_status != unjammed))   {
            MemberCounter++;
            if (MemberCounter == SoughtMemberIndex)  return loopMember->member_ptr;     }
    }
    
    return NULL;    // should never happen
}


// SetReturnCoordinate() sets the current 'search' variable to be the specified register.
// If we're trying to define something, draw a fibril to our base 'program counter' fibril.

void SetReturnCoordinate(Ulong ReturnBundleID)
{
    run_retrn retrn;

    SearchCoordinate._Bundle = BIF_ReturnBundle(ReturnBundleID);
    if (GL_Define.DefineOn == yz_true)  {
        DerefFibril(&(SearchCoordinate._Fibril));
        retrn = DrawFibril(SearchCoordinate._Bundle, PCCoordinate._Fibril, &SearchCoordinate._Fibril, 
                            PCCoordinate._Offset, PCCoordinate._IndexWidth, yz_true, PCCodeRef.PLL_index, 0, NULL);
        if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  SearchCoordinate._Fibril = NULL;  return;  }
        RefFibril(SearchCoordinate._Fibril);         }
    SearchCoordinate._Offset = 0;
    SearchCoordinate._IndexWidth = 1;
}
