/*
 *  extrnl.h
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
 

#ifndef Externals_h
#define Externals_h

#include "yazooh.h"
#include "intrpt.h"



// These definitions save a lot of time in the .c/.cpp files

#define BIF_InputBundle(ID) (((member *) FElement(&(BIF_ArgsCoordinate._Bundle->variable_ptr->mem.members), ID))->member_ptr)
#define BIF_AllFunctions (((member *) FElement(&(Zero->variable_ptr->mem.members), _yz_registers))->member_ptr)
#define BIF_ReturnBundle(ID) (((member *) FElement(&(BIF_AllFunctions->variable_ptr->mem.members), ID))->member_ptr)



// type of call() invocation -- only fast_call is currently being used

#define no_call 0
#define full_call 1



// Globals

typedef struct {
    char _ExtCallMode;
} yz_external_global_struct;

extern yz_external_global_struct yz_external_globals;

#define ExtCallMode yz_external_globals._ExtCallMode



// Prototypes

extern const Ulong FileReadBufferSize;

extern void(*InternalVarJumpTable[internal_vars_num])();
extern void(*BuiltInFunctionJumpTable[bi_commands_num])();

extern void ext_call();
extern void ext_fast_call();
extern void UserFunctionCall();

extern void ext_compile();
extern void ext_transform();
extern void ext_load();
extern void ext_save();
extern void ext_input();
extern void ext_print();
extern void PrintOneBundle(bundle *, Ulong, Ulong);
extern char ASCII_Hex(unsigned char);
extern void ext_read_string();
extern void ext_print_string();
extern void BundleAndString(Ulong);
extern void ext_clip();
//extern fibril *MakeClippedFibril(fibril *);
extern void ext_trap();
extern run_retrn SetCodeRef(variable *, code_ref *);
extern void ext_throw();
extern void ext_top();
extern void ext_size();

extern void ext_abs();
extern double Math_Abs(double);
extern void ext_round_down();
extern double Math_Round_Down(double);
extern void ext_round_up();
extern double Math_Round_Up(double);
extern void ext_log();
extern double Math_Log(double);
extern void ext_cos();
extern double Math_Cos(double);
extern void ext_sin();
extern double Math_Sin(double);
extern void ext_tan();
extern double Math_Tan(double);
extern void ext_acos();
extern double Math_ArcCos(double);
extern void ext_asin();
extern double Math_ArcSin(double);
extern void ext_atan();
extern double Math_ArcTan(double);
extern void ExtMathTemplate(double(*)(double));

extern void ext_random();
extern void ext_extract();
extern void ext_find();

extern void ext_YZ_variable_type();
extern void ext_YZ_variable_member_top();
extern void ext_YZ_variable_code_top();
extern void GetYazooVariableInfo(Ulong);

extern void ext_YZ_member_type();
extern void ext_YZ_member_code_top();
extern void ext_YZ_member_indices();
extern void ext_YZ_member_ID();
extern void ext_YZ_if_member_targeted();
extern void GetYazooMemberInfo(Ulong);
extern void ext_YZ_if_hidden_member();
extern void ext_YZ_ParentScriptInfo();

extern void ext_YZ_get_variable_code();
extern void ext_YZ_get_member_code();
extern void ext_YZ_get_variable_code_ID();
extern void ext_YZ_get_member_code_ID();
extern void ext_YZ_get_variable_code_offset();
extern void ext_YZ_get_member_code_offset();

extern void GetYazooCode(Ulong, unsigned char);
extern void SpringCleaning();

extern void ext_slong();
extern void ext_double();
extern void ext_string();
extern void ext_composite();
extern void ext_error_code();
extern void ext_error_index();
extern void ext_error_script();
extern void ext_warning_code();
extern void ext_warning_index();
extern void ext_warning_script();

extern void SetReturnCoordinate(Ulong);

extern Ulong MembersNum(variable *);
extern bundle *GetBIF_Member(Ulong);



// Built-in function ID numbers

#define _ext_call 1
#define _ext_protected_call 2
#define _ext_compile 3
#define _ext_transform 4
#define _ext_load 5
#define _ext_save 6
#define _ext_input 7
#define _ext_print 8
#define _ext_read_string 9
#define _ext_print_string 10
#define _ext_clip 11
#define _ext_trap 12
#define _ext_throw 13
#define _ext_top 14
#define _ext_size 15
#define _ext_abs 16
#define _ext_round_down 17
#define _ext_round_up 18
#define _ext_log 19
#define _ext_cos 20
#define _ext_sin 21
#define _ext_tan 22
#define _ext_acos 23
#define _ext_asin 24
#define _ext_atan 25
#define _ext_random 26
#define _ext_extract 27
#define _ext_find 28
#define _ext_YZ_variable_type 29
#define _ext_YZ_variable_member_top 30
#define _ext_YZ_variable_code_top 31
#define _ext_YZ_member_type 32
#define _ext_YZ_member_code_top 33
#define _ext_YZ_member_indices 34
#define _ext_YZ_member_ID 35
#define _ext_YZ_if_member_targeted 36
#define _ext_YZ_if_hidden_member 37
#define _ext_YZ_get_variable_code 38
#define _ext_YZ_get_member_code 39
#define _ext_YZ_get_variable_code_ID 40
#define _ext_YZ_get_member_code_ID 41
#define _ext_YZ_get_variable_code_offset 42
#define _ext_YZ_get_member_code_offset 43
#define _ext_GetParentScriptInfo 44
#define _ext_SpringCleaning 42



// Register IDs

#define R_slong 1
#define R_double 2
#define R_string 3
#define R_composite 4
#define R_error_code 5
#define R_error_index 6
#define R_error_script 7
#define R_warning_code 8
#define R_warning_index 9
#define R_warning_script 10

#endif
