/*
 *  hobbsh.c(pp) - executes bytecode commands
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "yazooh.h"
#include "cmpile.h"
#include "intrpt.h"
#include "hobbsh.h"
#include "extrnl.h"
#include "yzmain.h"
#include <limits.h>
#include <float.h>
#include <string.h>



// Globals

yz_bytecode_global_struct yz_bytecode_globals;

const Ulong GL_MaxRecursions = 100;    // maximum nesting of running codes, either through
                                       // construction {..} or function calls f()




// *** Routines to parse bytecode operators ***
// Each routine name is preceded by a '_', to differentiate it from the actual operator name.
// These should be roughly in the order of their operator ID numbers.


// _func_return() -- throws the return flag (a kind of error code) and sets the address of the return variable.

void _func_return()
{
    PCCodePtr++;

    CallHobbishFunction();
    if (err_code != passed)  return;

    DerefFibril(&ReturnCoordinate._Fibril);
    ReturnCoordinate = SearchCoordinate;
    RefFibril(ReturnCoordinate._Fibril);
    err_code = return_flag;        // not a real error -- just to get RunHobbish/Constructor to stop
}


// _internal_var() returns the address of the requested register.

void _internal_var()
{
    PCCodePtr++;

    InternalVarJumpTable[*PCCodePtr]();
    PCCodePtr++;

    GL_Path.StemMember = NULL;
}


// _built_in_function() runs a built-in Yazoo function and returns the appropriate register, if applicable.

void _built_in_function()
{
    bundle_coordinate HoldArgsCoordinate = BIF_ArgsCoordinate, TempArgsCoordinate;
    define_params HoldGLDefine = GL_Define;
    void(*theBuiltInFunction)();
    Slong BIFunctionID;
    
    
        // read in the function ID
    
    PCCodePtr++;

    BIFunctionID = *PCCodePtr;
    theBuiltInFunction = BuiltInFunctionJumpTable[BIFunctionID];
    PCCodePtr++;
    
    GL_Define.DefineOn = yz_true;
    GL_Define.FinalStep = yz_false;
    GL_Define.RemoveMode = yz_false;
    GL_Define.Flags = 0;            // can_add_members = yz_false;
    
    
        // get the argument variable
    
    CallHobbishFunction();
    if (err_code != passed)  return;
    TempArgsCoordinate = SearchCoordinate;
    SearchCoordinate._Fibril = NULL;      // for referencing
    
    BIF_ArgsCoordinate = TempArgsCoordinate;
    
    
        // run the function
    
    ReturnCoordinate._Bundle = NULL;
    theBuiltInFunction();


        // set the return variable and restore the old args variable
    
    SearchCoordinate._Fibril = NULL;
    SearchCoordinate._Offset = 0;
    SearchCoordinate._IndexWidth = 1;
    GL_Path.StemMember = NULL;

    DerefFibril(&(BIF_ArgsCoordinate._Fibril));
    BIF_ArgsCoordinate = HoldArgsCoordinate;
    
    GL_Define = HoldGLDefine;
}


// _user_function() runs a user-defined function and returns the return variable.

void _user_function()
{
    bundle_coordinate FunctionCoordinate, HoldArgsCoordinate = ArgsCoordinate;
    linkedlist BackupCodeList, *theCodeList;
    Ulong counter, HoldCodeNumber;
    define_params HoldGLDefine = GL_Define;
    LL_retrn linkedLretrn;
    run_retrn retrn;
    Boolean CleanUpCodeLL = yz_false;
    
    
        // First determine which code marker N to skip to.  N = 0 would run the constructor.
        // In theory, N can be anything, but the compiler only generates N = 1 function calls.

    PCCodePtr++;
    
    GL_Define.DefineOn = yz_true;
    GL_Define.FinalStep = yz_false;
    GL_Define.RemoveMode = yz_false;
    GL_Define.Flags = 0;            // can_add_members = yz_false;    
    
    
        // Get the function variable.
    
    CallCodeFunction();           // Assume SC._Bundle != 0. (?)
    if ((SearchCoordinate._Bundle == NULL) || (err_code != passed))  return;    // SC._Bundle == NULL if called in constructor mode
    FunctionCoordinate = SearchCoordinate;
    HoldCodeNumber = CodeNumber;
    
    BackupCodeList.memory = NULL;       // BeginExec() will likely change GL_Object.CodeList
    theCodeList = GL_Object.CodeList;
    if (theCodeList == NULL)  {  GL_Define = HoldGLDefine;  return;  }
        
        
        // Technical point:  make a backup of GL_Object.CodeList since that could be overwritten if it's stored in the CodeLL object.
        
    if ((theCodeList->elementnum > 1) || (theCodeList == &(GL_Object.CodeLL)))   {
        linkedLretrn = FNewLinkedList(&BackupCodeList, GL_Object.CodeList->elementnum, sizeof(code_ref), 0, uncleared);
        if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr);  return;  }
        theCodeList = &BackupCodeList;
        CleanUpCodeLL = yz_true;

        for (counter = 1; counter <= theCodeList->elementnum; counter++)    {
            *(code_ref *) FElement(&BackupCodeList, counter) = *(code_ref *) FElement(GL_Object.CodeList, counter);
            Ref_CodeRef((code_ref *) FElement(&BackupCodeList, counter));
    }   }


        // Now get the function arguments.
    
    CallHobbishFunction();
    if (err_code != passed)    {
        if (CleanUpCodeLL == yz_true)    {
            for (counter = 1; counter <= theCodeList->elementnum; counter++)   {        // deref & delete backup code list
                Deref_CodeRef((code_ref *) FElement(&BackupCodeList, counter));   }
            FDeleteLinkedList(&BackupCodeList);   }
        return;      }

    HoldArgsCoordinate = ArgsCoordinate;
    ArgsCoordinate = SearchCoordinate;  // Don't deref ArgsCoordinate._Fibril -- it's stored in HoldArgsCoordinate
    SearchCoordinate._Fibril = NULL;       // to prevent double derefencing
    
    
        // Run the function.
        
    ReturnCoordinate._Bundle = NULL;
    for (counter = 1; counter <= theCodeList->elementnum; counter++)  {
        BeginExecution((code_ref *) FElement(theCodeList, counter), yz_true, FunctionCoordinate._Offset,
                                    FunctionCoordinate._IndexWidth, HoldCodeNumber);
        if (err_code != passed)  break;     }     // return_flag shows up in err_code
        
        
        // deref & delete backup code list
    
    if (CleanUpCodeLL == yz_true)      {
        for (counter = 1; counter <= theCodeList->elementnum; counter++)    {
            Deref_CodeRef((code_ref *) FElement(&BackupCodeList, counter));     }
        FDeleteLinkedList(&BackupCodeList);    }

    if (err_code == return_flag)  err_code = passed;    // a return flag is a legitimate excuse for leaving the function
    
    
        // Set SearchCoordinate to the return variable, and draw a fibril to it if necessary (f().q :: ...).
        
    if (err_code == passed)    {
        if (ReturnCoordinate._Bundle == NULL)  GL_Object.Type = no_type;    // otherwise, f:: { .. args()[1] .. }, .. , f(8) won't work
        RefFibril(ReturnCoordinate._Fibril);          // indirectly, ref the new SearchCoordinate._Fibril
        DerefFibril(&(SearchCoordinate._Fibril));     // (the args function has no return statement)
        SearchCoordinate = ReturnCoordinate;
        
        if ((HoldGLDefine.DefineOn == yz_true) && (SearchCoordinate._Bundle != NULL))    {   // if in def mode, draw our own fibril from PCCoord
            DerefFibril(&SearchCoordinate._Fibril);
            retrn = DrawFibril(SearchCoordinate._Bundle, PCCoordinate._Fibril, &SearchCoordinate._Fibril,
                                    PCCoordinate._Offset, PCCoordinate._IndexWidth, yz_true, PCCodeRef.PLL_index, 0, NULL);
            if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  SearchCoordinate._Fibril = NULL;  return;  }
            RefFibril(SearchCoordinate._Fibril);     }
    }


        // restore the old args variable and clean up

    GL_Path.StemMember = NULL;

    DerefFibril(&(ArgsCoordinate._Fibril));
    ArgsCoordinate = HoldArgsCoordinate;
    ReturnCoordinate._Bundle = NULL;
    GL_Define = HoldGLDefine;
}


// _get_args() returns the args variable

void _get_args()
{
    PCCodePtr++;

    RefFibril(ArgsCoordinate._Fibril);
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = ArgsCoordinate;
    GL_Path.StemMember = NULL;
}


// _jump_if_true() performs a goto by the specified offset if the condition is true

void _jump_if_true()
{
    Slong *JumpPC;

    PCCodePtr++;
    JumpPC = PCCodePtr + *PCCodePtr;
    PCCodePtr++;
    CallLogicalFunction();
    if (GL_Conditional != yz_false)  PCCodePtr = JumpPC;
}


// _jump_if_false() performs a goto by the specified offset if the condition is false

void _jump_if_false()
{
    Slong *JumpPC;

    PCCodePtr++;
    JumpPC = PCCodePtr + *PCCodePtr;
    PCCodePtr++;
    CallLogicalFunction();
    if (GL_Conditional == yz_false)  PCCodePtr = JumpPC;
}


// _jump_always() is an unconditional goto

void _jump_always()
{
    PCCodePtr++;
    PCCodePtr += *(signed long *) PCCodePtr;
}


// _exit_Yazoo() sets the exit flag
// maybe somewhat unnecessary since a 'return' might do the trick as well(?)

void _exit_Yazoo()
{
    SetError(finished_signal, PCCodePtr);
    return;
}


// _code_number() changes the code # to run in a function:  e.g. f#0() to run the constructo

void _code_number()
{
    bundle_coordinate HoldSearchCoordinate;
    
    
    PCCodePtr++;
    
    
        // Get the left-hand argument:  the function
    
    CallCodeFunction();
    if (err_code != passed)  return;
    if (SearchCoordinate._Bundle->variable_ptr->type != composite_type)
        {  SetError(not_composite_err, PCCodePtr-1);  return;  }
    HoldSearchCoordinate = SearchCoordinate;
    SearchCoordinate._Fibril = NULL;     // to prevent double-derefs
    
    
        // Get the right-hand argument:  the code # to run.
    
    CallNumericFunction();
    if (err_code != passed)  return;
    
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = HoldSearchCoordinate;      // we return the first argument's coordinate
    
    if (!((DPRegister >= 0) && (DPRegister <= ULONG_MAX)))  {  SetError(out_of_range_err, PCCodePtr-1);  return;  }
    CodeNumber = (Ulong) DPRegister;
}


// _sub_code() effects a code substitution:  'a << b' returns the code of 'b' and the variable 'a'

void _sub_code()
{
    fibril **loopFibril, *OldFibril;
    bundle_coordinate HoldSearchCoordinate;
    define_params HoldGLDefine;
    step_params HoldPath;
    Ulong counter, HoldCodeNumber;
    Boolean CustomizeCode;
    run_retrn retrn;
    LL_retrn linkedLretrn;


        // Set appropriate flags so that a new fibril will be drawn to the left-hand argument.
        // We will fiddle with this fibril later.
    
    PCCodePtr++;
    GL_Path.StemMember = NULL;
    HoldGLDefine = GL_Define;
    GL_Define.DefineOn = yz_true;
    GL_Define.FinalStep = yz_false;
    GL_Define.RemoveMode = yz_false;
    GL_Define.Flags = 0;            // can_add_members = yz_false;


        // Get the left-hand argument:  the variable to substitute into.
    
    CallHobbishFunction();
    if (err_code != passed)  return;
    if (SearchCoordinate._Bundle->variable_ptr->type != composite_type)
        {  SetError(not_composite_err, PCCodePtr-1);  return;  }
    HoldSearchCoordinate = SearchCoordinate;
    SearchCoordinate._Fibril = NULL;     // to prevent double-derefs
    HoldPath = GL_Path;
    GL_Define = HoldGLDefine;


        // Get the right-hand argument:  containing the code to substitute.
        // If it's an existing code -- i.e. not an inlined code block -- we'll have to fool with its existing fibril.
        
    if (*PCCodePtr == code_block)  CustomizeCode = yz_true;
    else  CustomizeCode = yz_false;
    InitGLObject();
    CallCodeFunction();
    if (err_code != passed)  return;
    HoldCodeNumber = CodeNumber;
    
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = HoldSearchCoordinate;      // we return the first argument's coordinate
    GL_Path = HoldPath;


        // A bit tricky.  Copy the code into the global GL_Object.CodeLL variable -- Yazoo doesn't just use the code
        // in SearchCoordinate because, by the definition of substitution, we sometimes use code from a different variable.
        
    if (GL_Object.CodeList != &(GL_Object.CodeLL))     {
        DeleteObjectCodeLL(&(GL_Object.CodeLL));
        if (GL_Object.CodeList != NULL)  {
            linkedLretrn = FNewLinkedList(&(GL_Object.CodeLL), GL_Object.CodeList->elementnum, sizeof(code_ref), 0, uncleared);
            if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
            FCopyElements(GL_Object.CodeList, 1, &(GL_Object.CodeLL), 1, GL_Object.CodeList->elementnum);    }
        else  {
            linkedLretrn = FNewLinkedList(&(GL_Object.CodeLL), 0, sizeof(code_ref), 0, uncleared);
            if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }    }

        RefGLObjectCodeLL();
        GL_Object.CodeList = &(GL_Object.CodeLL);     }
        
        
        // Fix the anchoring fibrils of each code in GL_Object.CodeLL:  have their stems drawn either from PCCoordinate (inlined {})
        // or from the original stem of the anchors from the function that donated the code.
        // The 2nd possibility means that the anchors are different but their stems are the same:  only 1st step on search path differs.
        
    for (counter = 1; counter <= GL_Object.CodeLL.elementnum; counter++)  {
        loopFibril = &(((code_ref *) FElement(&(GL_Object.CodeLL), counter))->anchor);     // we use loopF to change GL_O.CodeLL[n]'s anchor
        OldFibril = *loopFibril;
        if ((OldFibril == NULL) && (CustomizeCode == yz_false))      {     // a function donate the code; anchor had no stem
            retrn = DrawFibril(SearchCoordinate._Bundle, NULL, loopFibril,
                                PCCoordinate._Offset, PCCoordinate._IndexWidth, yz_true, PCCodeRef.PLL_index, 0, NULL);
            RefFibril(*loopFibril);   }
        else  {
            if (CustomizeCode == yz_true)     // inlined code was substituted
                retrn = DrawFibril(SearchCoordinate._Bundle, PCCoordinate._Fibril, loopFibril,
                                    PCCoordinate._Offset, PCCoordinate._IndexWidth, yz_true, PCCodeRef.PLL_index, 0, NULL);
            else                              // a function donate the code; reuse the old anchor's stem
                retrn = DrawFibril(SearchCoordinate._Bundle, OldFibril->stem, loopFibril,
                                    PCCoordinate._Offset, PCCoordinate._IndexWidth, yz_true, PCCodeRef.PLL_index, 0, NULL);
            RefFibril(*loopFibril);   // since GL_O.CodeLL's anchor changed
            DerefFibril(&OldFibril);         }
        if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }     }
    
    CodeNumber = HoldCodeNumber;
}


// _append_code() conjoins two code lists:  a : b.  This is what inheritance means in Yazoo.

void _append_code()
{
    linkedlist FirstCodeLL, *FirstCodePtr;
    LL_retrn linkedLretrn;
    bundle_coordinate HoldSearchCoordinate;
    step_params HoldPath;
    Ulong CopyBottom, counter;
    unsigned char FirstType;
    Boolean HoldCanAddMembers = GetFlag(GL_Define.Flags, can_add_members), HoldGLDefineOn = GL_Define.DefineOn, RefInFCLL = yz_false;
    
    
        // Make sure we're drawing a fibril, and then get the left-hand argument.
    
    GL_Define.DefineOn = yz_false;
    PCCodePtr++;
    SetFlag(GL_Define.Flags, can_add_members, yz_false);     // assumes existing code to append to
    
    GL_Path.StemMember = NULL;
    CallCodeFunction();
    if (err_code != passed)  return;

    DerefFibril(&(SearchCoordinate._Fibril));
    HoldSearchCoordinate = SearchCoordinate;
    HoldPath = GL_Path;
    
    FirstType = GL_Object.Type;
    
    
        // This is a slightly delicate spot.  If there is a code in GL_Object.CodeLL, then copy the linked list into our own
        // FirstCodeLL variable and fool Yazoo into thinking CodeLL is free, so there are no problems with the 2nd argument.
        // Otherwise, we just copy the code into FirstCodeLL the normal way, adding references.
    
    FirstCodeLL.memory = NULL;
    if (GL_Object.CodeList == NULL)   FirstCodePtr = NULL;
    else if (GL_Object.CodeLL.memory == NULL)  {
        linkedLretrn = FNewLinkedList(&FirstCodeLL, GL_Object.CodeList->elementnum, sizeof(code_ref), 0, uncleared);
        if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
        FCopyElements(GL_Object.CodeList, 1, &FirstCodeLL, 1, FirstCodeLL.elementnum);
        for (counter = 1; counter <= FirstCodeLL.elementnum; counter++)
            Ref_CodeRef((code_ref *) FElement(&FirstCodeLL, counter));
        RefInFCLL = yz_true;
        FirstCodePtr = &FirstCodeLL;       }
    else  {
        RefInFCLL = yz_true;
        FirstCodeLL = GL_Object.CodeLL;
        FirstCodePtr = &FirstCodeLL;
        GL_Object.CodeLL.memory = NULL;       }
    
    
        // Now get the 2nd argument.  GL_Object.CodeLL should always be NULL here.
        
    GL_Define.DefineOn = yz_false;
    SetFlag(GL_Define.Flags, can_add_members, HoldCanAddMembers);
    CallCodeFunction();
    if (err_code != passed)  {
        if (RefInFCLL == yz_true)               // get rid of our copy of the code lists if we exit due to an error
            DeleteObjectCodeLL(&FirstCodeLL);   // this works even if FirstCodeLL was copied from GL_Object.CodeLL, since the latter was zeroed
        return;     }


        // Prepare GL_Object.CodeLL for holding the concatenated code list.
    
    if (GL_Object.CodeList != &(GL_Object.CodeLL))  DeleteObjectCodeLL(&(GL_Object.CodeLL));
    
    
        // error checking -- make sure all are bona fide codes, not ulongs or something
        
    if ((FirstType != no_type) && (GL_Object.Type != no_type))
    if (((FirstType == composite_type) && (GL_Object.Type != composite_type))
                    || ((FirstType != composite_type) && (GL_Object.Type == composite_type)))  {
        if (RefInFCLL == yz_true)      {
            DeleteObjectCodeLL(&FirstCodeLL);      }
        SetError(type_mismatch_err, PCCodePtr-1);
        return;           }
    if ((FirstType < composite_type) && (GL_Object.Type < composite_type) && (FirstType != GL_Object.Type))         {
        SetError(type_mismatch_err, PCCodePtr-1);
        return;           }
    if (GL_Object.Type == no_type)  GL_Object.Type = FirstType;
    
    
        // reference the arg2 codes since we will make a copy of them
        
    if ((FirstCodePtr != NULL) && (GL_Object.CodeList != NULL))  {
    for (counter = 1; counter <= GL_Object.CodeList->elementnum; counter++)  {
        Ref_CodeRef((code_ref *) FElement(GL_Object.CodeList, counter));
    }}


        // Another ticklish spot.  We do different things for various combinations of arg1/arg2 have/don't have any codes,
        // but the general procedure where they both have codes is to concatenate them in GL_Object.CodeLL.
        
    if (FirstCodePtr != NULL)    {
        if (GL_Object.CodeList != NULL)   {
            CopyBottom = FirstCodePtr->elementnum+1;
            linkedLretrn = FAddElements(FirstCodePtr, GL_Object.CodeList->elementnum, uncleared);
            if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
            
            FCopyElements(GL_Object.CodeList, 1, FirstCodePtr, CopyBottom, GL_Object.CodeList->elementnum);     }
            
        GL_Object.CodeList = &(GL_Object.CodeLL);    // no arg2 codes -- just use CodeLL from arg1
        GL_Object.CodeLL = FirstCodeLL;     }
        
    else if (RefInFCLL == yz_true)       // if there's no arg2 code, but there is an arg1 code
        DeleteObjectCodeLL(&FirstCodeLL);

                            // if no arg1 code or arg2 code, then GL_Object.CodeList will still be NULL, so we're good.
                            

        // Clean up and go.  The codes have been, one way or another, safely bundled into something under the GL_Object.CodeList pointer.
    
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = HoldSearchCoordinate;
    GL_Path = HoldPath;
    GL_Define.DefineOn = HoldGLDefineOn;
}


// _remove_member() deletes a member from a variable.

void _remove_member()  {  Prune(pn_member);  }



// _forced_equate() copies data between two variables whose sizes match (or can be made to match by resizing strings
// in the target variable), irrespective of the types of the variables.

void _forced_equate()
{
    bundle_coordinate SourceCoordinate, DestCoordinate, HoldThatCoordinate;
    step_params HoldPath;
    define_params HoldGLDefine = GL_Define;
    void *TempStorage = NULL, *StringPtr;
    Ulong SourceDataSize = 0, DestDataSize = 0, OldSCIndexWidth, index_counter;//, ExtraSourceDataSize = 0;
    signed long SourceStringSize = no_string, DestStringSize = no_string;
    Boolean CanResizeDestination = yz_false, InlinedSource, ResizedFromZero = yz_false;


        // Get the left-hand argument (the target variable).
    
    GL_Define.DefineOn = yz_true;
    GL_Define.FinalStep = yz_false;
    GL_Define.RemoveMode = yz_false;
    GL_Define.Flags = 0;            // can_add_members = yz_false;
    
    PCCodePtr++;
    GL_Path.StemMember = NULL;
    GL_Define.FinalStep = yz_false;
    if (*PCCodePtr == step_to_all)  CanResizeDestination = yz_true;
    CallHobbishFunction();
    if ((err_code == passed) && (SearchCoordinate._Bundle == NULL))  SetError(not_a_variable_err, PCCodePtr-1);
    if (err_code != passed)  return;

    DestCoordinate = SearchCoordinate;
    RefBundle(DestCoordinate._Bundle);
    SearchCoordinate._Fibril = NULL;           // reference remains for DestCoord._Fibril
    OldSCIndexWidth = GL_Path.OldSearchIndexWidth;

    HoldPath = GL_Path;
    HoldThatCoordinate = ThatCoordinate;      // hold the old variable 'that', in case it gets overwritten by the right argument.
    ThatCoordinate = SearchCoordinate;
    ThatCoordinate._Fibril = NULL;
    
    InlinedSource = yz_true;
    ExtCallMode = no_call;


        // Load the right-hand argument.  We do this in different ways depending on the type of the left-hand variable.
        
    if (*PCCodePtr == constant_Slong)  {
        SourceDataSize = sizeof(signed long);
        StringPtr = (void *) (PCCodePtr+1);
        CallHobbishFunction();          }
    else if (*PCCodePtr == constant_float)  {
        SourceDataSize = sizeof(double);
        StringPtr = (void *) (PCCodePtr+1);
        CallHobbishFunction();              }
    else if (*PCCodePtr == constant_string)  {
        SourceDataSize = (Ulong) *(PCCodePtr+1);
        StringPtr = (void *) (PCCodePtr+2);
        CallHobbishFunction();               }
        
    else  {        // a more complicated right-hand argument:  we size it manually after reading it
        InlinedSource = yz_false;
        CallHobbishFunction();
        if ((err_code == passed) && (SearchCoordinate._Bundle == NULL))  SetError(not_a_variable_err, PCCodePtr-1);
        if (err_code != passed)  {
            DerefFibril(&(SearchCoordinate._Fibril));
            SearchCoordinate = DestCoordinate;
            DerefBundle(&(DestCoordinate._Bundle));
            return;              }
            
        SourceCoordinate = SearchCoordinate;
        ThatCoordinate = HoldThatCoordinate;            // 'that' might have been overwritten
        SizeBundle(SourceCoordinate._Bundle, SourceCoordinate._Offset, SourceCoordinate._IndexWidth, &SourceDataSize, &SourceStringSize);    }


        // Get the byte size of the source variable.
        
    SizeBundle(DestCoordinate._Bundle, DestCoordinate._Offset, DestCoordinate._IndexWidth, &DestDataSize, &DestStringSize);
    if (SourceStringSize != no_string)  SourceDataSize += SourceStringSize;     // we don't care what form the source data is in
    
    
        // If the source and distination variable sizes don't match, we can try resizing the latter if the destination path ended in [*].
        
    if (((SourceDataSize > DestDataSize) && (DestStringSize == no_string)) || (SourceDataSize < DestDataSize))  {

        if (OldSCIndexWidth == 0)  {     // zero indices in source -- hard to deal with, so just throw an error
            SetError(unequal_data_size_err, PCCodePtr-1);
            DerefFibril(&(SearchCoordinate._Fibril));
            SearchCoordinate = DestCoordinate;
            DerefBundle(&(DestCoordinate._Bundle));
            return;                }
        if (HoldPath.StemMember == NULL)  {  SetError(undefined_member_err, PCCodePtr-1);  return;  }


            // if [*] was present and current top of that list is zero, then size it to 1 so we can measure its byte-size per index
            
        if ((CanResizeDestination == yz_true) && (HoldPath.StemMember->indices == 0))     {        
            ResizeBundle(HoldPath.StemCoordinate._Bundle->variable_ptr, HoldPath.StemMember, 1);
            DestCoordinate._IndexWidth = OldSCIndexWidth;
            SizeBundle(DestCoordinate._Bundle, DestCoordinate._Offset, DestCoordinate._IndexWidth, &DestDataSize, &DestStringSize);
            ResizedFromZero = yz_true;         }
            

            // If sizes don't match and the dest. path ended in a [*], try to resize the destination variable.
            
        if (((SourceDataSize > DestDataSize) && (DestStringSize == no_string)) || (SourceDataSize < DestDataSize))  {
            if ((DestDataSize != 0) && (SourceDataSize != DestDataSize))      {     // 2nd condition redundant?

                if ((CanResizeDestination == yz_true) && (SourceDataSize % (DestDataSize/HoldPath.StemMember->indices) == 0))  {
                    ResizeBundle(HoldPath.StemCoordinate._Bundle->variable_ptr, HoldPath.StemMember,     // we can
                                                            (HoldPath.StemMember->indices*SourceDataSize)/DestDataSize);
                    DestCoordinate._IndexWidth = (DestCoordinate._IndexWidth*SourceDataSize)/DestDataSize;
                    DestDataSize = SourceDataSize;         }
                
                else  {       // we can't -- throw an error (and, if we sized it from 0 --> 1, reset it to size 0)
                    if (ResizedFromZero == yz_true)  ResizeBundle(HoldPath.StemCoordinate._Bundle->variable_ptr, HoldPath.StemMember, 0);
                    SetError(unequal_data_size_err, PCCodePtr-1);
                    DerefFibril(&(SearchCoordinate._Fibril));
                    SearchCoordinate = DestCoordinate;
                    DerefBundle(&(DestCoordinate._Bundle));
                    return;           }        }

            else  {    // if the dest. variable is still size-0, there's no hope.
                if (ResizedFromZero == yz_true)  ResizeBundle(HoldPath.StemCoordinate._Bundle->variable_ptr, HoldPath.StemMember, 0);
                SetError(unequal_data_size_err, PCCodePtr-1);
                DerefFibril(&(SearchCoordinate._Fibril));
                SearchCoordinate = DestCoordinate;
                DerefBundle(&(DestCoordinate._Bundle));
                return;          }
    }   }
    if (DestStringSize != no_string)  DestStringSize = SourceDataSize-DestDataSize;
    
    
        // set up a temporary storage buffer for the data to copy
        
    if (InlinedSource == yz_false)     {
        TempStorage = (void *) malloc((size_t) SourceDataSize + sizeof(char));      //  avoid size-0 malloc
        if (TempStorage == NULL)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
        StringPtr = TempStorage;
        
            // read the data in
        for (index_counter = 0; index_counter < SourceCoordinate._IndexWidth; index_counter++)
            ReadBundle(SourceCoordinate._Bundle, SourceCoordinate._Offset+index_counter, 1, &StringPtr, &SourceStringSize);
        StringPtr = TempStorage;       }


        // write the data from the buffer to the target variable
        
    for (index_counter = 0; index_counter < DestCoordinate._IndexWidth; index_counter++)
        WriteBundle(DestCoordinate._Bundle, DestCoordinate._Offset+index_counter, 1, &StringPtr, &DestStringSize);


        // free the memory & clean up
    
    if (InlinedSource == yz_false)  free(TempStorage);

    GL_Path = HoldPath;
    GL_Define = HoldGLDefine;
    
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = DestCoordinate;
    DerefBundle(&(DestCoordinate._Bundle));
}


// _def_general() handles all define/define-at/equate operators.
// Most operators are handled with the general procedure, which comes first.  Equate is done separately for speed.
// Unlike forced-equate, def-general requires data types to have compatible types (the same composite structure,
// interconvertible primitive types), which does not imply byte-size equivalence.

void _def_general()
{
    linkedlist *theCodeList = NULL;
    bundle_coordinate SourceCoordinate, DestCoordinate, HoldThatCoordinate, HoldSearchCoordinate;
    code_ref *loopCodeRef;
    bundle *NewBundle, *ProxyBundle;
    fibril *CodeStemFibril;
    Ulong counter, FirstToCustomize, ProxyOffset, OldSCIndexWidth, AliasIndices;
    Ulong StartCommandIndex = PCCodePtr - PCCodeRef.code_ptr + 1;
    Boolean HoldDefineOn = GL_Define.DefineOn, CanResizeDestination = yz_false, IfCanJam = yz_true, IfUnjammed, ErrorInConstructor;
    Boolean IfAddMember, IfProxy = yz_false;
    Slong *DGCommandPtr, *SubjectCommand, *ObjectCommand, TDGN_store;
    define_params HoldDefineParams;
    step_params HoldPath;
    run_retrn retrn;
    
    
        // read in the flags
    
    DGCommandPtr = PCCodePtr;
    PCCodePtr++;
    GL_Define.Flags = *PCCodePtr;
    if (GetFlag(GL_Define.Flags, unjammable) == yz_true)  IfCanJam = yz_false;
    if (GetFlag(GL_Define.Flags, make_proxy) == yz_true)  IfProxy = yz_true;
    PCCodePtr++;
    
    
        // equate resizes arrays if sizes don't match & [*] is present at the end of the target variable path
    
    if (*PCCodePtr == step_to_all)  CanResizeDestination = yz_true;
    
    
        // The following block handles all non-equate operators.  (It should do equate too, but would be slow.)
    
    SubjectCommand = PCCodePtr;
    if (GL_Define.Flags != equ_flags)   {
        GL_Define.RemoveMode = yz_false;
        
        GL_Define.DefineOn = yz_false;
        HoldDefineParams = GL_Define;
        
// Begin define block -- get the destination variable
        GL_Path.StemMember = NULL;
        GL_Define.DefineOn = yz_true;
        GL_Define.FinalStep = yz_true;
        
        CallHobbishFunction();
        if (err_code != passed)  return;
        
        IfAddMember = AddedAMember;         // actually the last one was NOT added, but the boolean variable was set anyway
// End define block
        
        TDGN_store = TempDefGenName;
        if ((*SubjectCommand >= search_member) && (*SubjectCommand <= resize) && (IfAddMember == yz_true))  {
            retrn = AddMember(SearchCoordinate._Bundle, &(GL_Path.StemMemberNumber), GL_Path.Indices, SearchCoordinate._IndexWidth,
                              IfProxy, &GL_Path.StemMember, GetFlag(HoldDefineParams.Flags, hidden_member));
            if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  DerefBundle(&(GL_Object.Target._Bundle));  return;  }
            
            GL_Path.StemMember->member_ID = TempDefGenName;      // to hide it from the object block
            TempDefGenName--;
            GL_Path.StemMember->type = no_type;     }
        
// Begin object block -- get the source variable
        GL_Define.RemoveMode = yz_false;
        GL_Define.DefineOn = yz_false;
        
        InitGLObject();
        HoldSearchCoordinate = SearchCoordinate;
        HoldPath = GL_Path;        // important -- the last step will not be taken
        SearchCoordinate._Fibril = NULL;
        ObjectCommand = PCCodePtr;       // we can get the pointer to a constant string from this
        CallCodeFunction();
        TempDefGenName = TDGN_store;
        
        if ((err_code == void_member_err) || (err_code == void_proxy_err))  {
        if (GetFlag(GL_Define.Flags, relink_target) == yz_true)  {
            GL_Object.IfVoid = yz_true;
            err_code = passed;
        }}
        
        if (err_code != passed)  {
            DerefFibril(&(HoldSearchCoordinate._Fibril));
            if ((*SubjectCommand >= search_member) && (*SubjectCommand <= resize) && (IfAddMember == yz_true))  {
                retrn = FindMemberID(HoldSearchCoordinate._Bundle, TDGN_store, &HoldPath.StemMember, &HoldPath.StemMemberNumber);
                if (retrn == passed)  HoldPath.StemMember->member_ID = HoldPath.StemMemberID;     }
            return;     }
        
        if ((SearchCoordinate._Bundle != NULL) && (GL_Object.Type == block_type))
            GL_Object.TypeSize = SearchCoordinate._Bundle->variable_ptr->mem.data.elementsize;
        
        GL_Object.Target = SearchCoordinate;
        DerefFibril(&(SearchCoordinate._Fibril));
        SearchCoordinate = HoldSearchCoordinate;
// End object block
        
        GL_Path = HoldPath;
        
        
            // See if we have to add a variable; if so, find it and step into the member.
            // Trick:  the step will fail & cause an error.  The same would have happened if the member existed but was void.
            // The void-member error is caught later down, and signals to def-general that it should also construct a variable, flags permitting.
        
        if ((*SubjectCommand < search_member) || (*SubjectCommand > insert_indices_down))  {
            IfAddMember = yz_false;
            FailedLastStep = yz_false;
            IfUnjammed = yz_false;       }
        
        else    {
            
            if (IfAddMember == yz_true)  {          // refind the member in case the member # changed (i.e. something was removed beneath it)
                retrn = FindMemberID(SearchCoordinate._Bundle, TDGN_store, &GL_Path.StemMember, &GL_Path.StemMemberNumber);
                if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
                GL_Path.StemMember->member_ID = GL_Path.StemMemberID;       }
//                retrn = AddMember(SearchCoordinate._Bundle, &(GL_Path.StemMemberNumber), GL_Path.Indices, SearchCoordinate._IndexWidth,
//                                        IfProxy, &GL_Path.StemMember, GetFlag(HoldDefineParams.Flags, hidden_member));
//                if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  DerefBundle(&(GL_Object.Target._Bundle));  return;  }
//                
//                GL_Path.StemMember->member_ID = GL_Path.StemMemberID;       }
//                GL_Path.StemMember->type = no_type;     }
            
            
                // If we're defining something, we have to have provided a member (so f() :: ... is illegal).
                // Step into that member, and if we fail, we'll try to fix it below.
                
            StepSearchBundleAndFibril(GL_Path.StemMember, GL_Path.StemMemberNumber, GL_Path.Offset, GL_Path.Indices);
            if (err_code != passed)  FailedLastStep = yz_true;
            
            if ((err_code == unjammed_member_err) || (err_code == unjammed_proxy_err))  IfUnjammed = yz_true;
            else  IfUnjammed = yz_false;               }      // guaranteed true otherwise; see Navigate()
        
        GL_Define = HoldDefineParams;     // clears DefineOn.  Need to do this in case we bail out due to error
        
        if ((FailedLastStep == yz_false) && (err_code != passed))  return;
        err_code = passed;
        
        
            // if we couldn't step & we're not allowed to fix it, throw an error
        
        if ((FailedLastStep == yz_false) && (GetFlag(GL_Define.Flags, update_members) == yz_true))  {
            if (GL_Path.StemMember == NULL)  SetError(undefined_member_err, PCCodePtr-1);
            else if (GL_Path.StemMember->member_ptr != NULL)  {
                if ((GL_Path.StemMember->member_ptr->variable_ptr->type != proxy_type) && (IfProxy == yz_true))  {
                    SetError(type_mismatch_err, DGCommandPtr);
                    err_index = StartCommandIndex;
                    return;
        }   }   }
        
        if (GL_Object.Target._Bundle != NULL) {     // really not sure what this is supposed to do -- seems never to have run on my watch
        if (((GL_Object.Target._Bundle->variable_ptr->instances != GL_Object.Target._IndexWidth) || (GL_Object.Target._Offset != 0)) // latter condition
                   && (GetFlag(GL_Define.Flags, relink_target) == yz_true) && (PCCoordinate._IndexWidth != 1))  // for a[0] :: {b[2], c=@b[2]} (IW = 0)
                {  SetError(incomplete_variable_err, SubjectCommand);  return;  }  }
        
        
            // Here we try to resize the destination variable if the sizes don't match.
            
        AliasIndices = SearchCoordinate._IndexWidth;    // AliasIndices:  a proper count of indices even if the last member was void
        if (FailedLastStep == yz_true)  AliasIndices *= GL_Path.Indices;
        
        if ((AliasIndices != GL_Object.Target._IndexWidth) && (GL_Object.Target._Bundle != NULL) && (GL_Object.IfVoid == yz_false))   {
            if ((CanResizeDestination == yz_true) && (GL_Path.StemMember != NULL))  {
                Ulong CompareIndices;
                
                if (GL_Path.OldSearchIndexWidth == 0)  {  SetError(mismatched_indices_err, DGCommandPtr);  return;  }
                
                if (GL_Path.StemMember->indices == 0)  CompareIndices = GL_Path.OldSearchIndexWidth;     // in case of a[*] where top(a) == 0
                else if (FailedLastStep == yz_false)  CompareIndices = SearchCoordinate._IndexWidth/GL_Path.StemMember->indices;
                else  CompareIndices = SearchCoordinate._IndexWidth;
                if (GL_Object.Target._IndexWidth % CompareIndices != 0)  {  SetError(mismatched_indices_err, DGCommandPtr);  return;  }
                
                if ((GetFlag(GL_Define.Flags, relink_target) == yz_true) && (GL_Path.StemMember->member_ptr != NULL))   {    // don't jam during Resize()
                    DerefBundle(&(GL_Path.StemMember->member_ptr));
                    SearchCoordinate._Bundle = NULL;     }
                
                        // ResizeBundle() only changes the member indices -- no variable resize -- if VarToUnlink was unlinked
                ResizeBundle(GL_Path.StemCoordinate._Bundle->variable_ptr, GL_Path.StemMember, GL_Object.Target._IndexWidth/CompareIndices);
                
                if (err_code != passed)  return;
                SearchCoordinate._IndexWidth = GL_Object.Target._IndexWidth;
                GL_Path.Indices = GL_Path.StemMember->indices;     // otherwise Relink() will get confused
        }   }
        
        
            // handle void targets
        
        if ((GL_Object.Type == no_type) || (IfProxy == yz_true))  GL_Object.IfVoid = yz_true;
        
        
            // Catch type-mismatch errors before they make us unlink the whole target.  Check member type, then variable type.
            
        if (GL_Path.StemMember != NULL)   {
            CheckType(&(GL_Path.StemMember->code_list), &(GL_Path.StemMember->type), GetFlag(GL_Define.Flags, update_members));
            if (err_code != passed)  {  err_index = StartCommandIndex;  return;  }
            
            if ((GetFlag(GL_Define.Flags, relink_target) == yz_false) && (SearchCoordinate._Bundle != NULL))  {
                CheckType(&(SearchCoordinate._Bundle->variable_ptr->code_list), &(SearchCoordinate._Bundle->variable_ptr->type),
                                                GetFlag(GL_Define.Flags, new_target));
                if (err_code != passed)  {  err_index = StartCommandIndex;  return;  }   }   }
        
        
            // If we're re-aliasing (a.k.a. relinking) a variable, handle that here
        
        if ((GetFlag(GL_Define.Flags, relink_target) == yz_true) || (IfUnjammed == yz_true))  {
            Boolean UnlinkOnly = yz_false;
            
            if (GetFlag(GL_Define.Flags, relink_target) == yz_false)
                UnlinkOnly = yz_true;     // in other words, if we're here only to get rid of an unjammed member
                
            RefFibril(GL_Path.StemCoordinate._Fibril);
            DerefFibril(&(SearchCoordinate._Fibril));
            SearchCoordinate = GL_Path.StemCoordinate;     // Note that GL_Path.StemMember is now in the same variable as SC.
            
            RelinkGLSearchMember(&GL_Define, &GL_Object, UnlinkOnly);
            if (err_code != passed)  return;
            if (GL_Object.IfVoid == yz_false)       // step into the newly-aliased variable
                StepSearchBundleAndFibril(GL_Path.StemMember, GL_Path.StemMemberNumber, GL_Path.Offset, GL_Path.Indices);    }
        
        
            // Here we create a new variable if that is called for.
        
        if ((GetFlag(GL_Define.Flags, new_target) == yz_true) && (SearchCoordinate._Bundle == NULL))  {
            RefFibril(GL_Path.StemCoordinate._Fibril);
            DerefFibril(&(SearchCoordinate._Fibril));
            SearchCoordinate = GL_Path.StemCoordinate;     // Note that GL_Path.StemMember is now in the same variable as SC.
            
            
                // The second line below prevents crashes due to relinking unlinked arrays (a[*]=@*) by a[6]:=@x.
                // In relinking existing members, GL_Path.Indices is reset by sti assuming a new member will be created.
            
            if (GL_Path.StemMember == NULL)  {  SetError(undefined_member_err, SubjectCommand);  return;  }
            if (GL_Path.Indices != GL_Path.StemMember->indices)    {
            if (GL_Path.StemMember->member_ptr != NULL)    {
            if (GL_Path.StemMember->member_ptr->variable_ptr->type != proxy_type)    {
                SetError(incomplete_variable_err, SubjectCommand);
                return;
            }}}
            
                // For the following, SC_IW & SC_O are their old values before the aborted step, even if SC_B was previously set to zero.
                // For proxies, they are the values they took in the proxy variable itself, not the composite stem.
            
            if (GL_Object.IfVoid == yz_false)   {
                if (GL_Path.StemMember->member_ptr == NULL)    {     // no proxy -- assuming that it is already unlinked
                    
                    if (SearchCoordinate._Bundle != NULL)    {
                    if ((SearchCoordinate._IndexWidth != SearchCoordinate._Bundle->variable_ptr->instances)
                                || (GL_Path.Indices != GL_Path.StemMember->indices))
                        {  SetError(incomplete_variable_err, PCCodePtr-1);  return;  }       }
                    
                    AddVB(&(GL_Path.StemMember->member_ptr), SearchCoordinate._IndexWidth*GL_Path.Indices, GL_Object.Type, IfCanJam);
                    if (err_code != passed)  return;    }
                
                else   {
                    for (counter = 0; counter < SearchCoordinate._IndexWidth*GL_Path.Indices; counter++)  {
                        ProxyBundle = SearchCoordinate._Bundle;
                        ProxyOffset = SearchCoordinate._Offset;
                        StepToMember(&ProxyBundle, &ProxyOffset, GL_Path.StemMember, GL_Path.Offset+counter);
                        AddVB((bundle **) FElement(&(ProxyBundle->variable_ptr->mem.data), ProxyOffset+1), 1, GL_Object.Type, IfCanJam);
                        if (err_code != passed)  return;
                }   }
                
                StepSearchBundleAndFibril(GL_Path.StemMember, GL_Path.StemMemberNumber, GL_Path.Offset, GL_Path.Indices);
        }   }
        
        GL_Define.DefineOn = HoldDefineOn;
        
        
            // Check/update the member code list.
            // Don't always need to do this since the member is not more restrictive than the variable.
            
        if (err_code == passed)  {
            if (GL_Path.StemMember == NULL)  SetError(undefined_member_err, SubjectCommand);
            else  {
                if (GetFlag(GL_Define.Flags, update_members) != yz_false)    {    
                    UpdateType(&(GL_Path.StemMember->code_list), &(GL_Path.StemMember->type), &(GL_Path.StemCoordinate), yz_true, yz_true);
        }   }   }
        
        
            // At last, we can check/update the variable code list, and run the constructors of the new object, if that is allowed.
        
        ErrorInConstructor = yz_false;
        if ((GL_Object.IfVoid == yz_false) && (err_code == passed) &&
                        ((GetFlag(GL_Define.Flags, new_target) == yz_true) || (GetFlag(GL_Define.Flags, relink_target) == yz_true)))  {
            
            
                // First check the variable codes.
                
            FirstToCustomize = SearchCoordinate._Bundle->variable_ptr->code_list.elementnum+1;
            if (GetFlag(GL_Define.Flags, new_target) != yz_false)     {
                UpdateType(&(SearchCoordinate._Bundle->variable_ptr->code_list), &(SearchCoordinate._Bundle->variable_ptr->type),
                                                &SearchCoordinate, yz_false, yz_false);    }
            if ((err_code == passed) && (GL_Object.Type == block_type) &&
                        (GL_Object.TypeSize != SearchCoordinate._Bundle->variable_ptr->mem.data.elementsize))
                SetError(type_mismatch_err, DGCommandPtr);
            
            
                // Now customize the codes.
                // To 'customize' is to make new anchors for the new codes.
            
            if ((err_code == passed) && (GetFlag(GL_Define.Flags, new_target) == yz_true))  {
                theCodeList = &(SearchCoordinate._Bundle->variable_ptr->code_list);
                for (counter = FirstToCustomize; counter <= theCodeList->elementnum; counter++)   {
                    
                    fibril *NewStemFibril;
                    
                    loopCodeRef = (code_ref *) FElement(theCodeList, counter);
                    
                        // if a new code, then link it to PCCoordinate._Fibril
                        // (use SC_F if possible since that has the correct stem_member field)
                    
                    NewStemFibril = NULL;
                    
                    if (loopCodeRef->anchor == NULL)  {
                        if (SearchCoordinate._Fibril->stem == PCCoordinate._Fibril)  loopCodeRef->anchor = SearchCoordinate._Fibril;
                        else  NewStemFibril = PCCoordinate._Fibril;       }
                    
                        // if an old code, draw a fibril from the old stem to the new variable
                    
                    else    {
                        CodeStemFibril = loopCodeRef->anchor->stem;
                        if (CodeStemFibril != NULL)
                            if (CodeStemFibril->bundle_ptr->variable_ptr->type == proxy_type)  CodeStemFibril = CodeStemFibril->stem;
                        
                        if (CodeStemFibril == NULL)  {
//                            retrn = AddFibril(SearchCoordinate._Bundle, NULL, &(loopCodeRef->anchor), 0, GL_Path.Indices, yz_false,
                            retrn = DrawFibril(SearchCoordinate._Bundle, NULL, &(loopCodeRef->anchor), 0, GL_Path.Indices, yz_false,
                                              SearchCoordinate._Fibril->source_code, 0, NULL);
                            if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }    }
                        else  NewStemFibril = CodeStemFibril;     }
                    
                    
                        // If we had to make a new stem fibril, make a new bundle in its variable for it.
                        // Then draw the new stem to a new anchor in our new variable.
                    
                    if (NewStemFibril != NULL)       {
                        retrn = AddBundle(NewStemFibril->bundle_ptr->variable_ptr, 0, 0, &NewBundle, yz_true);
                        if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
//                        retrn = AddFibril(SearchCoordinate._Bundle, NewStemFibril, &(loopCodeRef->anchor), 0, GL_Path.Indices, yz_false,
                        retrn = DrawFibril(SearchCoordinate._Bundle, NewStemFibril, &(loopCodeRef->anchor), 0, GL_Path.Indices, yz_false,
                                                                            SearchCoordinate._Fibril->source_code, 0, NewBundle);
                        if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
                        NewBundle->remote_fibril = loopCodeRef->anchor;      }
                    
                    Ref_CodeRef(loopCodeRef);
            }   }
            else if (GetFlag(GL_Define.Flags, run_constructor) == yz_true)
                theCodeList = &(SearchCoordinate._Bundle->variable_ptr->code_list);
            
            
                // Run all codes.  Don't start at FirstToCustomize (some may not have properly initialized the first time
                // (if run in constructor mode, it doesn't know if there have been errors, which are tolerated))
            
            if ((GetFlag(GL_Define.Flags, run_constructor) == yz_true) && (err_code == passed))  {
            for (counter = 1; counter <= theCodeList->elementnum; counter++)   {
                BeginExecution((code_ref *) FElement(theCodeList, counter), yz_true, SearchCoordinate._Offset, SearchCoordinate._IndexWidth, 0);
                if (err_code == return_flag)  err_code = passed;
                if (err_code != passed)  {  ErrorInConstructor = yz_true;  break;  }
        }   }}
        GL_Define = HoldDefineParams;    // in case it has been changed by the constructor
        
        DeleteObjectCodeLL(&(GL_Object.CodeLL));
        
        if (GL_Object.IfVoid == yz_true)  SearchCoordinate._Bundle = NULL;    // if we ended up nowhere, make sure Yazoo knows that
        
        if (err_code != passed)  {        // handle any lingering errors (e.g. from construction)
            if (ErrorInConstructor == yz_false)  SetError(err_code, DGCommandPtr);
            
            RefFibril(GL_Path.StemCoordinate._Fibril);
            DerefFibril(&(SearchCoordinate._Fibril));
            SearchCoordinate = GL_Path.StemCoordinate;     // Note that GL_Path.StemMember is now in the same variable as SC.
            RelinkGLSearchMember(&GL_Define, &GL_Object, yz_true);
            SearchCoordinate._Bundle = NULL;
            return;      }
        
        if ((GetFlag(GL_Define.Flags, post_equate) == yz_false) || (SearchCoordinate._Bundle == NULL))  return;   // if no data copy, exit here
        
        
            // We can do some simple data copying right here -- from numeric or string source types.
        
        if ((SearchCoordinate._IndexWidth != 1) && (*ObjectCommand >= addf) && (*ObjectCommand <= constant_string))     {
            SetError(multiple_indices_not_allowed_err, PCCodePtr-1);
            return;     }
        if ((*ObjectCommand >= addf) && (*ObjectCommand <= constant_float))    {
            SaveDPRegister(FElement(&(SearchCoordinate._Bundle->variable_ptr->mem.data), SearchCoordinate._Offset+1),
                                                SearchCoordinate._Bundle->variable_ptr->type);
            if (err_code == out_of_range_err)  {  err_code = passed;  SetWarning(out_of_range_err, PCCodePtr-1);  }
            return;         }
        else if (*ObjectCommand == constant_string)
            {  LoadString(&SearchCoordinate, (void **) &ObjectCommand);  return;  }
        
        if (*ObjectCommand == code_block)  return;    // to allow a := {}, since a :: {} and a = {} are both allowed
        
        
            // OK, we need to do something more elaborate, so we'll fall through to the bottom after the setup for equate
        
        DestCoordinate = SearchCoordinate;
        SourceCoordinate = GL_Object.Target;     }
    
    
        // Everything afterwards is for copying data.  The first block is the setup for pure equate, to make up for missing the last block.
    
    else    {       // simple equate (separate for speed)
        
        
            // Get the first (left-hand) argument:  the destination variable
        
        GL_Path.StemMember = NULL;
        CallHobbishFunction();
        if (err_code != passed)  return;
        DestCoordinate = SearchCoordinate;
        if (DestCoordinate._Bundle == NULL)  {  SetError(not_a_variable_err, SubjectCommand);  return;  }
        else if ((DestCoordinate._IndexWidth != 1) && (*PCCodePtr >= addf) && (*PCCodePtr <= constant_string))     {
            SetError(multiple_indices_not_allowed_err, PCCodePtr-1);        // otherwise it may incorrectly throw a 'not a variable' err
            return;     }
        RefFibril(DestCoordinate._Fibril);
        
        HoldPath = GL_Path;
        HoldThatCoordinate = ThatCoordinate;
        ThatCoordinate = SearchCoordinate;      // store that
        ThatCoordinate._Fibril = NULL;
        OldSCIndexWidth = GL_Path.OldSearchIndexWidth;
        
            // Here we process the source variable (2nd argument) differentially depending on what we're expecting.
        
        if ((DestCoordinate._Bundle->variable_ptr->type < block_type) && (DestCoordinate._IndexWidth == 1) && (CanResizeDestination == yz_false))   {
            CallNumericFunction();     // numeric?  If so, we can process it quickly
            if (err_code == passed)  {
                SaveDPRegister(FElement(&(DestCoordinate._Bundle->variable_ptr->mem.data), DestCoordinate._Offset+1),
                                                DestCoordinate._Bundle->variable_ptr->type);
                if (err_code == out_of_range_err)  {  err_code = passed;  SetWarning(out_of_range_err, PCCodePtr-1);  }   }
            DerefFibril(&(SearchCoordinate._Fibril));
            SearchCoordinate = DestCoordinate;     // these might have been changed in NumericFunction()
            GL_Path = HoldPath;
            ThatCoordinate = HoldThatCoordinate;
            return;    }
        
        if ((DestCoordinate._Bundle->variable_ptr->type == string_type) && (*PCCodePtr == constant_string))    {   // string?
            ThatCoordinate = HoldThatCoordinate;
            LoadString(&DestCoordinate, (void **) &PCCodePtr);
            DerefFibril(&(SearchCoordinate._Fibril));
            SearchCoordinate = DestCoordinate;     // these might have been changed in NumericFunction()
            GL_Path = HoldPath;
            return;     }
        
        
            // OK, we'll have to copy the old-fashioned way.  Get the source variable coordinate and fall through.
        
        CallHobbishFunction();
        SourceCoordinate = SearchCoordinate;
        ThatCoordinate = HoldThatCoordinate;
        if (err_code != passed)  {
            DerefFibril(&(DestCoordinate._Fibril));
            return;      }
        
        if ((DestCoordinate._Bundle->variable_ptr->type == string_type) && (SourceCoordinate._Bundle == NULL))  {
            SetError(type_mismatch_err, DGCommandPtr);
            DerefFibril(&(DestCoordinate._Fibril));
            return;      }
        
        DerefFibril(&(SearchCoordinate._Fibril));
        SearchCoordinate = DestCoordinate;
        GL_Path = HoldPath;
    }
    
    
        // Here's the general procedure for copying between two variables.  First check to see if we need to/can resize the
        // destination ([*] in dest. path), then copy the data.
    
    if (SourceCoordinate._IndexWidth != DestCoordinate._IndexWidth)  {
        if ((CanResizeDestination == yz_true) && (SourceCoordinate._IndexWidth % GL_Path.StemCoordinate._IndexWidth == 0)
                    && (GL_Path.StemMember != NULL))     {
            ResizeBundle(HoldPath.StemCoordinate._Bundle->variable_ptr, GL_Path.StemMember,
                                            SourceCoordinate._IndexWidth/GL_Path.StemCoordinate._IndexWidth);
            if (err_code != passed)  return;
            DestCoordinate._IndexWidth = SourceCoordinate._IndexWidth;    }
        else  {  SetError(mismatched_indices_err, DGCommandPtr);  return;  }    }
    
    if (SourceCoordinate._Bundle == NULL)  {  SetError(not_a_variable_err, SubjectCommand);  return;  }
    CopyBundleData(SourceCoordinate._Bundle, SourceCoordinate._Offset, DestCoordinate._Bundle, DestCoordinate._Offset, DestCoordinate._IndexWidth);
}


// LoadString():  reads a constant string into a variable.

void LoadString(bundle_coordinate *DestCoordinate, void **StringPtr)
{
    linkedlist *NewStringLL;
    LL_retrn linkedLretrn;
    
    if (DestCoordinate->_IndexWidth != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    *(char **) StringPtr += sizeof(Ulong);      // StringPtr initially points to the constant_string command
    
    
        // set the string linked list to the appropriate size/type
        
    NewStringLL = (linkedlist *) FElement(&(DestCoordinate->_Bundle->variable_ptr->mem.data), DestCoordinate->_Offset+1);
    if (NewStringLL->memory != NULL)  FDeleteLinkedList(NewStringLL);
    linkedLretrn = NewLinkedList(NewStringLL, *(Ulong *) *StringPtr, sizeof(char), 0, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    
    
        // copy the data, and push StringPtr past the end of the string (as that is frequently &PCCodePtr)
        
    FSetElements(NewStringLL, 1, NewStringLL->elementnum, *(char **) StringPtr+sizeof(Ulong));
    
    *(char **) StringPtr += sizeof(Ulong) + NewStringLL->elementnum;
    *(char **) StringPtr += Align(NewStringLL->elementnum);
    return;
}


// CheckType() ensures that the type of a member or variable matches, or is compatible with, the given expected type.
// No return value; just throws an error if there is a type mismatch.

void CheckType(linkedlist *TargetLL, unsigned char *SourceType, Boolean IfUpdate)
{
    Ulong counter;
    
    if ((GL_Object.Type == no_type) && (IfUpdate == yz_false))  return;     // the void is compatible with everything
    
    
        // check: non-void type which explicitly doesn't match the expected type
    
    if ((*SourceType != no_type) && (*SourceType != proxy_type) && (GL_Object.Type != *SourceType)
                            && ((GL_Object.Type != no_type) || (IfUpdate == yz_true)))
        {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
    
    
        // make sure we're not copying primitive <-- composite
        
    if (*SourceType != composite_type)  {   // so it's a primitive type or a proxy
        if (TargetLL->elementnum != 0)  SetError(type_mismatch_err, PCCodePtr-1);
        return;    }
        
        
        // composite:  we need all the codes currently in our list to also be present in the expected type
        
    if (GL_Object.CodeList == NULL)   {
        if (TargetLL->elementnum != 0)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
        return;     }
    
    if (TargetLL->elementnum > GL_Object.CodeList->elementnum)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
    
    for (counter = 1; counter <= TargetLL->elementnum; counter++)  {
        if (((code_ref *) FElement(TargetLL, counter))->code_ptr != ((code_ref *) FElement(GL_Object.CodeList, counter))->code_ptr)
            {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
    }
}


// UpdateType() changes the type of a member or variable to match what is given.
// Only partial type-checking:  it is assumed that CheckType() was already run.

void UpdateType(linkedlist *TargetLL, unsigned char *SourceType, bundle_coordinate *CheckIndCoord, Boolean IfCheckMemberInds, Boolean IfMemberCR)
{
    Ulong counter, CopyBottom;
    code_ref *loopCodeRef;
    LL_retrn linkedLretrn;
    
    
        // Make sure that we're changing the whole member and/or variable, then update the type field.
        
    if (*SourceType != GL_Object.Type)  {
        if(IfCheckMemberInds == yz_true)
        if (GL_Path.Indices != GL_Path.StemMember->indices)
            {  SetError(incomplete_member_err, PCCodePtr-1);  return;  }
        if (CheckIndCoord->_IndexWidth != CheckIndCoord->_Bundle->variable_ptr->instances)
            SetError(incomplete_variable_err, PCCodePtr-1);
        *SourceType = GL_Object.Type;
    }
    
    
        // If it's composite we have to also update the codes.  If we're good on that count, leave here.
        
    if (GL_Object.CodeList == NULL)  return;
    if (TargetLL->elementnum == GL_Object.CodeList->elementnum)  return;
    
    
        // Repeat the incomplete-member/variable check, since we likely left the type field alone.
        
    if(IfCheckMemberInds == yz_true)
    if (GL_Path.Indices != GL_Path.StemMember->indices)
        {  SetError(incomplete_member_err, PCCodePtr-1);  return;  }
    if (CheckIndCoord->_IndexWidth != CheckIndCoord->_Bundle->variable_ptr->instances)
        SetError(incomplete_variable_err, PCCodePtr-1);
    
    
        // Copy the new codes over.  (Changing the code list always involves adding codes.)
    
    CopyBottom = TargetLL->elementnum+1;
    
    linkedLretrn = FAddElements(TargetLL, GL_Object.CodeList->elementnum-CopyBottom+1, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }
    FCopyElements(GL_Object.CodeList, CopyBottom, TargetLL, CopyBottom, GL_Object.CodeList->elementnum-CopyBottom+1);
    
    if (IfMemberCR == yz_true)  {     // reference the new code refs if it's a member code list
        for (counter = CopyBottom; counter <= TargetLL->elementnum; counter++)  {
            loopCodeRef = (code_ref *) FElement(TargetLL, counter);
            loopCodeRef->anchor = NULL;     // since anchors will be 'customized' later, don't assign anchors just yet
            Ref_CodeRef(loopCodeRef);
    }   }
}


// *** Navigation commands ***


// _search_member():  looks back up the fibril path (search path) for the member with the given ID (name).
// Throws an error if no such member can be found anywhere along the fibril tree.
// SM_Step() does all the actual work.

void _search_member()  {  Navigate(&SM_Step);  }
void SM_Step()
{
    if (GL_Define.DefineOn == yz_false)  SearchMember(*PCCodePtr, &GL_Path.StemMember, &GL_Path.StemMemberNumber);
    else  SearchMember_DragFibril(*PCCodePtr, &GL_Path.StemMember, &GL_Path.StemMemberNumber);
    
    GL_Path.Offset = 0;
    if (GL_Path.StemMember == NULL)  GL_Path.Indices = 1;      // for defining
    else  GL_Path.Indices = GL_Path.StemMember->indices;
    GL_Path.StemMemberID = *PCCodePtr;
    PCCodePtr++;
}


// _step_to_memberID():  steps from the current variable into the member with the given ID (name).
// If that can't be found then this returns an error.
// Calls SID_Step() to perform the step.

void _step_to_memberID()  {  Navigate(&SID_Step);  }
void SID_Step()
{
    run_retrn retrn;
    
    
        // first get the initial steps along the path
    
    CallHobbishFunction();
    if (err_code != passed)  return;
    
    if (SearchCoordinate._Bundle == NULL)  {  SetError(not_a_variable_err, PCCodePtr-1);  return;  }
    if (SearchCoordinate._Bundle->variable_ptr->type != composite_type)  {  SetError(not_composite_err, PCCodePtr-1);  return;  }
    
    
        // look for the member in this variable
    
    retrn = FindMemberID(SearchCoordinate._Bundle, *PCCodePtr, &GL_Path.StemMember, &GL_Path.StemMemberNumber);
    if (retrn != passed)  SetError(retrn, PCCodePtr);
    
    GL_Path.Offset = 0;
    if (GL_Path.StemMember == NULL)  GL_Path.Indices = 1;
    else  GL_Path.Indices = GL_Path.StemMember->indices;
    GL_Path.StemMemberID = *PCCodePtr;
    PCCodePtr++;
}


// step_to_index():  steps from the current variable into the member having the given index.
// The heavy lifting is done by STI_Step().

void _step_to_index()  {  Navigate(&STI_Step);  }
void STI_Step()
{
    bundle_coordinate HoldSC;
    step_params HoldGLPath;
    Ulong SoughtIndex;
    Boolean HoldGLDefineOn, IfThin = yz_false;
    run_retrn retrn;
    
    
        // get the initial steps along the path
    
    CallHobbishFunction();
    if (err_code != passed)  return;
    
    if (SearchCoordinate._Bundle == NULL)  {  SetError(not_a_variable_err, PCCodePtr-1);  return;  }
    if (SearchCoordinate._Bundle->variable_ptr->type != composite_type)
        {  SetError(not_composite_err, PCCodePtr-1);  return;  }
    
    
        // get the index number to step into
    HoldSC = SearchCoordinate;
    SearchCoordinate._Fibril = NULL;      // to prevent dereferencing problems
    HoldGLPath = GL_Path;
    HoldGLDefineOn = GL_Define.DefineOn;
    GL_Define.DefineOn = yz_false;
    CallNumericFunction();
    GL_Define.DefineOn = HoldGLDefineOn;
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = HoldSC;
    GL_Path = HoldGLPath;
    
    
        // calculate the member number that holds that index
    
    SoughtIndex = (Ulong) DPRegister;
    if ((DPRegister == 0) && (HoldGLDefineOn == yz_true))  {  DPRegister = 1;  IfThin = yz_true;  }
    if (err_code != passed)  err_code = index_argument_err;
    else if (!(DPRegister >= 1))
        {  GL_Path.StemMemberNumber = 0;  SetError(index_argument_err, PCCodePtr-1);  return;  }
    else  {
        retrn = FindMemberIndex(SearchCoordinate._Bundle->variable_ptr, SoughtIndex, &(GL_Path.StemMember),
                                                                                        &(GL_Path.StemMemberNumber), &(GL_Path.Offset));
        if (retrn != passed)  SetError(retrn, PCCodePtr-1); }
    
    
        // Note that err_code != passed if we are defining something new -- fine, since the error code will be cleared later on.
    
    if (err_code != passed)  {
        GL_Path.Indices = (Ulong) DPRegister - GL_Path.Offset + 1;
        GL_Path.StemMemberNumber = SearchCoordinate._Bundle->variable_ptr->mem.members.elementnum+1;
        GL_Path.Offset = 0;         }
    else  GL_Path.Indices = 1;
    GL_Path.StemMemberID = 0;          // this is for when we create new members
    
    if (IfThin == yz_true)  GL_Path.Indices = 0;   // for defining new members; we handle this separately so we don't search for [0]
}


// step_to_index():  steps from the current variable into the member having the given index.
// Calls STIs_Step().

void _step_to_indices()  {  Navigate(&STIs_Step);  }
void STIs_Step()
{
    bundle_coordinate HoldSC;
    step_params HoldGLPath;
    Ulong StartingIndex, SecondOffset, SecondMemberNumber;
    Boolean HoldGLDefineOn;
    run_retrn retrn;
    
    
        // get the first steps in the path
    
    CallHobbishFunction();
    if (err_code != passed)  return;
    
    if (SearchCoordinate._Bundle == NULL)  {  SetError(not_a_variable_err, PCCodePtr-1);  return;  }
    if (SearchCoordinate._Bundle->variable_ptr->type != composite_type)  {  SetError(not_composite_err, PCCodePtr-1);  return;  }
    
    
        // get the first index
    
    HoldSC = SearchCoordinate;
    SearchCoordinate._Fibril = NULL;         // to prevent dereferencing problems
    HoldGLPath = GL_Path;
    HoldGLDefineOn = GL_Define.DefineOn;
    GL_Define.DefineOn = yz_false;
    CallNumericFunction();
    GL_Path = HoldGLPath;
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = HoldSC;
    
    StartingIndex = (Ulong) DPRegister;   // below:  sort of a kludge -- so Navigate() won't think it can fix the error
    if (err_code != passed)  err_code = index_argument_err;
    else if (StartingIndex < 1)  {  GL_Path.StemMemberNumber = 0;  SetError(index_argument_err, PCCodePtr-1);  }
    else  {
        retrn = FindMemberIndex(SearchCoordinate._Bundle->variable_ptr, StartingIndex, &(GL_Path.StemMember),
                                                                                        &(GL_Path.StemMemberNumber), &(GL_Path.Offset));
        if (retrn != passed)  SetError(retrn, PCCodePtr-1);     }
    
    
        // Forgive indices which are above the current top -- this can happen if we're defining (think "array[+N+1] :: ..").
        // We just need the code for the second index to run correctly.
        // Don't worry -- if this is the case then the second argument will throw another error.
    
    if (err_code != passed)  {
        if ((err_code == invalid_index_err) && (GL_Path.Offset == StartingIndex))   {
            err_code = passed;
            GL_Path.Offset = 0;    }
        else  {  GL_Define.DefineOn = HoldGLDefineOn;  return;  }    }
    
    
        // get the second index
    
    HoldSC = SearchCoordinate;
    SearchCoordinate._Fibril = NULL;     // to prevent dereferencing problems
    HoldGLPath = GL_Path;
    CallNumericFunction();
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = HoldSC;
    GL_Path = HoldGLPath;
    GL_Define.DefineOn = HoldGLDefineOn;
    
    if (DPRegister == StartingIndex-1)  {  GL_Path.Indices = 0;  DPRegister = StartingIndex;  }   // legal even if DPR == 0
    else if (!(DPRegister < StartingIndex))  GL_Path.Indices = (Ulong) (DPRegister - StartingIndex + 1);
    
    if (err_code != passed)  err_code = index_argument_err;
    else if (DPRegister < StartingIndex)  {  GL_Path.StemMemberNumber = 0;  SetError(index_argument_err, PCCodePtr-1);  }
    else  {
        retrn = FindMemberIndex(SearchCoordinate._Bundle->variable_ptr, (Ulong) DPRegister, &(GL_Path.StemMember), &SecondMemberNumber, &SecondOffset);
        if (retrn != passed)  SetError(retrn, PCCodePtr-1);         }
    
    GL_Path.StemMemberID = 0;
    if (err_code != passed)  return;
    
    
        // make sure both indices are in the same member
    
    if (SecondMemberNumber != GL_Path.StemMemberNumber)  SetError(step_multiple_members_err, PCCodePtr-1);
}


// step_to_all_indices():  steps from the current variable into all indices.
// Only works if there is only one member.
// This operator is what allows an automatic resize of the last dimension of an array.
// Calls STAI_Step().

void _step_to_all_indices() {  Navigate(&STAI_Step);  }
void STAI_Step()
{
        // get the prior path
    
    GL_Path.Indices = 1;         // for subsequent STAIs
    CallHobbishFunction();
    if (err_code != passed)  return;
    
        // do a bunch of checks
    
    if (SearchCoordinate._Bundle == NULL)  {  SetError(not_a_variable_err, PCCodePtr-1);  return;  }
    if (SearchCoordinate._Bundle->variable_ptr->type != composite_type)  {  SetError(not_composite_err, PCCodePtr-1);  return;  }
    if (SearchCoordinate._Bundle->variable_ptr->mem.members.elementnum == 0)  {
        GL_Path.Indices = 0;                 // for dg st* sm -N st*/sti/etc. ...  (passing arrays)
        GL_Path.OldSearchIndexWidth = SearchCoordinate._IndexWidth;
        GL_Path.StemMemberID = 0;
        GL_Path.StemMemberNumber = 1;
        GL_Path.Offset = 0;
        SetError(no_member_err, PCCodePtr-1);  return;  }
    
    if (SearchCoordinate._Bundle->variable_ptr->mem.members.elementnum != 1)  {  SetError(step_multiple_members_err, PCCodePtr-1);  return;  }
    
        // set all the path params
    
    GL_Path.StemMember = (member *) FElement(&(SearchCoordinate._Bundle->variable_ptr->mem.members), 1);
    GL_Path.StemMemberNumber = 1;
    GL_Path.Offset = 0;
    GL_Path.Indices = GL_Path.StemMember->indices;
    GL_Path.OldSearchIndexWidth = SearchCoordinate._IndexWidth;
}


// Navigate() is the main function for moving up/down a search path.
// This is called by the step/search functions, and in turn relies on their auxiliary functions for the appropriate step info.

void Navigate(void (*StepFunction)())
{
    define_params DefParams;
    object_params ObjParams;
    run_retrn retrn;
    char ErrType;
    
    PCCodePtr++;    // get past the command that called us
    
    if (GL_Define.DefineOn == yz_false)  {    // a short block if we're not defining anything -- just do some error checking
        StepFunction();
        GL_Path.StemCoordinate = SearchCoordinate;         // do this after StepFunction(), since search_member sets SC
        if (err_code == passed)  {
            if (GL_Path.StemMember->member_ptr == NULL)
                {  SetError(void_member_err, PCCodePtr-1);  }
            else if ((SearchCoordinate._IndexWidth > 1) && (GL_Path.Indices != GL_Path.StemMember->indices))
                {  SetError(incomplete_member_err, PCCodePtr-1);  }
            else  {
                StepSearchBundle(GL_Path.StemMember, GL_Path.Offset, GL_Path.Indices);   }      }
        return;    }
    
    
        // We're defining something, so we have to take care of fibrils.
    
    DefParams = GL_Define;         // save the old values of these, since the step functions will overwrite them
    ObjParams = GL_Object;
    SetFlag(GL_Define.Flags, new_target, yz_true);
    GL_Define.FinalStep = yz_false;
    GL_Define.RemoveMode = yz_false;
    SetFlag(GL_Define.Flags, hidden_member, 0);
    InitGLObject();
    
    
        // Do the step
    
    StepFunction();
    GL_Define = DefParams;
    GL_Path.StemCoordinate = SearchCoordinate;  // in case the step fails (keep this before AddVB to make sure AddMem() matches GLP_SC w/ GLP_SM)
    if (DefParams.RemoveMode == yz_true)  return;
    
    AddedAMember = yz_false;
    FailedLastStep = yz_false;
    ErrType = 0;
    
    
        // Check for errors; if OK then add a member if we are allowed.
        // Even though def-general takes care of adding the last member, we can add
        // intermediate steps on the path (as in "a.b.c :: " where 'a' is undefined).
    
    if ((err_code == passed) && (SearchCoordinate._IndexWidth > 1) && (GL_Path.Indices != GL_Path.StemMember->indices))
        {  SetError(incomplete_member_err, PCCodePtr-1);  return;  }
    
    
        // If we have to add a member, we will have tripped an error when we tried the step above.
    
    if (err_code != passed)  {         // also possible to have a mem err, in which case we simply catch it later
        if ((err_code != member_not_found_err) && (err_code != no_member_err)
                        && ((err_code != invalid_index_err) || (GL_Path.StemMemberNumber == 0)))  return;
        if (GetFlag(DefParams.Flags, can_add_members) == yz_false)  return;
        if (SearchCoordinate._IndexWidth != SearchCoordinate._Bundle->variable_ptr->instances)
            {  SetError(incomplete_variable_err, PCCodePtr-1);  return;  }
        
        err_code = passed;
        AddedAMember = yz_true;
        if (DefParams.FinalStep == yz_true)  return;         // let def-general handle the last one
        
            // add a member if we are allowed
        
        retrn = AddMember(SearchCoordinate._Bundle, &(GL_Path.StemMemberNumber), GL_Path.Indices,
                                        SearchCoordinate._IndexWidth, yz_false, &GL_Path.StemMember, GetFlag(DefParams.Flags, hidden_member));
        if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
        
        GL_Path.StemMember->member_ID = GL_Path.StemMemberID;
        GL_Path.StemMember->type = no_type;
        ErrType = 1;      }
    
    
        // This checks for a type mismatch in the case where the member exists but is void.
    
    else if ((GL_Path.StemMember->member_ptr == NULL) && (DefParams.FinalStep == yz_false)
                        && (GetFlag(DefParams.Flags, can_add_members) != yz_false))  {
        ErrType = 2;
        if ((GL_Path.StemMember->type != no_type) &&
                ((GL_Path.StemMember->type != composite_type) || (GL_Path.StemMember->code_list.elementnum != 0)))  {
            SetError(type_mismatch_err, PCCodePtr-1);
            return;
    }   }
    
    
        // If we're qualified to add a (empty composite) variable, do that too.
    
    if (DefParams.FinalStep == yz_true)  return;
    if ((ErrType != 0) && (DefParams.FinalStep == yz_false))    {
        AddVB(&(GL_Path.StemMember->member_ptr), SearchCoordinate._IndexWidth*GL_Path.StemMember->indices, composite_type, yz_true);
        if (err_code != passed)  return;     }
    
    
        // Step into the final variable.
    
    StepSearchBundleAndFibril(GL_Path.StemMember, GL_Path.StemMemberNumber, GL_Path.Offset, GL_Path.Indices);
    if (err_code != passed)  FailedLastStep = yz_true;
}


// AddVB():  stands for Add Variable Bundle.  Creates a variable and a bundle to go with it.

void AddVB(bundle **NewBundle, Ulong BundleSize, Ulong VariableType, Boolean IfCanJam)
{
    variable *NewVariable;
    Ulong DataSize;
    run_retrn retrn;
    
    if ((VariableType == composite_type) || (VariableType == no_type))  {
        retrn = AddCompositeVariable(&NewVariable);
        if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }  }
    else  {
        if (VariableType == block_type)  DataSize = GL_Object.TypeSize;
        else  DataSize = PrimitiveSizes[(Ulong) VariableType];
        
        retrn = AddPrimitiveVariable(&NewVariable, VariableType, DataSize);
        if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }  }
    
    retrn = AddBundle(NewVariable, BundleSize, 0, NewBundle, IfCanJam);
    if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
    RefBundle(*NewBundle);       // OK, since NewBundle is always a member or proxy slot.
}


// RelinkGLSearchMember() sets the member in GL_Path.StemMember to point to a given bundle.
// Used when aliasing or voiding a member.

void RelinkGLSearchMember(define_params *DefParams, object_params *ObjParams, Boolean UnlinkOnly)
{
    bundle **theProxy, *ProxyBundle, *OTB_Copy;
    Ulong ProxyOffset, counter;
    Boolean ifProxy = yz_true, IfCanJam = yz_true;
    run_retrn retrn;
    
    if (GetFlag(DefParams->Flags, unjammable) == yz_true)  IfCanJam = yz_false;
    
    
        // first make sure the aliasing is legal
    
    if (GL_Path.StemMember == 0)  {  SetError(undefined_member_err, PCCodePtr-1);  return;  }
    if ((ObjParams->IfVoid == yz_false) && (UnlinkOnly == yz_false) && (GetFlag(DefParams->Flags, new_target) == yz_false))  {
        if ((ObjParams->Target._IndexWidth != SearchCoordinate._IndexWidth*GL_Path.Indices) && (GetFlag(DefParams->Flags, new_target) == yz_false))
            {  SetError(mismatched_indices_err, PCCodePtr-1);  return;  }
        if (ObjParams->Target._Bundle == NULL)
            {  SetError(illegal_target_err, PCCodePtr-1);  return;  }   }
    
    if (GL_Path.StemMember != NULL)   {      // redundant condition?
        
        if (GL_Path.StemMember->member_ptr == NULL)  ifProxy = yz_false;
        else if (GL_Path.StemMember->member_ptr->variable_ptr->type == proxy_type)  ifProxy = yz_true;
        else ifProxy = yz_false;
        
        
            // for safety, first reference the target bundle just in case the initial unlink-bundle deletes it
        
        OTB_Copy = ObjParams->Target._Bundle;
        if ((ObjParams->IfVoid == yz_false) && (UnlinkOnly == yz_false))
            RefBundle(OTB_Copy);
        
        
            // re-alias the member
        
        if (ifProxy == yz_false)   {
            if (GL_Path.Indices != GL_Path.StemMember->indices)  SetError(incomplete_variable_err, PCCodePtr-1);
            else if (SearchCoordinate._IndexWidth != SearchCoordinate._Bundle->variable_ptr->instances)
                    SetError(incomplete_variable_err, PCCodePtr-1);
            else if ((ObjParams->IfVoid == yz_true) || (GL_Path.StemMember->member_ptr != ObjParams->Target._Bundle))  {
                
                if (GL_Path.StemMember->member_ptr != NULL)    {     // first unlink the member if it is pointing to something
                    DerefBundle(&(GL_Path.StemMember->member_ptr));    }
                
                if ((ObjParams->IfVoid == yz_false) && (UnlinkOnly == yz_false))  {   // if we're not voiding it, then relink it
                    if (*(ObjParams->Target._Bundle->variable_ptr->references) == 0)  {  SetError(target_deleted_err, PCCodePtr-1);  return;  }
                    retrn = AddBundle(ObjParams->Target._Bundle->variable_ptr, ObjParams->Target._IndexWidth, ObjParams->Target._Offset,
                                                                                                &(GL_Path.StemMember->member_ptr), IfCanJam);
                    if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
                    RefBundle(GL_Path.StemMember->member_ptr);
        }   }   }
        
        else   {   // re-alias a proxy
            for (counter = 0; counter < GL_Path.Indices; counter++)   {
                ProxyBundle = SearchCoordinate._Bundle;
                ProxyOffset = SearchCoordinate._Offset;
                StepToMember(&ProxyBundle, &ProxyOffset, GL_Path.StemMember, GL_Path.Offset+counter);
                theProxy = (bundle **) FElement(&(ProxyBundle->variable_ptr->mem.data), ProxyOffset+1);
                if ((ObjParams->IfVoid == yz_true) || (*theProxy != ObjParams->Target._Bundle))  {
                    if (*theProxy != NULL)  {
                        DerefBundle(theProxy);         }
                    if ((ObjParams->IfVoid == yz_false) && (UnlinkOnly == yz_false))   {
                        if (*(ObjParams->Target._Bundle->variable_ptr->references) == 0)  {  SetError(target_deleted_err, PCCodePtr-1);  return;  }
                        retrn = AddBundle(ObjParams->Target._Bundle->variable_ptr, 1, ObjParams->Target._Offset, theProxy, IfCanJam);
                        if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
                        RefBundle(*theProxy);
        }   }   }   }
        
        if ((ObjParams->IfVoid == yz_false) && (UnlinkOnly == yz_false))
            DerefBundle(&OTB_Copy);    // undo the extra reference of the target bundle
    }
}


// _resize() and _resize_start() are both the [^..] operator.  These scale an array by setting the width of its lone member.
// _resize_start() is used when the [^..] 'begins' a sentence (as in, it's the lowest-OoO operation).
// _resize() is used when the expression is embedded in a larger expression:  x = array[^1].
// DoResize() does all the actual work.

void _resize()  {  Navigate(&DoResize);  }
void _resize_start()  {  PCCodePtr++;  DoResize();  }

void DoResize()
{
    bundle_coordinate HoldSC;
    step_params HoldGLPath;
    Boolean HoldGLDefineOn = GL_Define.DefineOn;
   
   
        // get the beginning of the path
    
    GL_Path.StemMember = NULL;
    GL_Define.FinalStep = yz_false;
    
    CallHobbishFunction();
    if (err_code != passed)  return;
    
    if (SearchCoordinate._Bundle == NULL)  {  SetError(not_a_variable_err, PCCodePtr-1);  return;  }
    if (SearchCoordinate._Bundle->variable_ptr->type != composite_type)  {  SetError(not_composite_err, PCCodePtr-1);  return;  }
    
    
        // read in the new top index
    
    HoldSC = SearchCoordinate;
    SearchCoordinate._Fibril = NULL;      // to prevent dereferencing problems
    HoldGLPath = GL_Path;
    GL_Define.DefineOn = yz_false;
    
    CallNumericFunction();
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = HoldSC;
    GL_Path = HoldGLPath;
    GL_Define.DefineOn = HoldGLDefineOn;
    
    if (err_code != passed)  return;
    if (!((DPRegister >= 0) && (DPRegister <= ULONG_MAX)))  {  SetError(index_argument_err, PCCodePtr-1);  return;  }
    if (SearchCoordinate._Bundle->variable_ptr->mem.members.elementnum > 1)  {  SetError(step_multiple_members_err, PCCodePtr-1);  return;  }
    
    GL_Path.Indices = (Ulong) DPRegister;
    GL_Path.StemMemberNumber = 1;
    GL_Path.Offset = 0;
    GL_Path.StemMemberID = 0;
    
    
        // assuming everything's OK, resize the member
    
    if (SearchCoordinate._Bundle->variable_ptr->mem.members.elementnum == 0)  {  SetError(no_member_err, PCCodePtr-1);  }
    else    {
        GL_Path.StemMember = (member *) FElement(&(SearchCoordinate._Bundle->variable_ptr->mem.members), 1);
        if (SearchCoordinate._IndexWidth != SearchCoordinate._Bundle->variable_ptr->instances)
            {  SetError(incomplete_variable_err, PCCodePtr-1);  return;  }
        
        ResizeBundle(SearchCoordinate._Bundle->variable_ptr, GL_Path.StemMember, (Ulong) DPRegister);    // member, member_number, offset
        if (err_code != passed)  return;       }
}


// The next five routines correspond to [+ (1, 3), +[ (2, 4), and delete (5).  They insert (1-4) or delete (5) indices from members/variables.
// Routines 1-2 are used in the middle of expressions; 3-4 when the insertion operation constitutes a complete sentence.
// MasterInsert() and Prune() do the dirty work.

void _insert_indices_up()  {  MasterInsert(yz_true, yz_true);  }
void _insert_indices_down()  {  MasterInsert(yz_false, yz_true);  }
void _insert_indices_up_start()  {  MasterInsert(yz_true, yz_false);  }
void _insert_indices_down_start()  {  MasterInsert(yz_false, yz_false);  }
void _delete_indices()  {  Prune(pn_indices);  }


// MasterInsert() adds indices to a member and the variable it points to.

void MasterInsert(Boolean IfUp, Boolean if_final_step)
{
    bundle_coordinate HoldSC;
    step_params HoldGLPath;
    Ulong StartingIndex;
    Boolean HoldGLDefineOn, HoldFinalStep = GL_Define.FinalStep;
    run_retrn retrn;
    
    PCCodePtr++;
    
    
        // get the first steps on the search path
    
    GL_Path.StemMember = NULL;
    GL_Define.FinalStep = yz_false;
    
    CallHobbishFunction();
    if (err_code != passed)  return;
    if (SearchCoordinate._Bundle->variable_ptr->type != composite_type)  {  SetError(not_composite_err, PCCodePtr-1);  return;  }
    if (GL_Path.StemMember == NULL)  {  SetError(undefined_member_err, PCCodePtr-1);  return;  }
    
    HoldSC = SearchCoordinate;
    SearchCoordinate._Fibril = NULL;       // to prevent dereferencing problems
    HoldGLPath = GL_Path;
    HoldGLDefineOn = GL_Define.DefineOn;
    GL_Define.DefineOn = yz_false;
    
    
        // find the first index to insert
    
    CallNumericFunction();
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = HoldSC;
    GL_Path = HoldGLPath;
    
    if (err_code != passed)  return;
    if (!(DPRegister >= 1))  {  SetError(index_argument_err, PCCodePtr-1);  return;  }
    StartingIndex = (Ulong) DPRegister;
    
    if ((IfUp == yz_true) || (StartingIndex == 1))  {
        retrn = FindMemberIndex(SearchCoordinate._Bundle->variable_ptr, (Ulong) DPRegister, &(GL_Path.StemMember),
                                                                                        &(GL_Path.StemMemberNumber), &(GL_Path.Offset));
        if (retrn != passed)  SetError(retrn, PCCodePtr-1);     }
    
        // 2nd condition on next line:  a([+top(a)+1]) is legal
    if (((IfUp == yz_false) && (DPRegister > 1)) || ((err_code == invalid_index_err) && (StartingIndex == GL_Path.Offset)))   {
        if (StartingIndex == 1)  {      // if all indices have been deleted, fix the error as long as there is a member to add to
            if (SearchCoordinate._Bundle->variable_ptr->mem.members.elementnum > 0)   {
                GL_Path.StemMember = (member *) FElement(&(SearchCoordinate._Bundle->variable_ptr->mem.members), 1);
                GL_Path.StemMemberNumber = 1;
                GL_Path.Offset = 0;
                err_code = passed;     }
            else  {  SetError(index_argument_err, PCCodePtr-1);  return;  }   }
        else  {
            err_code = passed;
            retrn = FindMemberIndex(SearchCoordinate._Bundle->variable_ptr, (Ulong) DPRegister-1, &(GL_Path.StemMember),
                                                                                            &(GL_Path.StemMemberNumber), &(GL_Path.Offset));
            if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
            GL_Path.Offset++;       // by default, append to the index just below
    }   }
    
    if ((err_code == passed) && (GL_Path.StemMember->member_ptr == NULL))  SetError(void_member_err, PCCodePtr-1);
    if (err_code != passed)  return;
    
    if (SearchCoordinate._IndexWidth != SearchCoordinate._Bundle->variable_ptr->instances)
        {  SetError(incomplete_variable_err, PCCodePtr-1);  return;  }
    
    
        // now read in the second index (there are two for these operators)
    
    HoldGLPath = GL_Path;           // GL_Path was changed since the first NumericFunction()
    SearchCoordinate._Fibril = NULL;   // this is to prevent dereferencing problems
    CallNumericFunction();
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = HoldSC;
    GL_Path = HoldGLPath;
    GL_Define.DefineOn = HoldGLDefineOn;
    
    if (!(DPRegister >= StartingIndex))  {  SetError(index_argument_err, PCCodePtr-1);  return;  }
    GL_Path.Indices = (Ulong) (DPRegister - StartingIndex + 1);
    
    
        // the following takes care of the resize of the member and the variable
    
    ResizeIndices(SearchCoordinate._Bundle->variable_ptr, GL_Path.StemMember, GL_Path.StemMemberNumber, GL_Path.Offset, GL_Path.Indices);
    if (GL_Define.DefineOn == yz_true)  AddedAMember = yz_false;
    if (err_code != passed)  return;
    
    
        // go into our newly-sized variable
    
    if ((if_final_step == yz_false) || ((HoldGLDefineOn == yz_true) && (HoldFinalStep == yz_true)))  return;
    if (SearchCoordinate._IndexWidth == 1)  {
        if (GL_Define.DefineOn == yz_false)  StepSearchBundle(GL_Path.StemMember, GL_Path.Offset, GL_Path.Indices);
        else  StepSearchBundleAndFibril(GL_Path.StemMember, GL_Path.StemMemberNumber, GL_Path.Offset, GL_Path.Indices);     }
    else  {
        if (GL_Define.DefineOn == yz_false)  StepSearchBundle(GL_Path.StemMember, 0, GL_Path.StemMember->indices);
        else  StepSearchBundleAndFibril(GL_Path.StemMember, GL_Path.StemMemberNumber, 0, GL_Path.StemMember->indices);      }
}


// Prune() removes either members or indices from a member and its target variable.
// Used both by _remove() and _delete().

void Prune(Boolean OffendingBranch)
{
    PCCodePtr++;
    
    
        // Get the path to the condemned.
        // Unlike MasterInsert(), we don't read the indices separately -- we just step into whatever we want deleted.
    
    GL_Define.DefineOn = yz_true;
    SetFlag(GL_Define.Flags, can_add_members, yz_false);
    GL_Define.RemoveMode = yz_true;    
    CallHobbishFunction();
    
    if (SearchCoordinate._Bundle == NULL)  {      // some basic checks
        if (err_code == void_proxy_err)
            SearchCoordinate._Bundle = SearchCoordinate._Fibril->stem->bundle_ptr;
        else if (err_code != passed)  return;     }
    else if ((err_code == passed) && (SearchCoordinate._IndexWidth != SearchCoordinate._Bundle->variable_ptr->instances))
        {  SetError(incomplete_variable_err, PCCodePtr-1);  }
    if (err_code != passed)  return;
    
    if (GL_Path.StemMember == NULL)  {  SetError(undefined_member_err, PCCodePtr-1);  return;  }
    
    
        // case 1:  remove a member
    
    if (OffendingBranch == pn_member)   {    // GL_Path.RemoveMode prevents the last step, so SC points to the stem bundle.
        if (GL_Path.Indices != GL_Path.StemMember->indices)  {  SetError(incomplete_member_err, PCCodePtr-1);  return;  }
        RemoveMember(SearchCoordinate._Bundle->variable_ptr, GL_Path.StemMemberNumber);    }
    
    
        // case 2:  delete indices
    
    else  {
        if (GL_Path.StemMember->member_ptr == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  }
        ResizeIndices(SearchCoordinate._Bundle->variable_ptr, GL_Path.StemMember, GL_Path.StemMemberNumber,
                      GL_Path.Offset, -GL_Path.Indices);  }
    
    GL_Path.StemMember = NULL;          // clean our knives & go
    SearchCoordinate._Bundle = NULL;
    GL_Define.DefineOn = yz_false;
}


// ResizeBundle() adds or subtracts indices to get NewTop total indices in a variable.
// Assumes there is one and only one member.  The top is an index, not an offset.

void ResizeBundle(variable *StemVariable, member *StemMember, Ulong NewTop)
{
    if (NewTop > StemMember->indices)  ResizeIndices(StemVariable, StemMember, 1, StemMember->indices, NewTop-StemMember->indices);
    else if (NewTop < StemMember->indices)  ResizeIndices(StemVariable, StemMember, 1, NewTop, NewTop-StemMember->indices);
}


// ResizeIndices() adds NewIndices indices at the given location in the given member.
// NewIndices can be either positive (insertion of new indecs) or negative (deletion of existing indices starting at InsertionOffset).
// Calls AddMemory() once for each index spanned by SearchCoordinate._IndexWidth.  GL_Path points to a member in SearchCoordinate's bundle.

void ResizeIndices(variable *StemVariable, member *StemMember, Ulong StemMemberNumber, Ulong InsertionOffset, signed long NewIndices)
{
    member *HoldGLMember;
    signed long counter, CounterTop, count_to_five;
    Ulong c2, CBO_Indices;
    run_retrn retrn;
    
    
        // First check to see if there are overlaps with existing members before resizing, as this is disallowed.
        // We don't want to have a partial resize if there's an overlap error midway through.
        
    if (StemMember->member_ptr != NULL)    {
        Ulong last_index_top = 0;		// next line:  almost correct, unless two members w/ different dimensions share a variable
        if (StemMember->indices != 0)  last_index_top = StemMember->member_ptr->variable_ptr->instances / StemMember->indices;
        
        for (c2 = 0; c2 < last_index_top; c2++)  {    // in general our PCCoordinate straddles multiple indices
            if (NewIndices > 0)  CBO_Indices = NewIndices;
            else  CBO_Indices = -NewIndices;
            
                // Checking involves a five-step procedure, calling CheckBundleOverlap each time.
                // Just run it five times and pass the counter.  Need to UnflagBundles() after each CBO() call.
                
            for (count_to_five = 1; count_to_five <= 5; count_to_five++)    {
                retrn = CheckBundleOverlap(StemMember->member_ptr, c2*StemMember->indices + InsertionOffset, CBO_Indices, count_to_five);
                UnflagBundles(StemMember->member_ptr, busy_overlap_flag);
                if (retrn != passed)  break;         }
            if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  };    }
        
        HoldGLMember = GL_Path.StemMember;   // delete GLPath.StemMember for safety (AddMem() won't know it stores correct value after resize)
        
        
            // Start from the highest index and work backwards, so we don't delete from under our own feet
            // (element numbers above the deletion point will change when we delete).
            
        if (StemMember->indices > 0)  CounterTop = (StemMember->member_ptr->instances/StemMember->indices)-1;
        else  CounterTop = StemVariable->instances-1;
        for (counter = CounterTop; counter >= 0; counter--)  {                            // StemMember should never be 'unjammed'?
            retrn = AddMemory(StemMember->member_ptr, counter*StemMember->indices + InsertionOffset, NewIndices);
            UnflagVariables(StemMember->member_ptr->variable_ptr, busy_add_flag);
            if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }         }
        
        GL_Path.StemMember = HoldGLMember;         }
    
    
        // lastly, update the member indices
    
    StemMember->indices += NewIndices;
    if ((StemMemberNumber < StemVariable->first_array) && (StemMember->indices != 1))
        StemVariable->first_array = StemMemberNumber;
}



// _this_var() returns 'this'.

void _this_var()
{
    PCCodePtr++;
    RefFibril(PCCoordinate._Fibril);
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = PCCoordinate;
    GL_Path.StemMember = NULL;
}


// _that_var() returns 'that' -- the variable to the left of the '=' (if there is one; otherwise returns *).

void _that_var()
{
    PCCodePtr++;
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = ThatCoordinate;
    GL_Path.StemMember = NULL;
}


// _no_var() returns the void (* or nothing).

void _no_var()
{
    PCCodePtr++;
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate._Bundle = NULL;
    GL_Path.StemMember = NULL;
    GL_Object.IfVoid = yz_true;
    GL_Object.CodeList = NULL;
}



// _if_eq() tests an expression for equality, and stores the result in GL_Conditional.

void _if_eq()
{
    double FirstArgument;
    bundle_coordinate DestCoordinate;
    char *FirstString = NULL, *SecondString = NULL;
    Ulong FirstStringLength, SecondStringLength, MC_retrn;
    
    PCCodePtr++;
    
    
        // get the first argument
    
    Compare_ReadArgument(&FirstString, &FirstStringLength);
    if (err_code != passed)  return;
    DestCoordinate = SearchCoordinate;


        // If the first argument was a string, get the second argument and expect another string.
    
    if (FirstString != NULL)   {
        Compare_ReadArgument(&SecondString, &SecondStringLength);
        if (err_code != passed)  return;
        if (SecondString == NULL)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
        if (FirstStringLength != SecondStringLength)  {  GL_Conditional = yz_false;  return;  }
        
        MC_retrn = memcmp((void *) FirstString, (void *) SecondString, (size_t) FirstStringLength);     // do the comparison
        if (MC_retrn == 0)  GL_Conditional = yz_true;
        else  GL_Conditional = yz_false;
        return;    }

    else if (DestCoordinate._Bundle != NULL)   {


            // If the first argument was a general (composite) object, get arg 2 and do a general comparison via CompareBundleData().
        
        if ((DestCoordinate._Bundle->variable_ptr->type > double_type) || (DestCoordinate._IndexWidth != 1))    {
            CallHobbishFunction();
            if (err_code != passed)  return;
            
            if (SearchCoordinate._IndexWidth != DestCoordinate._IndexWidth)  {  SetError(mismatched_indices_err, PCCodePtr-1);  return;  }
            if (SearchCoordinate._Bundle == NULL)  {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
            GL_Conditional = yz_true;         // CBD only sets it false
            CompareBundleData(SearchCoordinate._Bundle, SearchCoordinate._Offset, DestCoordinate._Bundle, DestCoordinate._Offset,
                                                        DestCoordinate._IndexWidth);
            return;         }
        
        
            // The first argument was numeric, so get the number and fall through to process the second argument.
        
        LoadDPRegister(FElement(&(DestCoordinate._Bundle->variable_ptr->mem.data), DestCoordinate._Offset+1), DestCoordinate._Bundle->variable_ptr->type);
    }

    FirstArgument = DPRegister;

    CallNumericFunction();      // get 2nd argument -- must be numeric

    if (FirstArgument == DPRegister)  GL_Conditional = yz_true;
    else  GL_Conditional = yz_false;
}


// Compare_ReadArgument() reads a bytecode argument and, if it's a string, sets StringPtr and StringSize.

void Compare_ReadArgument(char **StringPtr, Ulong *StringSize)
{
    linkedlist *theString;
    LL_retrn linkedLretrn;
    
    *StringPtr = NULL;
    SearchCoordinate._Bundle = NULL;
    
    if (*PCCodePtr == constant_string)  {    // look ahead so we can handle the string case manually
        CallHobbishFunction();
        *StringPtr = GL_Object.StringPtr;
        *StringSize = GL_Object.TypeSize;   }
        
    else if ((NumericJumpTable[*PCCodePtr] != &_illegal) && (NumericJumpTable[*PCCodePtr] != &Num_Adapter))  {   // do numbers manually
        CallHobbishFunction();
        if (err_code != passed)  return;  }

    else  {        // it's something else -- make sure it returns a variable
        CallHobbishFunction();
        if (err_code != passed)  return;
        if (SearchCoordinate._Bundle == NULL)  {  SetError(not_a_variable_err, PCCodePtr-1);  return;  }   }
    
    
        // if it's a string in a variable, set StringPtr and StringSize here
    
    if (SearchCoordinate._Bundle != NULL)  {
    if ((SearchCoordinate._Bundle->variable_ptr->type == string_type) && (SearchCoordinate._IndexWidth == 1))  {

        theString = (linkedlist *) FElement(&(SearchCoordinate._Bundle->variable_ptr->mem.members), SearchCoordinate._Offset+1);
        if (theString->memory == NULL)  {  SetError(undefined_string_err, PCCodePtr-1);  return;  }
        linkedLretrn = FDefragmentLinkedList(theString);
        if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }

        *StringSize = theString->elementnum;
        if (*StringSize == 0)  *StringPtr = ((char *) theString->memory) + sublistHeaderSize;
        else  *StringPtr = (char *) FElement(theString, 1);    }   }
}

void _if_ne()
{
    _if_eq();
    GL_Conditional = 1-GL_Conditional;
}


void _if_eq_at()
{
    bundle_coordinate FirstCoordinate;
    
    PCCodePtr++;
    
    
    // get the first argument
    
    iea_one_arg();
    if (err_code != passed)  return;
    FirstCoordinate = SearchCoordinate;
    
    iea_one_arg();
    if (err_code != passed)  return;
    
    if (FirstCoordinate._Bundle == NULL)    {
        if (SearchCoordinate._Bundle == NULL)  GL_Conditional = yz_true;
        else  GL_Conditional = yz_false;         }
    else if (SearchCoordinate._Bundle == NULL)  {
        GL_Conditional = yz_false;         }
    else if ((FirstCoordinate._Bundle->variable_ptr == SearchCoordinate._Bundle->variable_ptr)
                && (FirstCoordinate._IndexWidth == SearchCoordinate._IndexWidth)
                && (FirstCoordinate._Offset == SearchCoordinate._Offset))
        GL_Conditional = yz_true;
    else  GL_Conditional = yz_false;
}

void iea_one_arg()
{
    Slong *ArgBeginningPtr, *ErrPtr;
    Ulong hold_err_code;

    ArgBeginningPtr = PCCodePtr;
    CallHobbishFunction();
    if ((err_code == void_member_err) || (err_code == void_proxy_err))  {
        
        hold_err_code = err_code;
        err_code = passed;
        ErrPtr = PCCodePtr;
        PCCodePtr = ArgBeginningPtr;
        CallSkipFunction();
        
        if (ErrPtr == PCCodePtr)
            SearchCoordinate._Bundle = NULL;
        else
            err_code = hold_err_code;           }
}

void _if_ne_at()
{
    _if_eq_at();
    GL_Conditional = 1-GL_Conditional;
}


// The next three functions return true or false based on the truth values of their two arguments.
// They make use of the following three functions and LogicTemplate().

void _if_and()  {  LogicTemplate(&Logic_And);  }
void _if_or()  {  LogicTemplate(&Logic_Or);  }
void _if_xor()  {  LogicTemplate(&Logic_Eor);  }

void Logic_And(Boolean FirstArgument)  {  if ((FirstArgument) && (GL_Conditional))  GL_Conditional = yz_true;  else  GL_Conditional = yz_false;  }
void Logic_Or(Boolean FirstArgument)  {  if ((FirstArgument) || (GL_Conditional))  GL_Conditional = yz_true;  else  GL_Conditional = yz_false;  }
void Logic_Eor(Boolean FirstArgument)  {  if (FirstArgument != GL_Conditional)  GL_Conditional = yz_true;  else  GL_Conditional = yz_false;  }

// LogicTemplate() evaluates two logical arguments and invokes Logic_Operation() to turn these into one logical value.
void LogicTemplate(void(*Logic_Operation)(Boolean))
{
    Boolean FirstArgument;

    PCCodePtr++;
    CallLogicalFunction();             // get first arg
    if (err_code != passed)  return;
    FirstArgument = GL_Conditional;
    
    CallLogicalFunction();             // get second arg
    Logic_Operation(FirstArgument);    // and evaluate the conditional
}


// _if_not() returns the opposite of _if_eq().

void _if_not()
{
    PCCodePtr++;
    CallLogicalFunction();
    GL_Conditional = 1-GL_Conditional;
}


// The next 6 routines compute various mathematical functions of two numeric arguments, returning a numerical value.
// The following 4 routines compute various logical functions of two numeric arguments, returning a logical true or false.
// These use the respective 10 subsequent routines:  Math_..() and MathTemplate().

void _addf()  {  MathTemplate(Math_Add);  }
void _subf()  {  MathTemplate(Math_Sub);  }
void _mulf()  {  MathTemplate(Math_Mul);  }
void _divf()  {  MathTemplate(Math_Div);  }
void _powerf()  {  MathTemplate(Math_Pow);  }
void _modi()  {  MathTemplate(Math_Mod);  }

void _if_lt()  {  MathTemplate(Math_LT);  }
void _if_le()  {  MathTemplate(Math_LE);  }
void _if_gt()  {  MathTemplate(Math_GT);  }
void _if_ge()  {  MathTemplate(Math_GE);  }

void Math_Add(double FirstArgument)  {  DPRegister += FirstArgument;  }
void Math_Sub(double FirstArgument)  {  DPRegister = FirstArgument-DPRegister;  }
void Math_Mul(double FirstArgument)  {  DPRegister *= FirstArgument;  }
void Math_Div(double FirstArgument)  {
            if (DPRegister == 0)  {  SetWarning(divide_by_zero_warning, PCCodePtr-1);  }
            DPRegister = FirstArgument/DPRegister;  }
void Math_Pow(double FirstArgument)  {  DPRegister = pow(FirstArgument, DPRegister);  }
void Math_Mod(double FirstArgument)  {  
            if (DPRegister == 0)  DPRegister = NAN;
            else  DPRegister = (double) ((Ulong) FirstArgument % (Ulong) DPRegister);  }

void Math_LT(double FirstArgument)  {  if (FirstArgument < DPRegister)  GL_Conditional = yz_true;  else  GL_Conditional = yz_false;  }
void Math_LE(double FirstArgument)  {  if (FirstArgument <= DPRegister)  GL_Conditional = yz_true;  else  GL_Conditional = yz_false;  }
void Math_GT(double FirstArgument)  {  if (FirstArgument > DPRegister)  GL_Conditional = yz_true;  else  GL_Conditional = yz_false;  }
void Math_GE(double FirstArgument)  {  if (FirstArgument >= DPRegister)  GL_Conditional = yz_true;  else  GL_Conditional = yz_false;  }


// MathTemplate() evaluates two numeric arguments, and then calls some function of these arguments.

void MathTemplate(void(*Math_Operation)(double))
{
    double FirstArgument;

    PCCodePtr++;
    CallNumericFunction();           // arg 1
    if (err_code != passed)  return;
    FirstArgument = DPRegister;
    
    CallNumericFunction();           // arg 2
    SearchCoordinate._Bundle = NULL;
    GL_Path.StemMember = NULL;
    Math_Operation(FirstArgument);   // compute the function (return value isn't specified because result is stored in global variables)
}



// _constant_Slong() returns the given signed long constant embedded in the bytecode.

void _constant_Slong()
{
    PCCodePtr++;
    SearchCoordinate._Bundle = NULL;
    GL_Path.StemMember = NULL;
    GL_Object.Type = Slong_type;
    
    DPRegister = (double) *PCCodePtr;
    
    PCCodePtr++;
}


// _constant_float() returns the given double-precision floating point number embedded in the bytecode.

void _constant_float()
{
    PCCodePtr++;
    SearchCoordinate._Bundle = NULL;
    GL_Path.StemMember = NULL;
    GL_Object.Type = double_type;
    
    DPRegister = *(double *) PCCodePtr;
    
    PCCodePtr += sizeof(double)/sizeof(Ulong);
}


// _code_block() stores a pointer to the code that is encountered in the global GL_Object variable;
// then skips the program counter to the end of the code.

void _code_block()
{
    PCCodePtr++;
    OneCodeRef->PLL_index = PCCodeRef.PLL_index;
    OneCodeRef->references = PCCodeRef.references;
    OneCodeRef->code_ptr = PCCodePtr;
    OneCodeRef->anchor = NULL;          // will be 'customized' later

    GL_Object.CodeList = &OneCodeLL;
    GL_Object.Type = composite_type;
    SearchCoordinate._Bundle = NULL;
    GL_Path.StemMember = NULL;
    
    RunSkipMode(0);     // get over the code without executing any of it
    PCCodePtr++;
}


// _constant_string() loads a pointer to an embedded string constant into GL_Object

void _constant_string()
{
    PCCodePtr++;
    GL_Object.Type = string_type;

    GL_Object.TypeSize = (Ulong) *PCCodePtr;
    PCCodePtr++;

    SearchCoordinate._Bundle = NULL;
    GL_Path.StemMember = NULL;
    
    GL_Object.StringPtr = (char *) PCCodePtr;
    PCCodePtr += (Ulong) ceil(1.*GL_Object.TypeSize/sizeof(Ulong));
}



// The next 10 routines define various data types.

void _ubyte_cmd()  {  PCCodePtr++;  SetPrimitiveArgType(Ubyte_type);  }
void _sbyte_cmd()  {  PCCodePtr++;  SetPrimitiveArgType(Sbyte_type);  }
void _ushort_cmd()  {  PCCodePtr++;  SetPrimitiveArgType(Ushort_type);  }
void _sshort_cmd()  {  PCCodePtr++;  SetPrimitiveArgType(Sshort_type);  }
void _ulong_cmd()  {  PCCodePtr++;  SetPrimitiveArgType(Ulong_type);  }
void _slong_cmd()  {  PCCodePtr++;  SetPrimitiveArgType(Slong_type);  }
void _float_cmd()  {  PCCodePtr++;  SetPrimitiveArgType(single_type);  }
void _double_cmd()  {  PCCodePtr++;  SetPrimitiveArgType(double_type);  }
void _string_cmd()  {  PCCodePtr++;  SetPrimitiveArgType(string_type);  }

void _block_cmd()  {     // this one is slightly more involved since we have to read the number of bytes in this block
    PCCodePtr++;
    CallNumericFunction();      // get bytes #; call before SPAT() so we don't overwrite our type

    SetPrimitiveArgType(block_type);      // which steps the PC
    if (!( (DPRegister >= 0) && (DPRegister <= ULONG_MAX) ))  SetError(invalid_block_size_err, PCCodePtr-1);
    else GL_Object.TypeSize = (Ulong) DPRegister;
    GL_Object.StringPtr = NULL;

    SearchCoordinate._Bundle = NULL;
    GL_Path.StemMember = NULL;
}

// SetPrimitiveArgType() sets the global variables to indicate that a type, but not a variable, was specified.
void SetPrimitiveArgType(Ulong theType)
{
    GL_Object.Type = theType;
    GL_Object.CodeList = NULL;
    SearchCoordinate._Bundle = NULL;
    GL_Path.StemMember = NULL;
}


// A 'code' marker signifies to the interpreter that a code boundary was encountered.
// Not sure under what cases the _code_marker() command is actually run..

void _code_marker()  {  PCCodePtr++;  SearchCoordinate._Bundle = NULL;  GL_Path.StemMember = NULL;  GL_Object.CodeList = NULL;  }


// _illegal() is used in jump tables for forbidden commands -- it just throws an error.

void _illegal()  {  SetError(illegal_command_err, PCCodePtr);  }



// InitGLObject() resets the GL_Object type and clears GL_Object.CodeLL, in preparation for applying GL_Object to another object.

void InitGLObject()
{
    DeleteObjectCodeLL(&(GL_Object.CodeLL));
    
    GL_Object.Type = no_type;
    GL_Object.IfVoid = yz_false;
    GL_Object.CodeList = NULL;
}


// RefGLObjectCodeLL() references all codes in GL_Object.CodeLL.

void RefGLObjectCodeLL()
{
    Ulong counter;
    
    if (GL_Object.CodeLL.memory == NULL)  return;
    for (counter = 1; counter <= GL_Object.CodeLL.elementnum; counter++)    {
        Ref_CodeRef((code_ref *) FElement(&(GL_Object.CodeLL), counter));   }
}


// DeleteObjectCodeLL() dereferences all code refs in theCodeLL, then deletes the linked list.

void DeleteObjectCodeLL(linkedlist *theCodeLL)
{
    Ulong counter;
    
    if (theCodeLL->memory == NULL)  return;
    for (counter = 1; counter <= theCodeLL->elementnum; counter++)   {
        Deref_CodeRef((code_ref *) FElement(theCodeLL, counter));    }
    FDeleteLinkedList(theCodeLL);
}




// Num_Adapter() is used to read a number from a constant or variable.

void Num_Adapter()
{
    variable *theVariable;

    SearchCoordinate._Bundle = NULL;
    CallHobbishFunction();        // read in an argument
    if (err_code != passed)  return;

    if (SearchCoordinate._Bundle == NULL)       // Can happen w/ append.  Known bugs:  a = proxy:proxy, a = 2:3 are legal.
        {  SetError(not_a_variable_err, PCCodePtr-1);  return;  }
    if (SearchCoordinate._IndexWidth != 1)  {  SetError(multiple_indices_not_allowed_err, PCCodePtr-1);  return;  }
    theVariable = SearchCoordinate._Bundle->variable_ptr;
    if (theVariable->type > double_type)  {  SetError(non_numeric_err, PCCodePtr-1);  return;  }    // must be numeric

    LoadDPRegister(FElement(&(theVariable->mem.data), SearchCoordinate._Offset+1), theVariable->type);
}


// VT_Adapter() gives a quick way of adapting the standard Hobbish routines when a VarType
// or code argument is expected (i.e. a SearchCoordinate and GL_Object.CodeList).
// It returns both a variable and a code (the function code) when the user provides only the variable.

void VT_Adapter()
{
    DeleteObjectCodeLL(&(GL_Object.CodeLL));

    SearchCoordinate._Bundle = NULL;
    GL_Object.CodeList = NULL;
    GL_Object.Type = no_type;
    
    CallHobbishFunction();      // read the argument
    if (err_code != passed)  return;

    GL_Define.DefineOn = yz_false;
    if (SearchCoordinate._Bundle != NULL)  {     // set the global type/code variables
        GL_Object.CodeList = &(SearchCoordinate._Bundle->variable_ptr->code_list);
        GL_Object.Type = SearchCoordinate._Bundle->variable_ptr->type;    }     // else do nothing -- it might have been a string, etc.
}


// Reads a numeric argument from a math function and sets the global type variable
// to double or slong depending on the outer operator of the argument.

void VT_NumericAdapter()
{
    Slong StoreFirstCommand = *PCCodePtr;
    
    CallHobbishFunction();
    if (err_code != passed)  return;

    SearchCoordinate._Bundle = NULL;
    GL_Object.CodeList = NULL;
    GL_Path.StemMember = NULL;

    if (StoreFirstCommand == modi)  GL_Object.Type = Slong_type;
    else  GL_Object.Type = double_type;
}


// Next 16 routines:  used for reading & writing numeric types.  Called by the jump tables at the very end.

void LoadDPR_UByte(void *SourcePtr)  {  DPRegister = (double) *(unsigned char *) SourcePtr;  }
void LoadDPR_SByte(void *SourcePtr)  {  DPRegister = (double) *(signed char *) SourcePtr;  }
void LoadDPR_UShort(void *SourcePtr)  {  DPRegister = (double) *(unsigned short *) SourcePtr;  }
void LoadDPR_SShort(void *SourcePtr)  {  DPRegister = (double) *(signed short *) SourcePtr;  }
void LoadDPR_ULong(void *SourcePtr)  {  DPRegister = (double) *(unsigned long *) SourcePtr;  }
void LoadDPR_SLong(void *SourcePtr)  {  DPRegister = (double) *(signed long *) SourcePtr;  }
void LoadDPR_Single(void *SourcePtr)  {  DPRegister = (double) *(float *) SourcePtr;  }
void LoadDPR_Double(void *SourcePtr)  {  DPRegister = (double) *(double *) SourcePtr;  }

void SaveDPR_UByte(void *DestPtr)  {
    if (!((DPRegister <= UCHAR_MAX) && (DPRegister >= 0)))  SetError(out_of_range_err, PCCodePtr-1);
    else *(unsigned char *) DestPtr = (unsigned char) DPRegister;  }
void SaveDPR_SByte(void *DestPtr)  {
    if (!((DPRegister <= SCHAR_MAX) && (DPRegister >= SCHAR_MIN)))  SetError(out_of_range_err, PCCodePtr-1);
    else *(signed char *) DestPtr = (signed char) DPRegister;  }
void SaveDPR_UShort(void *DestPtr)  {
    if (!((DPRegister <= USHRT_MAX) && (DPRegister >= 0)))  SetError(out_of_range_err, PCCodePtr-1);
    else *(unsigned short *) DestPtr = (unsigned short) DPRegister;  }
void SaveDPR_SShort(void *DestPtr)  {
    if (!((DPRegister <= SHRT_MAX) && (DPRegister >= SHRT_MIN)))  SetError(out_of_range_err, PCCodePtr-1);
    else *(signed short *) DestPtr = (signed short) DPRegister;  }
void SaveDPR_ULong(void *DestPtr)  {
    if (!((DPRegister <= ULONG_MAX) && (DPRegister >= 0)))  SetError(out_of_range_err, PCCodePtr-1);
    else *(unsigned long *) DestPtr = (unsigned long) DPRegister;  }
void SaveDPR_SLong(void *DestPtr)  {
    if (!((DPRegister <= LONG_MAX) && (DPRegister >= LONG_MIN)))  SetError(out_of_range_err, PCCodePtr-1);
    else *(signed long *) DestPtr = (signed long) DPRegister;  }
void SaveDPR_Single(void *DestPtr)  {
    if (!((DPRegister <= FLT_MAX) && (DPRegister >= -FLT_MAX)))  SetError(out_of_range_err, PCCodePtr-1);
    else *(float *) DestPtr = (float) DPRegister;  }
void SaveDPR_Double(void *DestPtr)  {  *(double *) DestPtr = (double) DPRegister;  }



// ****************************************
// The next routines are used by the skip jump table when hopping over a chunk of code to get to the end, or to the next code marker.
// They are perhaps similar to the CheckHobbish routines, but leave most of the checks out (for speed).
// These are grouped by the kind of arguments/parameters they expect in the bytecode.

void SkipNoArgs()  {  PCCodePtr++;  }
void SkipOneArg()  {  PCCodePtr++;  CallSkipFunction();  }
void SkipTwoArgs()  {  PCCodePtr++;  CallSkipFunction();  if (err_code == passed)  CallSkipFunction();  }
void SkipThreeArgs() {  PCCodePtr++;  CallSkipFunction();  if (err_code == passed)  CallSkipFunction();  if (err_code == passed)  CallSkipFunction();  }

void SkipLong()  {  PCCodePtr++;  PCCodePtr++;  }
void SkipDouble()  {  PCCodePtr++;  PCCodePtr += sizeof(double)/sizeof(Ulong);  }
void SkipString()  {  Slong StrLength;  PCCodePtr++;  StrLength = *PCCodePtr;  PCCodePtr += 1 + (Ulong) ceil(1.*StrLength/sizeof(Ulong));  }

void SkipLongAndOneArg()  {  PCCodePtr++;  PCCodePtr++;  CallSkipFunction();  }
void SkipLongAndTwoArgs()  {  PCCodePtr++;  PCCodePtr++;  CallSkipFunction();  if (err_code == passed)  CallSkipFunction();  }
void SkipOneArgAndLong()  {  PCCodePtr++;  CallSkipFunction();  PCCodePtr++;  }

void SkipCode()  {  PCCodePtr++;  RunSkipMode(0);  PCCodePtr++;  }


// ****************************************

// CheckHobbish() tests Hobbish code as far as possible before it's loaded.
// This works because the code is static during runtime; the dynamic checks that Yazoo performs are a chronic burden.
// 4 checks:  make sure each command is valid (< commands_num, not illegal_command); jumps are within the Hobbish block;
// jumps go to start of a sentence; and there is a null statement at the end.

void CheckHobbish()
{
    LL_retrn linkedLretrn;
    Ulong JumpCounter, SentenceCounter;
    
    
        // We keep track of all the goto addresses and sentence beginnings.
    
    err_index = 1;
    linkedLretrn = FNewLinkedList(&JumpList, 0, sizeof(Slong **), 200, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr);  return;  }
    linkedLretrn = FNewLinkedList(&JumpFromList, 0, sizeof(Slong **), 200, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr);  return;  }
    linkedLretrn = FNewLinkedList(&SentenceList, 0, sizeof(Slong **), 200, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr);  return;  }


        // Check the outermost layer of bytecode.
    
    CheckHobbishSentences();


        // Ensure that all gotos point to the start of some sentence.
    
    if (err_code == passed)  {
    for (JumpCounter = 1; JumpCounter <= JumpList.elementnum; JumpCounter++)    {
        for (SentenceCounter = 1; SentenceCounter <= SentenceList.elementnum; SentenceCounter++)   {
            if (*(Slong **) FElement(&JumpList, JumpCounter) == *(Slong **) FElement(&SentenceList, SentenceCounter))  break;
            else if (SentenceCounter == SentenceList.elementnum)  {
                PCCodePtr = *(Slong **) FElement(&JumpFromList, JumpCounter);    // to set err_index correctly below
                SetError(bad_jump_err, PCCodePtr);      // do this manually, since the code isn't yet loaded into memory
                JumpCounter = JumpList.elementnum;
                break;
    }}  }   }

    FDeleteLinkedList(&JumpList);      // clean up
    FDeleteLinkedList(&SentenceList);
    FDeleteLinkedList(&JumpFromList);


        // If we didn't get to the end of the code, we'll set a warning -- at least it won't crash the interpreter.
    
    if ((err_code == passed) && (PCCodePtr != EndCodePtr))  SetWarning(inaccessible_code_warning, PCCodePtr);
}


// CheckHobbishSentences() evaluates one sentence after another to make sure each command is valid.
// Stops at the null terminating character, or if there's an error.

void CheckHobbishSentences()
{
    LL_retrn linkedLretrn;

    while (*PCCodePtr != null_cmd)   {


            // keep track of our sentence beginnings for CheckHobbish()
            
        linkedLretrn = FAddElements(&SentenceList, 1, uncleared);
        if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr);  return;  }
        *(Slong **) FElement(&SentenceList, SentenceList.elementnum) = PCCodePtr;
        
        
            // run each outermost command
        
        Check_CallHobbishFunction(start_arg);
        if (err_code != passed)  break;         }
    
    
        // add the null terminator as a valid jump point
    
    linkedLretrn = FAddElements(&SentenceList, 1, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr);  return;  }
    *(Slong **) FElement(&SentenceList, SentenceList.elementnum) = PCCodePtr;
    
    if (err_code == passed)  PCCodePtr++;   // get past the null terminator
}


// These next few routines skip over Yazoo commands having various combinations of arguments/inlined constants.
// They check to make sure the code doesn't spill past EndCodePtr, and the jump tables take care of illegal commands.

void Check_YazooCmd(Slong CurrentCmd)      // checks bread-and-butter operators
{
    if (*PCCodePtr == def_general)  PCCodePtr++;     // do an extra step to get over the flags
    PCCodePtr++;
    if ((left_args[CurrentCmd] != no_arg) && (err_code == passed))  Check_CallHobbishFunction(left_args[CurrentCmd]);
    if ((right_args[CurrentCmd] != no_arg) && (err_code == passed))  Check_CallHobbishFunction(right_args[CurrentCmd]);
}

void Check_FunctionCmd(Slong CurrentCmd)      // checks user-defined/built-in functions
{
    PCCodePtr ++;
    if (CurrentCmd == built_in_function)    {
        if ((*PCCodePtr < 0) || (*PCCodePtr >= bi_commands_num))  {
            SetError(nonexistent_built_in_function_err, PCCodePtr);
            return;         }
        PCCodePtr++;
        if (PCCodePtr >= EndCodePtr)  {  PCCodePtr = EndCodePtr;  SetError(code_overflow_err, PCCodePtr);  return;  }    }
    /*    else if (CurrentCmd == user_function)   {
        if (*PCCodePtr < 0)  {
            SetError(illegal_command_err, PCCodePtr);
            return;
    }   }*/

    if (CurrentCmd == user_function)  Check_CallHobbishFunction(var_arg);
    
    if (err_code == passed)  Check_CallHobbishFunction(var_arg);
}

void Check_LongCmd(Slong CurrentCmd)          // checks commands that have long integer constants
{
    PCCodePtr++;
    if (CurrentCmd == step_to_memberID)  {
        if ((left_args[CurrentCmd] != no_arg) && (err_code == passed))  Check_CallHobbishFunction(left_args[CurrentCmd]);
        if ((right_args[CurrentCmd] != no_arg) && (err_code == passed))  Check_CallHobbishFunction(right_args[CurrentCmd]);   }
    
    if (CurrentCmd == internal_var)  {
        if ((*PCCodePtr < 0) || (*PCCodePtr >= internal_vars_num))  {
            SetError(nonexistent_register_err, PCCodePtr);
    }   }
    
    if (err_code != passed)  return;
    PCCodePtr++;
    if (PCCodePtr >= EndCodePtr)  {  PCCodePtr = EndCodePtr;  SetError(code_overflow_err, PCCodePtr);  return;  }

    if (CurrentCmd != step_to_memberID)  {
        if ((left_args[CurrentCmd] != no_arg) && (err_code == passed))  Check_CallHobbishFunction(left_args[CurrentCmd]);
        if ((right_args[CurrentCmd] != no_arg) && (err_code == passed))  Check_CallHobbishFunction(right_args[CurrentCmd]);    }
}

void Check_DoubleCmd(Slong CurrentCmd)    // checks inlined doubles
{
    PCCodePtr += sizeof(double)/sizeof(Ulong) + 1;
    if (PCCodePtr >= EndCodePtr)  {  PCCodePtr = EndCodePtr;  SetError(code_overflow_err, PCCodePtr);  return;  }
}

void Check_StringCmd(Slong CurrentCmd)    // checks inlined strings
{
    Slong StringSize;
    
    PCCodePtr++;
    StringSize = *PCCodePtr;
    if (StringSize < 0)  {  SetError(illegal_command_err, PCCodePtr);  return;  }
    
    StringSize += Align(StringSize);
    
    PCCodePtr++;
    
    if (PCCodePtr + (Ulong) ceil(1.*StringSize/sizeof(Ulong)) >= EndCodePtr)      // step over the width of the string
        {  PCCodePtr = EndCodePtr;  SetError(code_overflow_err, PCCodePtr);  return;  }
    PCCodePtr += (Ulong) ceil(1.*StringSize/sizeof(Ulong));
}

void Check_CodeCmd(Slong CurrentCmd)      // checks a code block
{
    PCCodePtr++;
    CheckHobbishSentences();      // each code block contains many sentences
}

void Check_JumpCmd(Slong CurrentCmd)     // cheks conditional/unconditional jumps
{
    signed long JumpOffset;
    LL_retrn linkedLretrn;
    
    PCCodePtr++;
    JumpOffset = *PCCodePtr;

    if (PCCodePtr + JumpOffset >= EndCodePtr)  {  SetError(bad_jump_err, PCCodePtr);  return;  }
    
    
        // Add the jump position to CheckHobbish()'s JumpList.
    
    linkedLretrn = FAddElements(&JumpList, 1, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr);  return;  }
    linkedLretrn = FAddElements(&JumpFromList, 1, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr);  return;  }
    *(Slong **) FElement(&JumpList, JumpList.elementnum) = PCCodePtr + JumpOffset;
    *(Slong **) FElement(&JumpFromList, JumpFromList.elementnum) = PCCodePtr;
    
    PCCodePtr++;
    if (PCCodePtr >= EndCodePtr)  {  PCCodePtr = EndCodePtr;  SetError(code_overflow_err, PCCodePtr);  return;  }

    if ((right_args[CurrentCmd] != no_arg) && (err_code == passed))  Check_CallHobbishFunction(right_args[CurrentCmd]);
}

void Check_IndicesCmd(Slong CurrentCmd)     // checks array operators [] of various sorts
{
    PCCodePtr++;
    if ((left_args[CurrentCmd] != no_arg) && (err_code == passed))  Check_CallHobbishFunction(left_args[CurrentCmd]);
    if ((right_args[CurrentCmd] != no_arg) && (err_code == passed))  Check_CallHobbishFunction(right_args[CurrentCmd]);

    if (err_code != passed)  return;
    Check_CallHobbishFunction(data_arg);
    if (err_code != passed)  return;

    if ((CurrentCmd == step_to_indices) || (CurrentCmd == insert_indices_up) || (CurrentCmd == insert_indices_down))  {
        Check_CallHobbishFunction(data_arg);
        if (err_code != passed)  return;     }
}

void Check_CallHobbishFunction(Slong TableID)     // runs a bytecode function via the check jump tables
{
    if (*PCCodePtr >= commands_num)  {  SetError(illegal_command_err, PCCodePtr);  return;  }
    if (PCCodePtr + 1 >= EndCodePtr)  {  PCCodePtr = EndCodePtr;  SetError(code_overflow_err, PCCodePtr);  return;  }
    
    AllowedCMDJumpTable[TableID][*PCCodePtr](*PCCodePtr);
}

void IllegalCmd(Slong CurrentCmd)           // the checker encountered a command that is out of place -- throw an error
{
    SetError(illegal_command_err, PCCodePtr);
}




// ****************************************
// The following are the main loops of the Yazoo interpreter.
// These routines read in the appropriate commands and call them from the jump tables.


// BeginExecution() stores the current program counter & related information on the 'stack',
// then moves it to a new location and starts running a new code.

void BeginExecution(code_ref *theCode, Boolean if_store_PC, Ulong NewPCOffset, Ulong NewPC_IW, Ulong CodesToSkip)
{
    code_ref HoldPCCodeRef = PCCodeRef;
    Slong *HoldPCCodePtr = PCCodePtr, *HoldStartCodePtr = StartCodePtr;
    define_params HoldDefineParams = GL_Define;
    object_params HoldObjectParams = GL_Object;
    step_params HoldStepParams = GL_Path;
    void *OldStackTop, *NewStackTop;
    LL_retrn linkedLretrn;
    
    
        // don't go too many levels deep or we will blow the (real) stack
    
    if (GL_RecursionCounter >= GL_MaxRecursions)  {  SetError(recursion_depth_err, PCCodePtr);  return;  }
    GL_RecursionCounter++;


        // point to the code we will run
        
    PCCodeRef = *theCode;
    StartCodePtr = *(Slong **) FElement(&(MasterCodeList.data), theCode->PLL_index);
    PCCodePtr = PCCodeRef.code_ptr;

    
        // store the old state on our 'stack'
    
    if (if_store_PC == yz_true)  {
        linkedLretrn = PushStack(&PCStack, &NewStackTop);
        if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr);  PCCodeRef = HoldPCCodeRef;  return;  }
        *(bundle_coordinate *) NewStackTop = PCCoordinate;
    
        linkedLretrn = PushStack(&PCStack, &NewStackTop);
        if (linkedLretrn != passed)    {  SetError(out_of_mem_err, PCCodePtr);  PCCodeRef = HoldPCCodeRef;  return;  }
        *(bundle_coordinate *) NewStackTop = SearchCoordinate;   }


        // move our memory location to the code's variable
    
    PCCoordinate._Fibril = theCode->anchor;
    PCCoordinate._Bundle = PCCoordinate._Fibril->bundle_ptr;
    PCCoordinate._Offset = NewPCOffset;
    PCCoordinate._IndexWidth = NewPC_IW;
    
    RefBundle(PCCoordinate._Bundle);
    RefFibril(PCCoordinate._Fibril);
    SearchCoordinate._Fibril = NULL;       // no deref since it is now a copy of HoldSearchCoordinate._Fibril

    GL_Define.DefineOn = yz_false;
    GL_Object.CodeLL.memory = NULL;        // don't dereference, since we have it backed up in HoldObjectParams
    err_code = passed;


        // Skip over codes we are not running (i.e. find the Nth 'code' marker), and go.
    
    if (CodesToSkip != 0)   {
        RunSkipMode(CodesToSkip);
        if (*PCCodePtr == code_marker)  {
            PCCodePtr++;
            RunHobbish();   }   }
    else  RunHobbish();


        // dereference the old location variables
    
    DeleteObjectCodeLL(&(GL_Object.CodeLL));

    DerefBundle(&(PCCoordinate._Bundle));
    DerefFibril(&(PCCoordinate._Fibril));


        // pull the old location info off our 'stack'
    
    if (if_store_PC == yz_true)   {
        DerefFibril(&(SearchCoordinate._Fibril));
    
        PopStack(&PCStack, &OldStackTop);
        SearchCoordinate = *(bundle_coordinate *) OldStackTop;
    
        PopStack(&PCStack, &OldStackTop);
        PCCoordinate = *(bundle_coordinate *) OldStackTop;    }
    
    
        // restore the old code pointer
    
    StartCodePtr = HoldStartCodePtr;
    PCCodePtr = HoldPCCodePtr;
    PCCodeRef = HoldPCCodeRef;
    GL_Define = HoldDefineParams;
    GL_Object = HoldObjectParams;
    GL_Path = HoldStepParams;

    GL_RecursionCounter--;
}


// RunHobbish() runs a succession of bytecode sentences, until there is an error or else we hit the end or a 'code' marker.

void RunHobbish()
{
    GL_Define.DefineOn = yz_false;

    while ((*PCCodePtr != null_cmd) && (*PCCodePtr != code_marker))   {
        CallStartFunction();
        if (err_code != passed)  return;
    }
    PCCodePtr++;         // over the 0
}


// RunSkipMode() skips over a run of bytecode sentences, stopping at the first null terminator or 'code' marker, or upon error.
// This avoids running the commands by calling a different 'skip' jump table.

void RunSkipMode(Ulong CodesToSkip)
{
    while (*PCCodePtr != null_cmd)  {
        if ((*PCCodePtr == code_marker) && (CodesToSkip != 0))    {
            CodesToSkip--;
            if (CodesToSkip == 0)  return;    }
        CallSkipFunction();
        if (err_code != passed)  return;
    }
}


// These next 6 routines run various functions by calling the function-ID term in the appropriate jump table.
// Commands sometimes behave differently when they begin a sentence; hence the CallStartFunction() function's table
// sometimes has pointers to different functions than the main tables.
// The other jump tables could in principle be put into one, except that the compiler uses them to see what arguments are legal where.

void CallStartFunction()  {  StartJumpTable[*PCCodePtr]();  }
void CallHobbishFunction()  {  HobbishJumpTable[*PCCodePtr]();  }
void CallLogicalFunction()  {  HobbishJumpTable[*PCCodePtr]();  }
void CallNumericFunction()  {  NumericJumpTable[*PCCodePtr]();  }
void CallCodeFunction()  {  CodeNumber = 1;  CodeJumpTable[*PCCodePtr]();  }
void CallSkipFunction()  {  SkipJumpTable[*PCCodePtr]();  }


// Lastly, the jump tables, which store addresses of the operator-handling functions above in this file, in order of ID number.
// This is much faster than dynamically searching for the appropriate function.

void(*SkipJumpTable[commands_num])() = {     // for skipping over code
    &_illegal, &SkipOneArg, &SkipLong, &SkipLongAndOneArg, &_illegal,
        &SkipTwoArgs, &SkipNoArgs, &SkipLong, &SkipLongAndOneArg, &SkipLongAndOneArg,
    &SkipNoArgs, &SkipLongAndTwoArgs, &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs,
        &SkipTwoArgs, &SkipOneArg, &SkipLong, &SkipOneArgAndLong, &SkipTwoArgs,
    &SkipThreeArgs, &SkipOneArg, &SkipTwoArgs, &SkipThreeArgs, &SkipThreeArgs,
        &SkipOneArg, &_illegal, &SkipNoArgs, &SkipNoArgs, &SkipNoArgs,
    &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs,
        &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs,
    &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs,
        &SkipTwoArgs, &SkipTwoArgs, &SkipTwoArgs, &_illegal, &_illegal,
    &SkipLong, &SkipDouble, &SkipString, &SkipCode, &SkipNoArgs,
        &SkipNoArgs, &SkipNoArgs, &SkipNoArgs, &SkipNoArgs, &SkipNoArgs,
    &SkipNoArgs, &SkipNoArgs, &SkipOneArg, &SkipNoArgs, &SkipNoArgs, &SkipNoArgs      };

void(*StartJumpTable[commands_num])() = {    // start-of-sentence:  resize/insert will not step (prevents error on proxies[^10])
    &_illegal, &_func_return, &_internal_var, &_built_in_function, &_illegal,
        &_user_function, &_get_args, &_jump_always, &_jump_if_true, &_jump_if_false,
    &_exit_Yazoo, &_def_general, &_forced_equate, &_code_number, &_sub_code,
        &_append_code, &_remove_member, &_search_member, &_step_to_memberID, &_step_to_index,
    &_step_to_indices, &_step_to_all_indices, &_resize_start, &_insert_indices_up_start, &_insert_indices_down_start,
        &_delete_indices, &_illegal, &_this_var, &_that_var, &_no_var,
    &_if_eq, &_if_ne, &_if_gt, &_if_ge, &_if_lt,
        &_if_le, &_if_and, &_if_or, &_if_xor, &_if_not,
    &_if_eq_at, &_if_ne_at, &_addf, &_subf, &_mulf,
        &_divf, &_powerf, &_modi, &_illegal, &_illegal,
    &_constant_Slong, &_constant_float, &_constant_string, &_code_block, &_ubyte_cmd,
        &_sbyte_cmd, &_ushort_cmd, &_sshort_cmd, &_ulong_cmd, &_slong_cmd,
    &_float_cmd, &_double_cmd, &_block_cmd, &_string_cmd, &_illegal, &_code_marker   };

void(*NumericJumpTable[commands_num])() = {     // numeric arguments
    &_illegal, &Num_Adapter, &Num_Adapter, &Num_Adapter, &_illegal,
        &Num_Adapter, &Num_Adapter, &_illegal, &_illegal, &_illegal,
    &_illegal, &Num_Adapter, &Num_Adapter, &Num_Adapter, &Num_Adapter,
        &Num_Adapter, &_illegal, &Num_Adapter, &Num_Adapter, &Num_Adapter,
    &Num_Adapter, &Num_Adapter, &Num_Adapter, &Num_Adapter, &Num_Adapter,
        &_illegal, &_illegal, &Num_Adapter, &Num_Adapter, &Num_Adapter,
    &_illegal, &_illegal, &_illegal, &_illegal, &_illegal,
        &_illegal, &_illegal, &_illegal, &_illegal, &_illegal,
    &_illegal, &_illegal, &_addf, &_subf, &_mulf,
        &_divf, &_powerf, &_modi, &_illegal, &_illegal,
    &_constant_Slong, &_constant_float, &_illegal, &_illegal, &_illegal,
        &_illegal, &_illegal, &_illegal, &_illegal, &_illegal,
    &_illegal, &_illegal, &_illegal, &_illegal, &_illegal, &_illegal    };

void(*CodeJumpTable[commands_num])() = {     // right argument of define:  something that denotes the new type
    &_illegal, &_illegal, &VT_Adapter, &VT_Adapter, &_illegal,
        &VT_Adapter, &VT_Adapter, &_illegal, &_illegal, &_illegal,
    &_illegal, &VT_Adapter, &VT_Adapter, &_code_number, &_sub_code,
        &_append_code, &_illegal, &VT_Adapter, &VT_Adapter, &VT_Adapter,
    &VT_Adapter, &VT_Adapter, &VT_Adapter, &VT_Adapter, &VT_Adapter,
        &_illegal, &_illegal, &VT_Adapter, &VT_Adapter, &VT_Adapter,
    &_illegal, &_illegal, &_illegal, &_illegal, &_illegal,
        &_illegal, &_illegal, &_illegal, &_illegal, &_illegal,
    &_illegal, &_illegal, &VT_NumericAdapter, &VT_NumericAdapter, &VT_NumericAdapter,
        &VT_NumericAdapter, &VT_NumericAdapter, &VT_NumericAdapter, &_illegal, &_illegal,
    &VT_Adapter, &VT_Adapter, &VT_Adapter, &_code_block, &_ubyte_cmd,
        &_sbyte_cmd, &_ushort_cmd, &_sshort_cmd, &_ulong_cmd, &_slong_cmd,
    &_float_cmd, &_double_cmd, &_block_cmd, &_string_cmd, &_illegal, &_illegal   };

void(*HobbishJumpTable[commands_num])() = {     // generic jump table for mid-sentence operators
    &_illegal, &_func_return, &_internal_var, &_built_in_function, &_illegal,
        &_user_function, &_get_args, &_jump_always, &_jump_if_true, &_jump_if_false,
    &_exit_Yazoo, &_def_general, &_forced_equate, &_code_number, &_sub_code,
        &_append_code, &_remove_member, &_search_member, &_step_to_memberID, &_step_to_index,
    &_step_to_indices, &_step_to_all_indices, &_resize, &_insert_indices_up, &_insert_indices_down,
        &_delete_indices, &_illegal, &_this_var, &_that_var, &_no_var,
    &_if_eq, &_if_ne, &_if_gt, &_if_ge, &_if_lt,
        &_if_le, &_if_and, &_if_or, &_if_xor, &_if_not,
    &_if_eq_at, &_if_ne_at, &_addf, &_subf, &_mulf,
        &_divf, &_powerf, &_modi, &_illegal, &_illegal,
    &_constant_Slong, &_constant_float, &_constant_string, &_code_block, &_ubyte_cmd,
        &_sbyte_cmd, &_ushort_cmd, &_sshort_cmd, &_ulong_cmd, &_slong_cmd,
    &_float_cmd, &_double_cmd, &_block_cmd, &_string_cmd, &_illegal, &_code_marker   };

void(*LoadDPR_JumpTable[8])(void *) =    // for storing a number in DPRegister
        {  &LoadDPR_UByte, &LoadDPR_SByte, &LoadDPR_UShort, &LoadDPR_SShort, &LoadDPR_ULong, &LoadDPR_SLong, &LoadDPR_Single, &LoadDPR_Double  };
                
void(*SaveDPR_JumpTable[8])(void *) =    // for getting a number out of DPRegister
        {  &SaveDPR_UByte, &SaveDPR_SByte, &SaveDPR_UShort, &SaveDPR_SShort, &SaveDPR_ULong, &SaveDPR_SLong, &SaveDPR_Single, &SaveDPR_Double  };
