/*
 *  hobbsh.h
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef Hobbish_h
#define Hobbish_h

#include "intrpt.h"



// define_params contains the parameters relevant to defining new members & variables

typedef struct {
    Ulong Flags;          // defined in yazooh.h
    Boolean DefineOn;     // yes only for first argument of def commands
    Boolean FinalStep;    // only used for outer step command
    Boolean RemoveMode;   // set only by remove_member()
} define_params;


// object_params describes what kind of object we stepped across; set by various commands: ulong_cmd, search_var, etc.

typedef struct {
    unsigned char Type;
    Boolean IfVoid;            // set by *
    unsigned long TypeSize;    // for block and string types only
    char *StringPtr;           // for constant strings, the ptr in the code where the string begins

    bundle_coordinate Target;  // the member bundle pointed to by ::=@, =@;  set by sub_code
    linkedlist *CodeList;
    linkedlist CodeLL;         // used for communicating code lists not stored in variables (from :s)
} object_params;



// step_params describes the last step in a path; used for communicating between def/navigation commands

typedef struct {
    member *StemMember;
    Ulong StemMemberNumber;
    signed long StemMemberID;
    Ulong Offset;
    Ulong Indices;

    bundle_coordinate StemCoordinate;     // used sometimes when defining variables; stores the coordinate before an unsuccessful step
    Ulong OldSearchIndexWidth;
} step_params;


// constants for Prune()

#define pn_member 0
#define pn_indices 1




// *** Famous & global variables ***

typedef struct {
    
    
        // Zero -- the top-level variable, with its entourage of fibrils, codes, etc.
    
    bundle *_Zero, *_Zero_workspace;
    fibril *_ZeroSuspensor;
    
    
        // these hold the error/warning codes & locations
    
    Ulong _err_code, _err_index, _warning_code, _warning_index;
    code_ref _err_script, _warning_script;
    
    
        // DPRegister holds numeric quantities from inlined constants, variables, etc.
    
    double _DPRegister;
    
    
        // GL_Conditional stores the true/false result of a conditional expression.
    
    Boolean _GL_Conditional;
    
    
        // PC stands for Program Counter.
        // PCCodePtr is the standard PC -- points to the active location in the bytecode.
    
    Slong *_PCCodePtr, *_StartCodePtr, *_EndCodePtr;
    
    
        // the following convey type, path, define-status, etc. of objects that Yazoo encounters
    
    define_params _GL_Define;
    object_params _GL_Object;
    step_params _GL_Path;
    
    
        // PCCoordinate/SearchCoordinate are the Yazoo memory addresses of the
        // active function/the current search path (a.b) respectively.
    
    bundle_coordinate _PCCoordinate, _SearchCoordinate;
    
    
        // A series of lesser Yazoo memory addresses, holding args, that, the return variable, etc.  BIF = built-in-function.
    
    bundle_coordinate _BIF_ArgsCoordinate, _ArgsCoordinate, _ReturnCoordinate, _ThatCoordinate;
    
    
        // Miscellaneous globals
    
    Ulong _CodeNumber;                     // set by the # operator
    Ulong _GL_MaxDigits;                   // for printing numbers
    Ulong _GL_RecursionCounter;            // current nesting of running codes
    Slong _TempDefGenName;                 // temp. name for new members while the object block is running
    code_ref *_OneCodeRef, _PCCodeRef;
    linkedlist _OneCodeLL;                 // for passing codes from code blocks; keep this permanently defined
    Boolean _AddedAMember, _FailedLastStep;
    linkedlist _JumpList, _JumpFromList, _SentenceList;         // these are used for checking code
} yz_bytecode_global_struct;

extern yz_bytecode_global_struct yz_bytecode_globals;

#define Zero yz_bytecode_globals._Zero
#define Zero_workspace yz_bytecode_globals._Zero_workspace
#define ZeroSuspensor yz_bytecode_globals._ZeroSuspensor

#define err_code yz_bytecode_globals._err_code
#define err_index yz_bytecode_globals._err_index
#define warning_code yz_bytecode_globals._warning_code
#define warning_index yz_bytecode_globals._warning_index
#define err_script yz_bytecode_globals._err_script
#define warning_script yz_bytecode_globals._warning_script

#define DPRegister yz_bytecode_globals._DPRegister

#define GL_Conditional yz_bytecode_globals._GL_Conditional

#define PCCodePtr yz_bytecode_globals._PCCodePtr
#define StartCodePtr yz_bytecode_globals._StartCodePtr
#define EndCodePtr yz_bytecode_globals._EndCodePtr

#define GL_Define yz_bytecode_globals._GL_Define
#define GL_Object yz_bytecode_globals._GL_Object
#define GL_Path yz_bytecode_globals._GL_Path

#define PCCoordinate yz_bytecode_globals._PCCoordinate
#define SearchCoordinate yz_bytecode_globals._SearchCoordinate

#define BIF_ArgsCoordinate yz_bytecode_globals._BIF_ArgsCoordinate
#define ArgsCoordinate yz_bytecode_globals._ArgsCoordinate
#define ReturnCoordinate yz_bytecode_globals._ReturnCoordinate
#define ThatCoordinate yz_bytecode_globals._ThatCoordinate

#define CodeNumber yz_bytecode_globals._CodeNumber
#define GL_MaxDigits yz_bytecode_globals._GL_MaxDigits
#define GL_RecursionCounter yz_bytecode_globals._GL_RecursionCounter
#define TempDefGenName yz_bytecode_globals._TempDefGenName
#define OneCodeRef yz_bytecode_globals._OneCodeRef
#define PCCodeRef yz_bytecode_globals._PCCodeRef
#define OneCodeLL yz_bytecode_globals._OneCodeLL
#define AddedAMember yz_bytecode_globals._AddedAMember
#define FailedLastStep yz_bytecode_globals._FailedLastStep
#define JumpList yz_bytecode_globals._JumpList
#define JumpFromList yz_bytecode_globals._JumpFromList
#define SentenceList yz_bytecode_globals._SentenceList




// Prototypes

extern void _func_return();
extern void _internal_var();
extern void _built_in_function();
extern void _user_function();
extern void _get_args();
extern void _jump_always();
extern void _jump_if_true();
extern void _jump_if_false();
extern void _exit_Yazoo();

extern void _forced_equate();

extern void _code_number();
extern void _sub_code();
extern void _append_code();

extern void _remove_member();


extern void DefGeneral(Boolean, Boolean, Boolean);
extern void LoadString(bundle_coordinate *, void **);
extern void CheckType(linkedlist *, unsigned char *, Boolean);
extern void UpdateType(linkedlist *, unsigned char *, bundle_coordinate *, Boolean, Boolean);

extern void _search_member();
extern void SM_Step();
extern void _step_to_memberID();
extern void SID_Step();
extern void _step_to_index();
extern void STI_Step();
extern void _step_to_indices();
extern void STIs_Step();
extern void _step_to_all_indices();
extern void STAI_Step();
extern void _this_var();
extern void _that_var();
extern void _no_var();

extern void Navigate(void (*)());
extern void AddVB(bundle **, Ulong, Ulong, Boolean);
extern void RelinkGLSearchMember(define_params *, object_params *, Boolean);
extern void _resize_start();
extern void _resize();
extern void DoResize();
extern void ResizeBundle(variable *, member *, Ulong);
extern void _insert_indices_up();
extern void _insert_indices_down();
extern void _insert_indices_up_start();
extern void _insert_indices_down_start();
extern void MasterInsert(Boolean, Boolean);
extern void _delete_indices();
extern void Prune(Boolean);
extern void ResizeIndices(variable *, member *, Ulong, Ulong, signed long);

extern void _if_eq();
extern void Compare_ReadArgument(char **, Ulong *);
extern void _if_ne();
extern void _if_eq_at();
extern void iea_one_arg();
extern void _if_ne_at();
extern void _if_and();
extern void _if_or();
extern void _if_xor();

extern void Logic_And(Boolean);
extern void Logic_Or(Boolean);
extern void Logic_Eor(Boolean);
extern void LogicTemplate(void(*Logic_Operation)(Boolean));

extern void _if_not();

extern void _addf();
extern void _subf();
extern void _mulf();
extern void _divf();
extern void _powf();
extern void _modi();

extern void _if_lt();
extern void _if_le();
extern void _if_gt();
extern void _if_ge();

extern void Math_Add(double FirstArgument);
extern void Math_Sub(double FirstArgument);
extern void Math_Mul(double FirstArgument);
extern void Math_Div(double FirstArgument);
extern void Math_Pow(double FirstArgument);
extern void Math_Mod(double FirstArgument);

extern void Math_LT(double);
extern void Math_LE(double);
extern void Math_GT(double);
extern void Math_GE(double);

extern void MathTemplate(void(*Math_Operation)(double));

extern void _addf();
extern void _subf();
extern void _mulf();
extern void _divf();
extern void _powerf();
extern void _modi();

extern void _constant_Slong();
extern void _constant_float();
extern void _code_block();
extern void _constant_string();

extern void _ubyte_cmd();
extern void _sbyte_cmd();
extern void _ushort_cmd();
extern void _sshort_cmd();
extern void _ulong_cmd();
extern void _slong_cmd();
extern void _float_cmd();
extern void _double_cmd();
extern void _string_cmd();
extern void _block_cmd();
void SetPrimitiveArgType(Ulong);

extern void _illegal();

extern void InitGLObject();
extern void RefGLObjectCodeLL();
extern void DeleteObjectCodeLL(linkedlist *);

extern void Num_Adapter();
extern void VT_Adapter();
extern void VT_NumericAdapter();

extern void SkipNoArgs();
extern void SkipOneArg();
extern void SkipTwoArgs();
extern void SkipThreeArgs();
extern void SkipLong();
extern void SkipDouble();
extern void SkipString();
extern void SkipLongAndOneArg();
extern void SkipLongAndTwoArgs();
extern void SkipOneArgAndLong();
extern void SkipCode();

extern void CheckHobbish();
extern void CheckHobbishSentences();

extern void Check_YazooCmd(Slong);
extern void Check_FunctionCmd(Slong);
extern void Check_LongCmd(Slong);
extern void Check_DoubleCmd(Slong);
extern void Check_StringCmd(Slong);
extern void Check_CodeCmd(Slong);
extern void Check_JumpCmd(Slong);
extern void Check_IndicesCmd(Slong);
extern void IllegalCmd(Slong);

extern void BeginExecution(code_ref *, Boolean, Ulong, Ulong, Ulong);
extern void RunHobbish();
extern void RunSkipMode(Ulong);

extern void CallHobbishFunction();
extern void CallStartFunction();
extern void CallLogicalFunction();
extern void CallNumericFunction();
extern void CallCodeFunction();
extern void CallSkipFunction();
extern void Check_CallHobbishFunction(Slong);

#define SaveDPRegister(MemoryPtr, type) (SaveDPR_JumpTable[type](MemoryPtr))
#define LoadDPRegister(SourcePtr, type) (LoadDPR_JumpTable[type](SourcePtr))


extern void(*LoadDPR_JumpTable[8])(void *);
extern void(*SaveDPR_JumpTable[8])(void *);

extern void(*StartJumpTable[commands_num])();
extern void(*HobbishJumpTable[commands_num])();
extern void(*SkipJumpTable[commands_num])();
extern void(*NumericJumpTable[commands_num])();
extern void(*CodeJumpTable[commands_num])();
extern void(*HobbishJumpTable[commands_num])();


// misc

#define no_string -1

#endif
