/*
 *  llmast.c(pp) -- stack/PLL routines; compiles lnklst.c(pp)
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "lnklst.h"


// Prototypes (only user-accessible routines are listed here)

/*LL_retrn NewLinkedList(linkedlist *, unsigned long, unsigned short, unsigned char, Boolean);
LL_retrn DeleteLinkedList(linkedlist *);
LL_retrn InsertElements(linkedlist *, unsigned long, unsigned long, Boolean);
LL_retrn AddElements(linkedlist *, unsigned long, Boolean);
LL_retrn DeleteElements(linkedlist *, unsigned long, unsigned long);
LL_retrn DeleteElement(linkedlist *, unsigned long);
LL_retrn DefragmentLinkedList(linkedlist *);
LL_retrn CopyElements(linkedlist *, unsigned long, linkedlist *, unsigned long, unsigned long);
LL_retrn CompareElements(linkedlist *, unsigned long, linkedlist *, unsigned long, unsigned long);
LL_retrn FillElements(linkedlist *, unsigned long, unsigned long, char);
LL_retrn ClearElements(linkedlist *, unsigned long, unsigned long);
LL_retrn SetElements(linkedlist *, unsigned long, unsigned long, void *);
LL_retrn SetElement(linkedlist*, unsigned long, void *);
LL_retrn GetElements(linkedlist *, unsigned long, unsigned long, void *);
LL_retrn GetElement(linkedlist *, unsigned long, void *);
LL_retrn ElementExists(linkedlist *, unsigned long);
void * Element(linkedlist *, unsigned long);

LL_retrn FNewLinkedList(linkedlist *, unsigned long, unsigned short, unsigned char, Boolean);
void FDeleteLinkedList(linkedlist *);
LL_retrn FInsertElements(linkedlist *, unsigned long, unsigned long, Boolean);
LL_retrn FAddElements(linkedlist *, unsigned long, Boolean);
LL_retrn FDeleteElements(linkedlist *, unsigned long, unsigned long);
LL_retrn FDeleteElement(linkedlist *, unsigned long);
LL_retrn FDefragmentLinkedList(linkedlist *);
void FCopyElements(linkedlist *, unsigned long, linkedlist *, unsigned long, unsigned long);
LL_retrn FCompareElements(linkedlist *, unsigned long, linkedlist *, unsigned long, unsigned long);
void FFillElements(linkedlist *, unsigned long, unsigned long, char);
void FClearElements(linkedlist *, unsigned long, unsigned long);
void FSetElements(linkedlist *, unsigned long, unsigned long, void *);
void FSetElement(linkedlist*, unsigned long, void *);
void FGetElements(linkedlist *, unsigned long, unsigned long, void *);
void FGetElement(linkedlist *, unsigned long, void *);
LL_retrn FElementExists(linkedlist *, unsigned long);
void * FElement(linkedlist *, unsigned long);
void * FSkipElements(linkedlist *, sublistHeader **, unsigned long *, unsigned long);    // no equivalent SkipElements

LL_retrn NewPLL(pinned_LL *, Ulong, Ulong, unsigned char);
void DeletePLL(pinned_LL *);
LL_retrn AddPLLElement(pinned_LL *, void **, Ulong *);
void RefPLLElement(pinned_LL *, Ulong);
Ulong DerefPLLElement(pinned_LL *, Ulong);
void DerefPLLPtr(pinned_LL *, Ulong, Ulong *);

LL_retrn NewStack(stack *, Ulong, Ulong, unsigned char);
void DeleteStack(stack *);
LL_retrn PushStack(stack *, void **);
LL_retrn PopStack(stack *, void **);*/


// Linked list and Fast Linked List routines:  unbundle them here.

#include "lnklst.cpp"
#define LL_FastFlag
#include "lnklst.cpp"
#undef LL_FastFlag




// *************** PLL routines *****************

// PLL is for Pinned Linked List.  This is a list whose elements stay fixed at the same memory location.
// We can only add elements to the end of this list (which, by my present LL scheme, does not alter the memory location of elements
// below the insertion point).  When elements are deleted, we simply leave the LL alone, and make a note of the fact that an empty spot
// is there for the taking (more precisely, the lowest and highest empty spots are stored as lowest_free_spot and highest_free_spot).
// An element is considered deleted when the corresponding entry in 'references' goes to 0 -- i.e. the element
// is no longer being used by anybody.


// NewPLL() initializes a PLL by allocating its linkedlists

LL_retrn NewPLL(pinned_LL *BasePLL, Ulong NumElements, Ulong DataSize, unsigned char SpareRoom)
{
    LL_retrn linkedLretrn;

    linkedLretrn = NewLinkedList(&(BasePLL->data), NumElements, DataSize, SpareRoom, uncleared);
    if (linkedLretrn != passed)  return linkedLretrn;
    linkedLretrn = NewLinkedList(&(BasePLL->references), NumElements, sizeof(Ulong), SpareRoom, cleared);
    if (linkedLretrn != passed)  return linkedLretrn;
    BasePLL->lowest_free_spot = 1;
    BasePLL->highest_free_spot = NumElements;

    return passed;
}


// DeletePLL() deallocates a PLL by removing its linkedlists

void DeletePLL(pinned_LL *BasePLL)
{
    DeleteLinkedList(&(BasePLL->data));
    DeleteLinkedList(&(BasePLL->references));
}


// AddPLLElement() looks for an empty element, adding to the top of the linkedlist if none is to be found in the middle.
// Thus the user has no control over where in the linkedlist the new element actually goes.

LL_retrn AddPLLElement(pinned_LL *BasePLL, void **NewElPtr, Ulong *NewElIndex)
{
    Ulong counter;
    LL_retrn linkedLretrn;


        // Case 1:  add to the top of the list

    if (BasePLL->lowest_free_spot > BasePLL->data.elementnum)   {           // shouldn't move what's in the lower sublists
        linkedLretrn = FAddElements(&(BasePLL->data), 1, uncleared);        // with my current InsertElements scheme
        if (linkedLretrn != passed)  return linkedLretrn;
        linkedLretrn = FAddElements(&(BasePLL->references), 1, cleared);     // references must be initially set to 0
        if (linkedLretrn != passed)  return linkedLretrn;

        *NewElIndex = BasePLL->data.elementnum;
        *NewElPtr = FElement(&(BasePLL->data), *NewElIndex);
        BasePLL->lowest_free_spot = *NewElIndex+1;
        BasePLL->highest_free_spot = 1;         }   // cannot equal lowest_free_spot: we need it to be lower than any spots that open up


        // Case 2:  go to the lowest free element in the existing list

    else  {
        *NewElIndex = BasePLL->lowest_free_spot;
        *NewElPtr = FElement(&(BasePLL->data), *NewElIndex);
        *(Ulong *) FElement(&(BasePLL->references), *NewElIndex) = 0;
        if (BasePLL->lowest_free_spot >= BasePLL->highest_free_spot)  {
            BasePLL->lowest_free_spot = BasePLL->data.elementnum+1;
            BasePLL->highest_free_spot = 1;     }
        else  {
            for (counter = BasePLL->lowest_free_spot+1; counter <= BasePLL->highest_free_spot; counter++)  {
                if (*(Ulong *) FElement(&(BasePLL->references), counter) == 0)  {
                    BasePLL->lowest_free_spot = counter;
                    counter = BasePLL->highest_free_spot;
    }   }   }   }

    return passed;
}


// RefPLLElement() adds one to the number of references for a given element.

void RefPLLElement(pinned_LL *BasePLL, Ulong ElIndex)
{
    (*(Ulong *) Element(&(BasePLL->references), ElIndex))++;
    return;
}


// DerefPLLElement() decrements the # of references of an element given its index; updates lowest/highest_free_spot variables if refs = 0.
// Don't need a DeletePLLElement() -- below, if we run out of references, the element is considered free.

Ulong DerefPLLElement(pinned_LL *BasePLL, Ulong ElIndex)
{
    Ulong *ref_ptr = (Ulong *) Element(&(BasePLL->references), ElIndex);

    DerefPLLPtr(BasePLL, ElIndex, ref_ptr);

    return *ref_ptr;
}


// DerefPLLPtr() decrements the # of references given its index and a pointer to its reference list.
// (Should be faster than DerefPLLElement(), if the ref. list is available).
// Don't need to return the number of references, since the user already has ref_ptr.

void DerefPLLPtr(pinned_LL *BasePLL, Ulong ElIndex, Ulong *ref_ptr)
{
    (*ref_ptr)--;
    if (*ref_ptr == 0)  {
        if (ElIndex < BasePLL->lowest_free_spot)  BasePLL->lowest_free_spot = ElIndex;
        if (ElIndex > BasePLL->highest_free_spot)  BasePLL->highest_free_spot = ElIndex;    }

    return;
}



// *************** Stack routines *****************

// The stacks add to/remove from the top of the linked lists; the difference is that when elements are removed from the top,
// their storage space is not freed (to make it faster to push new elements later on).
// The only stack in Yazoo is for PCCodePtr, and Yazoo checks (via the recursion-depth error) to make sure that this doesn't get too big.
// This is somewhat redundant, since the LL routines also store extra space, so this may change.


// NewStack() initializes a stack by opening up its LL.
// InitialAllocation is (roughly) the size of free space in the linked list; it should not imply that the stack top doesn't start at zero.

LL_retrn NewStack(stack *theStack, Ulong ElementSize, Ulong InitialAllocation, unsigned char SpareRoom)
{
    LL_retrn retrn;

    retrn = FNewLinkedList(&(theStack->data), (Ulong) (InitialAllocation/(1.+SpareRoom)), ElementSize, SpareRoom, uncleared);
    if (retrn != passed)  return retrn;

    theStack->top = 0;

    return passed;
}


// DeleteStack() frees the LL memory in the stack.

void DeleteStack(stack *theStack)
{
    FDeleteLinkedList(&(theStack->data));

    return;
}


// PushStack() adds an element to (the top of) the stack

LL_retrn PushStack(stack *theStack, void **NewElPtr)
{
    LL_retrn retrn;

    if (theStack->top == theStack->data.elementnum)  {
        retrn = FAddElements(&(theStack->data), 1, uncleared);
        if (retrn != passed)  return retrn;         }

    (theStack->top)++;
    *NewElPtr = FElement(&(theStack->data), theStack->top);

    return passed;
}


// PopStack() decrements the stack top and sets OldElPtr to that element
// Due to the way stacks work, that new element will not be overwritten until PushStack() is called again.

LL_retrn PopStack(stack *theStack, void **OldElPtr)
{
    *OldElPtr = FElement(&(theStack->data), theStack->top);

    (theStack->top)--;

    return passed;
}
