/*
 *  lnklst.h
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


#ifndef LinkedList_h
#define LinkedList_h

#include "yazooh.h"


// Return type used by linked list routines

#define LL_retrn char


// shorthand for lnklst.c(pp)

#define SublistHeader(a) ((sublistHeader *) a)

#define linkedlistSize sizeof(linkedlist)
#define sublistHeaderSize sizeof(sublistHeader)


// misc. definitions

#define cleared yz_true
#define uncleared yz_false

#define mem_err 1
#define range_err 2
#define init_err 3
#define list_mismatch_err 4

#define equal 100
#define notequal 101



// There is one header for the entire linked list, which is stored in a variable of type linkedlist (below).
// The header points to the first sublist in the queue; each sublist begins with a 12(?)-byte block of memory
// of type sublistHeader (also below), which gives information specific to that particular sublist.

typedef struct sH_temp {
    unsigned long num_elements;     // number of elements that are officially registered to be in this sublist
    struct sH_temp *next_sublist;   // link to next sublist (NULL if none)
    unsigned long max_elements;     // number of elements that this sublist can hold in the space that was allocated
} sublistHeader;

typedef struct {                    // all these have to do with the whole linked list
    unsigned long elementnum;       // number of elements
    sublistHeader *memory;          // first sublist pointer
    unsigned short elementsize;     // bytes per element
    unsigned char spare_room;       // _percentage_ of extra room allocated (for speed if elements are added)
} linkedlist;


// stack -- a kind of list used for PCStack

typedef struct {
    Ulong top;
    linkedlist data;
} stack;


// pinned_LL -- used when we don't want the location of existing elements to change when others are inserted/deleted.
// Used by, e.g., VariableList, all bundles, fibrils, etc. -- anything we take pointers from.

typedef struct {
    linkedlist data;            // the actual data
    linkedlist references;      // # of pointers of each element in use -- unsigned longs
    Ulong lowest_free_spot;     // the first and last elements in the LL that are unallocated (not deleted, since we need
    Ulong highest_free_spot;    // to keep memory in the same spot (since variables hold mem_pointers))
} pinned_LL;



// Prototypes

extern LL_retrn NewLinkedList(linkedlist *, unsigned long, unsigned short, unsigned char, Boolean);
extern LL_retrn DeleteLinkedList(linkedlist *);
extern LL_retrn InsertElements(linkedlist *, unsigned long, unsigned long, Boolean);
extern LL_retrn AddElements(linkedlist *, unsigned long, Boolean);
extern LL_retrn DeleteElements(linkedlist *, unsigned long, unsigned long);
extern LL_retrn DeleteElement(linkedlist *, unsigned long);
extern LL_retrn DefragmentLinkedList(linkedlist *);
extern LL_retrn CopyElements(linkedlist *, unsigned long, linkedlist *, unsigned long, unsigned long);
extern LL_retrn CompareElements(linkedlist *, unsigned long, linkedlist *, unsigned long, unsigned long);
extern LL_retrn FillElements(linkedlist *, unsigned long, unsigned long, char);
extern LL_retrn ClearElements(linkedlist *, unsigned long, unsigned long);
extern LL_retrn SetElements(linkedlist *, unsigned long, unsigned long, void *);
extern LL_retrn SetElement(linkedlist*, unsigned long, void *);
extern LL_retrn GetElements(linkedlist *, unsigned long, unsigned long, void *);
extern LL_retrn GetElement(linkedlist *, unsigned long, void *);
extern LL_retrn ElementExists(linkedlist *, unsigned long);
extern void *Element(linkedlist *, unsigned long);

extern LL_retrn FNewLinkedList(linkedlist *, unsigned long, unsigned short, unsigned char, Boolean);
extern void FDeleteLinkedList(linkedlist *);
extern LL_retrn FInsertElements(linkedlist *, unsigned long, unsigned long, Boolean);
extern LL_retrn FAddElements(linkedlist *, unsigned long, Boolean);
extern LL_retrn FDeleteElements(linkedlist *, unsigned long, unsigned long);
extern LL_retrn FDeleteElement(linkedlist *, unsigned long);
extern LL_retrn FDefragmentLinkedList(linkedlist *);
extern void FCopyElements(linkedlist *, unsigned long, linkedlist *, unsigned long, unsigned long);
extern LL_retrn FCompareElements(linkedlist *, unsigned long, linkedlist *, unsigned long, unsigned long);
extern void FFillElements(linkedlist *, unsigned long, unsigned long, char);
extern void FClearElements(linkedlist *, unsigned long, unsigned long);
extern void FSetElements(linkedlist *, unsigned long, unsigned long, void *);
extern void FSetElement(linkedlist*, unsigned long, void *);
extern void FGetElements(linkedlist *, unsigned long, unsigned long, void *);
extern void FGetElement(linkedlist *, unsigned long, void *);
extern LL_retrn FElementExists(linkedlist *, unsigned long);
extern void *FElement(linkedlist *, unsigned long);
extern void *FSkipElements(linkedlist *, sublistHeader **, unsigned long *, unsigned long);    // no equivalent SkipElements

extern LL_retrn NewPLL(pinned_LL *, Ulong, Ulong, unsigned char);
extern void DeletePLL(pinned_LL *);
extern LL_retrn AddPLLElement(pinned_LL *, void **, Ulong *);
extern void RefPLLElement(pinned_LL *, Ulong);
extern Ulong DerefPLLElement(pinned_LL *, Ulong);
#define RefPLLPtr(ref_ptr) ((*ref_ptr)++)
extern void DerefPLLPtr(pinned_LL *, Ulong, Ulong *);

extern LL_retrn NewStack(stack *, Ulong, Ulong, unsigned char);
extern void DeleteStack(stack *);
extern LL_retrn PushStack(stack *, void **);
extern LL_retrn PopStack(stack *, void **);

#endif
